

from math import *
from extra_functions import *

from svm import *

from texture import *

from gtk import *
from extra_widgets import *




class AttributeSelector:
  """
  3 formats:
  - dictionary of (string,boolean)=(attr name,wether it is used)
  - list of attr names
  - list of expanded attr names (tags and camera models are exapanded)
  corresponding to actual SVM values  
  """

  fixed_attributes={
    'tags':True,
    'npixel':True,
    'exif_model':True,
    'exif_date':True,
    'nperson':True
    }

  def __init__(self):
    self.prev_pa={}

  def possible_attributes_dict(self):
    pa={}
    for (k,v) in AttributeSelector.fixed_attributes.iteritems():
      try:
        pa[k]=self.prev_pa[k]
      except KeyError:
        pa[k]=v

    texlas=allTexlas.get_classifier_dict()
    for (name,v) in texlas.iteritems():
      k='tex_'+name
      try:
        pa[k]=self.prev_pa[k]
      except KeyError:
        pa[k]=True
    return pa      

  def get_attribute_list(self):
    l=[name for (name,b) in self.possible_attributes_dict().items() if b]
    l.sort()
    return l
    
  def ask_attribute_list(self):
    pa=self.possible_attributes_dict().items()
    pa.sort()
    ret=ask_checkarray("attributes to include...",pa)
    if ret==None:
      return None
    self.prev_pa={}
    al=[]
    for i in range(len(pa)):
      (k,prev_b)=pa[i]
      b=ret[i]
      self.prev_pa[k]=b
      if b:
        al.append(k)
    return al

  def expand_attribute_list(self,attribute_list,all_im):
    exp_attr_list=[]
  
    for name in attribute_list:
      if name=='tags':
        # collect tags
        tl={}
        for im in all_im:
          for tag in im.tags:
            tl[tag]=None
        tl=tl.keys()
        tl.sort()
        for tag in tl:
          exp_attr_list.append('tag_'+tag)
      elif name=='exif_model':
        # collect models
        models={}
        for im in all_im:
          ei=im.get_exif_info()
          if ei:
            models[ei.model]=None
        models=models.keys()
        models.sort()
        for model in models:
          exp_attr_list.append('exif_model_'+model)
      else:
        exp_attr_list.append(name)
        
    return exp_attr_list
  
  def extract_attributes(self,exp_attribute_list,all_im,progress=None):
    texlas=allTexlas.get_classifier_dict()
    
    i=0
    all_attrs=[]
    for im in all_im:
      if progress:
        progress(i/float(len(all_im)))
        i+=1

      what=im.what_computed()

      attrs=[]

      pl=None
      
      for name in exp_attribute_list:
        # print "handline ",name
        if name=='npixel':
          attrs.append(im.get_npix())
        elif name=='nperson':
          attrs.append(len(im.get_persons()))
        elif beginswith(name,'tag_'):
          if name[4:] in im.tags:
            attrs.append(1.0)
          else:
            attrs.append(0.0)
        elif beginswith(name,'exif_'):
          ei=im.get_exif_info()
          if ei:
            if name=='exif_date':
              attrs.append(ei_date_to_secs(ei.date))
            elif beginswith(name,'exif_model_'):
              if ei.model==name[11:]:
                attrs.append(1.0)
              else:
                attrs.append(0.0)
            else: assert False
          else:            
            attrs.append(None)
        elif beginswith(name,'tex_'):
          texname=name[4:]
          cl=texlas[texname]
          imno=im.get_tex_imno()
          f_prop=textureManager.get_cache_entry(cl,imno)
          if f_prop=='not in cache':
            if pl==None: pl=textureManager.get_tex(imno)
            f_prop=textureManager.classify_patches(cl,pl)
            textureManager.add_cache_entry(cl,imno,f_prop)
          attrs.append(f_prop)
        else: assert False
        
      if pl!=None: pl.clear()
      all_attrs.append(attrs)
      
    return all_attrs
    


attributeSelector=AttributeSelector()


def find_scale_facts(samples):
  n=len(samples[0])
  maxes=samples[0][:]
  mins=samples[0][:]
  for s in samples[1:]:
    for i in range(n):
      if maxes[i]<s[i]: maxes[i]=s[i]
      if mins[i]>s[i]: mins[i]=s[i]
  as=[]
  bs=[]
  for i in range(n):    
    if maxes[i]==mins[i]:
      (a,b)=(1.0,0.0)
    else:
      a=1.0/(maxes[i]-mins[i])
      b=-a*mins[i]
    as.append(a)
    bs.append(b)
  return (as,bs)

def scale_all(all_attrs,(as,bs)):
  n=len(as)
  for s in all_attrs:
    for i in range(n):
      s[i]=as[i]*s[i]+bs[i]



def prepare_attrs(all_im,pos_im,neg_im):
  n=len(all_im)
  
  al=attributeSelector.ask_attribute_list()
  if al==None: return

  exp_al=attributeSelector.expand_attribute_list(al,pos_im+neg_im)

  m=len(exp_al)

  print "selected attributes:", exp_al

  pb=ProgressBarWindow("computing attributes",10)
  train_attrs=attributeSelector.extract_attributes(exp_al,pos_im+neg_im,
                                                   pb.set_fraction)
  all_attrs=attributeSelector.extract_attributes(exp_al,all_im,pb.set_fraction)
  pb.destroy()

  # replace unknown values with 0
  for im_attrs in train_attrs+all_attrs:
    for j in range(m):
      if im_attrs[j]==None:
        im_attrs[j]=0.0
  
  # scale
  scale_facts=find_scale_facts(train_attrs)

  print "scale factors=",scale_facts

  scale_all(train_attrs+all_attrs,scale_facts)

  print "train_attrs=",train_attrs
  # print "all_attrs=",all_attrs

  return (train_attrs,all_attrs)


svm_t=0



def learn_svm_model(pos_im,neg_im,train_attrs):
  
  print "learning"

  # labels
  # 0 = negative
  # 1 = positive
  if svm_t==0:
    labels=[1]*len(pos_im)+[0]*len(neg_im)
  elif svm_t==1:
    labels=[1]*len(pos_im)+[-1]*len(neg_im)
    
  print "labels=",labels

  problem = svm_problem(labels,train_attrs)

  if svm_t==0:
    param = svm_parameter(
      kernel_type=RBF,C = 1000,nr_weight = 0)
  elif svm_t==1:
    param = svm_parameter(
      svm_type=EPSILON_SVR,
      kernel_type=RBF,C = 1000,nr_weight = 0)

  # print svm_check_parameter(problem,param)
  
  model = svm_model(problem,param)

  # should check if model ok on train img

  return model


def classify_with_model(model,pos_im,neg_im,all_im,all_attrs):
  n=len(all_im)
  
  print "classifying"


  # print "labels=",model.get_labels()

  im_with_label=[]
  
  for i in range(n):
    # print 'class',im
    im=all_im[i]
    im_sam=all_attrs[i]
        
    #probas=model.predict_probability(im_sam)
    #print probas


    if svm_t==0:
      l=model.predict_values(im_sam)
      flabel=l[(1,0)]
    elif svm_t==1:
      flabel=model.predict(im_sam)
      print flabel

    # print im,label

    if im in pos_im:    gt_label=1
    elif im in neg_im:  gt_label=0
    else:               gt_label=-1

    if gt_label!=-1:
      print gt_label,flabel
      #if gt_label!=label:
      #  print "image misclassified",im

    im_with_label.append((flabel,im))

  im_with_label.sort()
  im_with_label.reverse()

  return im_with_label
  
def classify_with_attrs(all_im,pos_im,neg_im):
  x=prepare_attrs(all_im,pos_im,neg_im)
  if x==None: return 
  (train_attrs,all_attrs)=x

  model=learn_svm_model(pos_im,neg_im,train_attrs)
    
  return classify_with_model(model,pos_im,neg_im,all_im,all_attrs)

