

import gtk
from math import *
import time
import calendar
import random,os,stat,errno,pdb,struct



class Object:
  pass



def packs(l,n):
  # splits list l into lists of maximum size n
  l2=[]; l=l[:]
  while len(l)>n:
    l2.append(l[:n])
    del l[:n]
  if l: l2.append(l)
  return l2

def clusterize(l):
  a={}
  for key,i in l:
    try:
      a[key].append(i)
    except KeyError:
      a[key]=[i]
  keys=a.keys()
  keys.sort()
  clusters=[]
  for key in keys:
    clusters.append((str(key),a[key]))
  return clusters


def robust_date(ei):
  l=ei.date
  if l==None: return "No date"
  return l[:10]


def common_prefix(l):
  pref=l[0]
  for s in l[1:]:
    i=0
    while i<len(s) and i<len(pref) and s[i]==pref[i]:
      i+=1
    pref=pref[:i]
  return pref

def escape_sh(s):
  a=""
  for i in s:
    if i in ' \\"\'':
      a+='\\'+i
    else:
      a+=i
  return a

def rgb_to_luminance((r,g,b)):
  return 0.29900*r+0.58700*g+0.11400*b

# from ImageMagick's TransformHSL (magick/gem.c)

def rgb_to_hsl((red,green,blue)):
  r=red/255.0
  g=green/255.0
  b=blue/255.0
  vmax=max(r,g,b);
  vmin=min(r,g,b);
  hue=0.0;
  saturation=0.0;
  luminosity=((vmin+vmax)/2.0);
  delta=vmax-vmin;
  if delta == 0.0: return (0.0,0.0,luminosity)
  if luminosity < 0.5: den=vmin+vmax
  else: den=2.0-vmax-vmin
  saturation=delta/den
  if r == vmax:
    if g==vmin:
      hue=5.0+(vmax-b)/delta
    else:
      hue=7.0-(vmax-g)/delta
  elif g == vmax:
    if b==vmin:
      hue=1.0+(vmax-r)/delta
    else:
      hue=3.0-(vmax-b)/delta
  else:
    if r==vmin:
      hue=3.0+(vmax-g)/delta
    else:
      hue=5.0-(vmax-r)/delta
  # hue is 1-7
  return (hue,saturation,luminosity)




def shrink_big_image(im,maxpix,interp=gtk.gdk.INTERP_NEAREST):
  xd=im.get_width()
  yd=im.get_height()
  if xd*yd<maxpix:
    return im
  else:
    scale=sqrt(maxpix/float(xd*yd))
    t_xd=int(scale*xd)
    if t_xd==0: t_xd=1
    t_yd=int(scale*yd)
    if t_yd==0: t_yd=1
    return im.scale_simple(t_xd,t_yd,interp)


def breakup_cross_rels(mat,thr,lt=False):
  n=len(mat)
  eqv=range(n)
  for i,l in enumerate(mat):
    for j,v in enumerate(l):
      if (v>thr) ^ lt:
        # i is equivalent to j
        if i<j:
          if eqv[j]>i: eqv[j]=i
        else:
          if eqv[i]>j: eqv[i]=j
  tab=[None]*n
  for i,l in enumerate(eqv):
    if i==l:
      tab[i]=[i]
    else:
      base=l
      while eqv[base]!=base:
        base=eqv[base]
      tab[base].append(i)
  return [v for v in tab if v!=None]


    
      
  
def complete_sym_mat(m):
  n=len(m)
  m2=[]
  for i in range(n):
    l=m[i][:]
    l.append(0)
    for j in range(i+1,n):
      l.append(m[j][i])
    m2.append(l)
  return m2



def dump_cross_rels(fname,cross_rels):
  f=open(fname,"w")
  f.write("%d\n"%len(cross_rels))
  for l in cross_rels:
    f.write(reduce(lambda accu,x: accu+" %g"%x,l,"")+"\n")
  f.close()



def segs_inter(min0,max0,min1,max1):
  return not(max0<min1 or max1<min0)


def rectangles_inter(xmin0,ymin0,xmax0,ymax0,
                     xmin1,ymin1,xmax1,ymax1):
  return (segs_inter(xmin0,xmax0,xmin1,xmax1) and
          segs_inter(ymin0,ymax0,ymin1,ymax1))


def uniquify_list(l):
  ht={}
  ol=[]
  for i in l:
    if not ht.has_key(i):
      ol.append(i)
      ht[i]=None
  return ol

import mds

      
def cross_rels_to_mds_pos(cross_rels,transform=lambda x:x):
  natoms=len(cross_rels)
  tau=mds.DoubleArray(natoms*natoms)
  x=mds.DoubleArray(natoms*2)
   
  for i in range(natoms):
    l=cross_rels[i]
    for j in range(natoms):
      tau[i+j*natoms]=transform(l[j])
      
  # dump distances 

##  f=open("mds/cross_rels.dat","w")
##  f.write("%d\n"%natoms)
##  for i in range(natoms):
##    l=cross_rels[i]
##    for j in range(natoms):
##      f.write("%g "%l[j])
##    f.write("\n")

  info=mds.multi_dimensional_scaling(natoms,2,20,tau,x,None)
  assert info in (0,1)
  mds_pos=[]
  for i in range(natoms):
    mds_pos.append((x[i],x[i+natoms]))
  return mds_pos

def cross_rels_to_lapeig_pos(cross_rels,transform=lambda x:x):
  natoms=len(cross_rels)
  tau=mds.DoubleArray(natoms*natoms)
  x=mds.DoubleArray(natoms*2)
  for i in range(natoms):
    l=cross_rels[i]
    for j in range(natoms):
      tau[i+j*natoms]=transform(l[j])
  mds.laplacian_eigenmap(natoms,2,20,tau,x)
  pos=[]
  for i in range(natoms):
    pos.append((x[i],x[i+natoms]))
  return pos

def get_reasonable_order(cross_rels,transform=lambda x:x):
  natoms=len(cross_rels)
  tau=mds.DoubleArray(natoms*natoms)  
  for i in range(natoms):
    l=cross_rels[i]
    for j in range(natoms):
      tau[i+j*natoms]=transform(l[j])
  perm=mds.IntArray(natoms)
  mds.one_dim_matrix_permutation(natoms,tau,perm)
  pperm=[]
  for i in range(natoms):
    pperm.append(perm[i])
  return pperm


# scale a pixbuf so that it becomes smaller than maxpix pixels
def scale_pixbuf_maxpix(pixbuf,maxpix):
  xd=pixbuf.get_width(); yd=pixbuf.get_height()
  if xd*yd<maxpix:
    scale=1.0
  else:
    scale=sqrt(maxpix/float(xd*yd))
    t_xd=int(scale*xd)
    if t_xd==0: t_xd=1
    t_yd=int(scale*yd)
    if t_yd==0: t_yd=1
    pixbuf=pixbuf.scale_simple(t_xd,t_yd,gtk.gdk.INTERP_BILINEAR)
    
  return (scale,pixbuf)


# simple format to save thumbnails in image.dat

def savePixbuf(pixbuf,f):
  s=struct.pack("3i",
                pixbuf.get_width(),
                pixbuf.get_height(),
                pixbuf.get_rowstride())
  f.write(s)
  f.write(pixbuf.get_pixels())


def loadPixbuf(f):
  s=f.read(struct.calcsize("3i"))
  (w,h,rs)=struct.unpack("3i",s)
  return gtk.gdk.pixbuf_new_from_data(f.read(h*rs),gtk.gdk.COLORSPACE_RGB,
                               False,8,w,h,rs)


def ei_date_to_secs(s):
  # date/time handling is a bit of a mess in the python std lib.
  return calendar.timegm(time.strptime(s,"%Y:%m:%d %H:%M:%S"))


def make_tmpdir():
  base=os.getcwd()+"/descs"
  n=1
  while True:
    fname=base+"/%05d"%n
    try:
      os.mkdir(fname)
      return fname
    except OSError,e:
      if e.errno!=errno.EEXIST:
        raise e
    n+=1


def obj_to_dict(obj,attrs):
  dict={}
  for attr in attrs:
    try:
      dict[attr]=getattr(obj,attr)
    except AttributeError:
      pass
  return dict


def dict_to_obj(dict,obj):
  for k,v in dict.items():
    setattr(obj,k,v)

def beginswith(s,subs):
  return s[:len(subs)]==subs

def aligned_v((x0,y0),(x1,y1)):
  return x0*y1==x1*y0

def aligned_pt((x0,y0),(x1,y1),(x2,y2)):
  return aligned_v((x1-x0,y1-y0),(x2-x0,y2-y0))
  
def simplify_curve(l):
  i=0
  while i+2<len(l):
    if aligned_pt(l[i],l[i+1],l[i+2]):
      del l[i+1]
    else:
      i+=1

def clamp_curve(l,xd,yd):
  for i in range(len(l)):
    (x,y)=l[i]
    l[i]=(clamp(0,x,xd),clamp(0,y,yd))
  simplify_curve(l)

class ChangeBroadcaster:

  def __init__(self):
    self.callbacks=[]

  def register(self,f,msg=None):
    self.callbacks.append((f,msg))

  def unregister(self,f,msg=None):
    self.callbacks.remove((f,msg))    

  def broadcast(self,msg=None,*args):
    for (f,qmsg) in self.callbacks:
      if qmsg==None:
        f(*args)
      elif msg==None or msg==qmsg:
        f(msg,*args)

def random_split(l,n1):
  l1=[]
  l2=[]
  prev=-1
  s=random.sample(range(len(l)),n1); s.sort()
  # print s
  for i in s:
    l1.append(l[i])
    l2+=l[prev+1:i]
    prev=i
  l2+=l[prev+1:]
  return (l1,l2)

def clamp(vmin,v,vmax):
  if v<vmin: return vmin
  if v>vmax: return vmax
  return v

def list_to_set(b):
  bh={}
  for e in b: bh[e]=None
  return bh


def list_sub(a,b):
  """
  a' = a minus b
  return a inter b
  """
  bh=list_to_set(b)
  i=0
  ret=[]
  while i<len(a):
    if bh.has_key(a[i]):
      ret.append(a[i])
      del a[i]
    else:
      i+=1
  return ret
  

def list_add_unique(a,b):
  """
  a'=a+(b-a)
  """
  ah=list_to_set(a)
  a+=[bv for bv in b if not ah.has_key(bv)]


def list_inter(a,b):
  """ return a inter b """
  bh=list_to_set(b)
  return [av for av in a if bh.has_key(av)]

def issamefile(f1,f2):
  try:
    s1=os.stat(f1)
    s2=os.stat(f2)
  except OSError,e:
    if e.errno==errno.ENOENT:
      return False
    else:
      raise e
  # actually, all fields seem to be the same
  return s1.st_ino==s2.st_ino and s1.st_dev==s2.st_dev


def move_dir(src,dst):
  try:
    os.stat(dst)
  except OSError,e:
    if e.errno==errno.ENOENT:
      pass # perfect
    else:
      raise e
  else:
    raise "%s exists already !!"%dst
  # print "rename %s %s"%(src,dst)
  try:
    os.rename(src,dst)
  except OSError,e:
    if e.errno==errno.EXDEV: #Invalid cross-device link
      return False
    else:
      raise e
  # print "moved %s to %s"%(src,dst)
  return True

def copy_file(src,dst):
  open(dst,"w").write(open(src,"r").read())
  #os.system("ls -la %s %s"%(src,dst))


def copy_dir(src,dst):
  assert src[-1]=='/' and dst[-1]=='/'
  os.mkdir(dst)
  for fn in os.listdir(src):
    s=os.stat(src+fn)
    if stat.S_ISREG(s.st_mode):
      copy_file(src+fn,dst+fn)
  



class DataDir:
  """ manages a set of datafiles that must be saved along with the
  yorg file"""
  
  def __init__(self,dirkey):
    self.dirprefix=make_tmpdir()+'/'
    self.istmpdir=True
    self.dirkey=dirkey
    
  def fname(self,i):
    return self.dirprefix+"%08d.%s"%(i,self.dirkey)

  def make_dirprefix(self,fname):
    dstdir=fname[:fname.rfind('/')] # strip
    return dstdir+'/%s/'%self.dirkey
       
  def prepare_store(self,fname):
    """
    fname is like xxxx/yyyy.yorg/filename
    datafiles will be saved to xxxx/yyyy.yorg/<ty>/
    Three possibilites:
    1. they are already in the correct directory
      ==> nothing to do
    2. they are in a temp directory (from make_tmpdir())
      ==> they should be moved
    3. they are in another yorg directory
      ==> they should be copied
    """
    newdirprefix=self.make_dirprefix(fname)
    if issamefile(self.dirprefix,newdirprefix):
      pass
    elif self.istmpdir:
      if move_dir(self.dirprefix,newdirprefix):
        pass
      else:
        copy_dir(self.dirprefix,newdirprefix)
        print "warning, useless files left in temp dir",self.dirprefix
    else:
      # copy from other yorg
      copy_dir(self.dirprefix,newdirprefix)    
    self.dirprefix=newdirprefix
    self.istmpdir=False

def demangle_version_str(v):
  return [int(x) for x in v.split('.')]



if __name__=='__main__':
  print "testing datadir"
  dd=DataDir("yellow")
  print "dir:",dd.dirprefix
  open(dd.fname(15),"w").write("cocuou 15\n")
  open(dd.fname(12),"w").write("cocuou 12\n")
  os.system("rm -rf /tmp/xx.yorg /tmp/yy.yorg")
  os.mkdir("/tmp/xx.yorg")
  pdb.set_trace()
  dd.prepare_store("/tmp/xx.yorg/yd.dat")
  print "dir:",dd.dirprefix  
  os.mkdir("/tmp/yy.yorg")
  dd.prepare_store("/tmp/yy.yorg/yd.dat")
  print "dir:",dd.dirprefix

  


