
import gtk,random,gobject,pdb,time



class ColoredWidget(gtk.DrawingArea):
  def __init__(self,color=None):
    gtk.DrawingArea.__init__(self)
    if color==None:
      color=gtk.gdk.Color(random.randint(0,65536),
                          random.randint(0,65536),
                          random.randint(0,65536))
    self.modify_bg(gtk.STATE_NORMAL,color)



class HeightFromWidthWidget(gtk.Frame):
  """
  Keeps a constant aspect ratio ar=height/width.
  Implementaton requires 2 request-allocate chains
  """
  
  def __init__(self,ar=1.0):
    gtk.Frame.__init__(self)
    self.ar=ar
    self.h=0; self.tried_h=-1    
    self.connect('size-request',self.on_size_request)
    self.connect('size-allocate',self.on_size_allocate)

  def set_ar(self,ar):
    if self.ar!=ar:
      self.ar=ar
      self.queue_resize()   
    
  def on_size_request(self,widget,requisition):
    requisition.width=0
    requisition.height=self.h
    self.tried_h=self.h

  def on_size_allocate(self,widget, allocation):
    w=allocation.width
    h=int(allocation.width*self.ar)
    if allocation.height!=h and not (allocation.height>h and h==self.tried_h):
      self.h=h
      self.queue_resize()

class AltFrames(gtk.Frame):
  """ pseudo-container that displays one child at a time, according to a
  combo menu
  """
  __gsignals__ = {
    'switch' : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
                 (gobject.TYPE_INT,))
    }
  def __init__(self,title):
    gtk.Frame.__init__(self)    
    c1=gtk.VBox()
    self.add(c1)
    l1=gtk.HBox()
    c1.pack_start(l1,False)
    lab=gtk.Label(title)
    l1.pack_start(lab,False)
    self.cb=gtk.combo_box_new_text()
    l1.pack_start(self.cb,True)
    self.fr=gtk.Frame()
    c1.add(self.fr)
    self.widgets=[]
    self.cb.connect('changed', self.cb_changed)

  def cb_changed(self,cb):
    new_n=self.cb.get_active()
    self.switchto(new_n)
    self.emit('switch',new_n)

  def get_widget(self,no=-1):
    if no==-1:
      return self.widgets[self.cb.get_active()]
    else:
      return self.widgets[no]
    
  def add_widget(self,name,widget):
    self.cb.append_text(name)
    self.widgets.append(widget)
    if len(self.widgets)==1:
      self.switchto(0)

  def switchto(self,n):    
    self.cb.set_active(n)
    oldw=self.fr.get_child()
    if oldw: self.fr.remove(oldw)
    self.fr.add(self.widgets[n])
    self.fr.show_all() # force redraw

gobject.type_register(AltFrames)


class ProgressBarWindow(gtk.Window):
  """modal window with progress bar that calls back a function nstep times
  each call blocks the application, so it should be short (<1s). This avoids
  the use of multithreading"""   
  
  def __init__(self,title,nstep,callback=None,extrawidget=None):
    gtk.Window.__init__(self)
    self.set_title(title)    
    c=gtk.VBox()
    if extrawidget:
      # display of extra widget is always 1 step late due to the redisplay     
      c.add(extrawidget)
    self.pb=gtk.ProgressBar()
    c.add(self.pb)
    bs=gtk.Button("cancel")
    c.add(bs)
    self.add(c)
    bs.connect("clicked",self.stop_clicked)
    self.set_modal(True)
    self.show_all()
    self.nstep=nstep
    self.callback=callback
    self.process_events()
    self.cur=0
    self.stop=False
    self.t0=time.time()

  def print_time(self):
    print "ProgressBar total time:",time.time()-self.t0
    
  def loop(self):
    "returns nb of steps effectively processed (=no of last step+1)"
    if self.nstep==0:
      self.destroy()
      return 0
    for i in range(self.nstep):
      self.callback(i)
      self.pb.set_fraction(i/float(self.nstep))
      self.process_events()
      if self.stop:
        break
    self.destroy()
    self.process_events()
    return i+1

  def one_step(self):
    self.cur+=1
    self.pb.set_fraction(self.cur/float(self.nstep))
    self.process_events()
    return not self.stop

  def set_fraction(self,v):
    self.pb.set_fraction(v)
    self.process_events()
    if self.stop:
      print "stop ignored"
      self.stop=False
      
  def process_events(self):
    "flush event loop"
    while gtk.events_pending():
      gtk.main_iteration_do(False)
    
  def stop_clicked(self,arg):
    self.stop=True



def parse_submenu(desc):
  menu=gtk.Menu()
  for d in desc:
    if d=='_':
      menu.append(gtk.MenuItem())
    else:
      mi=gtk.MenuItem(d[0])
      # handle shortcut d[1]
      if type(d[2])==type([]):
        subm=parse_submenu(d[2])
        mi.set_submenu(subm)
      else:
        if len(d)==3:
          mi.connect("activate",d[2])
        else:
          mi.connect("activate",d[2],d[3])
      menu.append(mi)
  return menu

def build_menubar(desc):
  """ builds a menubar. Shortcuts not handled yet"""
  b1=gtk.MenuBar()
  for d in desc:
    if type(d[2])==type([]):
      subm=parse_submenu(d[2])
      mi=gtk.MenuItem(d[0])
      mi.set_submenu(subm)
      # handle shortcut d[1]
      b1.append(mi)
    else:
      raise "bad menu type"
  return b1

class RadioButtons(gtk.VBox):
  def __init__(self,buts,callback):
    gtk.VBox.__init__(self)
    ref=None
    self.buttons=[]
    self.cb=callback
    for text in buts:
      if text[:2]=='_L':
        self.pack_start(gtk.Label(text[2:]),False)
      else:
        b=gtk.RadioButton(ref,text)
        if not ref: ref=b
        self.pack_start(b,False)
        b.connect("toggled",self.on_toggled,len(self.buttons))
        self.buttons.append(b)
        
  def set(self,no):
    self.buttons[no].set_active(True)

  def on_toggled(self,source,no):
    if source.get_active():
      self.cb(no)



def file_chooser(title,action):
  if action==gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER:
    buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
             gtk.STOCK_OPEN, gtk.RESPONSE_OK)
  elif action==gtk.FILE_CHOOSER_ACTION_SAVE:
    buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
             gtk.STOCK_SAVE, gtk.RESPONSE_OK)    
  elif action==gtk.FILE_CHOOSER_ACTION_OPEN:
    buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
             gtk.STOCK_OPEN, gtk.RESPONSE_OK)    
  fc=gtk.FileChooserDialog(title,None,action,buttons)
  response = fc.run()
  if response != gtk.RESPONSE_OK:
    name=None
  else:
    name=fc.get_filename()
  fc.destroy()
  return name


def make_button(title,callback):
  b=gtk.Button(title)
  b.connect('pressed',callback)
  return b


class TagList(gtk.ScrolledWindow):
  """
  gtk.TreeView is slightly overengineered, hence this simple interface
  tags is just a list of strings
  """  
  def __init__(self,colnames=['tags']):
    gtk.ScrolledWindow.__init__(self)
    
    self.set_policy(gtk.POLICY_NEVER,gtk.POLICY_AUTOMATIC)
    model =apply(gtk.ListStore,(gobject.TYPE_STRING,)*len(colnames))
    
    # the tree view
    tv = gtk.TreeView(model)
    self.add(tv)
    self.treeview=tv
    # selection
    tv.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
    tv.set_headers_clickable(True)
    # column
    for no,coln in enumerate(colnames):
      cell = gtk.CellRendererText()
      column = gtk.TreeViewColumn(coln, cell, text=no)
      column.set_expand(no==0)
      column.set_clickable(True)
      tv.append_column(column)
      column.connect("clicked",self.col_clicked)
    # model.set_sort_column_id(0,gtk.SORT_ASCENDING) 
    
  def set_tags(self,tags):
    model=self.treeview.get_model()
    model.clear()
    for tag in tags:
      if type(tag)==type(''): tag=(tag,)
      model.append(tag)

  def col_clicked(self,treeviewcol):
    print "cl",treeviewcol,
    treeviewcol.set_sort_column_id(treeviewcol.get_sort_column_id())
    
  def get_selected_tags(self):
    sel=self.treeview.get_selection()
    (ls,rows)=sel.get_selected_rows()    
    return [ls[row][0] for (row,) in rows]

  def handle_double_click(self,cb):
    self.double_click_cb=cb
    self.treeview.connect("row-activated",on_row_activated)
    
  def double_click_callback(self,cb):
    self.double_click_cb=cb
    self.treeview.connect("row-activated",self.on_row_activated)
    
  def on_row_activated(self,treeview,path,view_column):
    self.double_click_cb(treeview.get_model()[path[0]][0])

  def select_no(self,no=0):
    # select the first one
    #model=self.treeview.get_model()
    self.treeview.get_selection().select_path((no,))


class SingleSelectTagList(TagList):
  """ there is always a single tag selected, initially tag 0 """

  def __init__(self,heads,init_tags):
    TagList.__init__(self,heads)
    self.handler_no=-1

    sel=self.treeview.get_selection()
    
    sel.set_mode(gtk.SELECTION_SINGLE)
    
    model=self.treeview.get_model()
    # model.set_sort_column_id(0,gtk.SORT_ASCENDING)     
    self.in_on_select=False
    self.selh=sel.connect("changed",self.on_select)
    self.set_tags_and_sel(init_tags,0)
    self.select_tag_cb=None
    
  def get_selected_no(self):
    return self.treeview.get_selection().get_selected_rows()[1][0][0]

  def set_select_tag_cb(self,f):
    self.select_tag_cb=f

  def set_tags_and_sel(self,tags,sel_no):
    assert len(tags)>0 and sel_no<len(tags)
    # do not call this from the select_tag callback or weird thing will happen
    assert not self.in_on_select
    sel=self.treeview.get_selection()
    sel.handler_block(self.selh)
    self.set_tags(tags)
    self.treeview.get_selection().select_path((sel_no,))
    sel.handler_unblock(self.selh)
    
  def on_select(self,sel):
    (ls,rows)=sel.get_selected_rows()
    # pdb.set_trace()
    self.in_on_select=True
    if self.select_tag_cb:
      self.select_tag_cb(rows[0][0])
    self.in_on_select=False
                       

def ask_string(prompt):
  buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
           gtk.STOCK_OK, gtk.RESPONSE_OK)
  win=gtk.Dialog(prompt,None,gtk.DIALOG_MODAL,buttons)
  en=gtk.Entry()
  win.vbox.add(en)
  win.show_all()
  response = win.run()
  if response != gtk.RESPONSE_OK:
    win.destroy()
    return None
  name=en.get_text()
  win.destroy()
  return name  
  

  
def message(st,modal=True):
  if modal: flags=gtk.DIALOG_MODAL
  else: flags=0
  dialog = gtk.MessageDialog(
    None, flags,
    gtk.MESSAGE_INFO, gtk.BUTTONS_CLOSE,
    st)
  dialog.connect('response', lambda dialog, response: dialog.destroy())
  dialog.show()

  
class SilentCheckButton(gtk.CheckButton):
  """ adds a set_active that does not trigger the clicked signal """

  def connect(self, detailed_signal, handler, *args):
    no=gtk.CheckButton.connect(self,detailed_signal, handler, *args)
    if detailed_signal=="clicked":
      self.cn=no
      self.cn_handler=handler
      self.cn_args=args
    return no
      
  def set_active_silent(self,b):
    self.disconnect(self.cn)
    self.set_active(b)
    self.cn=gtk.CheckButton.connect(self,"clicked",self.cn_handler,*self.cn_args)
    


def ask_checkarray(prompt,checks):
  buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
           gtk.STOCK_OK, gtk.RESPONSE_OK)
  win=gtk.Dialog(prompt,None,gtk.DIALOG_MODAL,buttons)
  buts=[]
  for v in checks:
    if type(v)==type(''):
      win.vbox.add(gtk.Label(v))      
    else:
      (name,default)=v
      but=gtk.CheckButton(name,False)
      but.set_active(default)
      buts.append(but)
      win.vbox.add(but)

  win.show_all()
  response = win.run()
  if response != gtk.RESPONSE_OK:
    win.destroy()
    return None
  ret=[but.get_active() for but in buts]
  win.destroy()
  return ret


gtk_color_black=gtk.gdk.color_parse('Black')
gtk_color_white=gtk.gdk.color_parse('White')




