
from math import *


have_photomole=True

try:
  from PhotoMole import *
except ImportError:
  print "PhotoMole not available. Using dummy geometry."
  have_photomole=False
  
import gtk,cPickle,pdb,time

from extra_functions import *
from extra_widgets import *

import TextureClassifier


def PixbufToGrayImage(pixbuf):
  lava_im=GrayImage(pixbuf.get_width(),pixbuf.get_height())
  TextureClassifier.colorImageToDoubleArray(pixbuf,lava_im.getData())
  return lava_im



class GeometryIndex(DataDir):

  def __init__(self):
    DataDir.__init__(self,"geometry")
    self.clear(self.dirprefix)

  def clear(self,tmpprefix):
    self.mf=MatchingDataBase(tmpprefix,5)
    ak=self.mf.getHighDimIndex()
    assert ak!=None
    from canonic_ak import acpk_data
    bs=BufferStream(Stream.INPUT,acpk_data)
    ak.load(bs)
    self.mf.syncStatus()

  def fname(self,i):
    return self.dirprefix+"%08d.corns"%i
            
  def add_image(self,pixbuf,filename):
    # unfortunately, the geomindex must be invalidated because
    # the image is not taken into account.
    crossRelations.reset()
    # may need resize
    (scale,pixbuf)=scale_pixbuf_maxpix(pixbuf,640*480)
    lava_im=PixbufToGrayImage(pixbuf)    
    # detect points and compute descriptors
    imno=self.mf.add(lava_im,filename)
    return (scale,imno)

  def is_compiled(self):
    return True

  def recompile(self):
    if self.mf.size()==0:
      return
    self.mf.undoPCA()
    self.mf.doPCA()
    self.mf.rebuildTree(5)
    crossRelations.reset()
    
  def match(self,imdesc):
    self.mf.setSearchThreshold(0.3,0.15)
    t0=time.time()
    self.mf.retrieve(imdesc)
    t1=time.time()    
    print "retrieve time: %.3f s"%(t1-t0) 

    
  def find_matches_raw(self,imdesc,use_geom=False):
    " return a Matching "
    if not use_geom:
      sc=None
    else:
      sc=LoweHough()
      sc.thisown=False
    self.mf.setSpatialConstraints(sc)
    self.match(imdesc)    
    return self.mf.getMatching()
  
  def save_to(self,fname):
    self.prepare_store(fname)
    self.mf.setPrefix(self.dirprefix)
    self.mf.store(fname)
        
  def load_from(self,fname):
    self.mf=MatchingDataBase()
    self.mf.load(fname)
    crossRelations.reset()
    self.dirprefix=self.make_dirprefix(fname)
    self.istmpdir=False

class DummyDesc:
  def getNDescs(self):
    return 0

class DummyGeomIndex:
  def __init__(self):
    self.cnt=0
    self.mf=Object()
    self.mf.getDescriptor=lambda n: DummyDesc()
  def add_image(self,i,n):
    self.cnt+=1
    return (1.0,self.cnt-1)
  def save_to(self,fname):
    pass
  def load_from(self,fname):
    pass
  

geomIndex=have_photomole and GeometryIndex() or DummyGeomIndex()

class CrossRelations:
  """
  contains the matches between images, represented as a list of weighted votes
  """
  
  def __init__(self):
    self.matchings_key=None
    self.mds_key=None

  def reset(self):
    self.reset_mds()
    if self.matchings_key!=None:
      del self.cached_matchings
    self.matchings_key=None

  def reset_mds(self):
    if self.mds_key!=None:
      self.mds_key=None
      del self.mds_pos
      
  def complete_matchings(self,images):
    key=self.make_key(images)
    if not (self.matchings_key==None or (key & ~self.matchings_key)):
      return # nothing to complete
    
    if self.matchings_key==None:
      self.matchings_key=0L
      self.cached_matchings=[None for i in range(geomIndex.mf.size())]      
    matchings=self.cached_matchings       
    use_geom=True
    pb=ProgressBarWindow("computing cross-relations",len(images))
    for im in images:
      imdesc=im.get_imdesc()
      if matchings[im.geom_imno]!=None: continue
      mat=geomIndex.find_matches_raw(imdesc,use_geom)
      
      # don't store the whole matching
      matchings[im.geom_imno]=[mat.getWeightedVotes(i.geom_imno) for i in images]
      # copy matching, it belongs to the MatchingFeatures
      # matchings[im.geom_imno]=Matching(mat)
      pb.one_step()
    pb.destroy()

    self.matchings_key|=key    
    self.reset_mds()

  def matchings_to_cross_rels(self,images):
    matchings=self.cached_matchings           
    cross_rels=[]
    #print "images %d matchings %d "%(len(images),len(matchings))
    #print "cached images",self.matchings_key
    for im in images:
      #mat=matchings[im.geom_imno]
      #l=[mat.getWeightedVotes(i.geom_imno) for i in images]
      l=self.cached_matchings[im.geom_imno]
      cross_rels.append(l[:])
    return cross_rels

  def make_key(self,images):
    key=0L
    for i in images:
      key|=1L<<i.geom_imno
    return key

  def get_cross_rels(self,images):
    self.complete_matchings(images)
    return self.matchings_to_cross_rels(images)

  def get_normalized_cross_rels(self,images):
    """ make symmetric & normalize by the self-matches """
    cross_rels=self.get_cross_rels(images)
    n=len(images)
    norms=[1.0/(sqrt(cross_rels[i][i])+1e-10) for i in range(n)]
    cr2=[]
    for i,l in enumerate(cross_rels):
      v=norms[i]
      l=[]
      for j in range(n):
        v2=norms[j]
        l.append((cross_rels[i][j]+cross_rels[j][i])*v*v2*0.5)
      cr2.append(l)
    return cr2

  def get_2image_matches(self,im1,im2,use_geom=True):
    imdesc1=geomIndex.mf.getDescriptor(im1.geom_imno)
    mat=geomIndex.find_matches_raw(imdesc1,use_geom)
    mc=mat.getMatches(im2.geom_imno)
    mc=cast_MatchContainer(mc)
    # pdb.set_trace()
    # mat.describe()
    matches=[]
    for i in range(mc.size()):
      match=mc[i]
      matches.append((match.qpt,match.dbpt,match.dist))
    return matches
 

  #########################################################################
  # MDS positions

  def get_mds_pos(self,images,fact):
    key=(self.make_key(images),fact)
    if self.mds_key!=key:
      cross_rels=self.get_normalized_cross_rels(images)     
      self.mds_pos=cross_rels_to_mds_pos(cross_rels,lambda x:1.0-x**(fact+0.05))
      self.mds_key=key
    return self.mds_pos

  def get_reasonable_order(self,images,fact):
    cross_rels=self.get_normalized_cross_rels(images)
    fact=pow(fact,10)
    perm=get_reasonable_order(cross_rels,lambda x:x>fact and x or 0)
    return [images[i] for i in perm]

  def get_lapeig_pos(self,images,fact):
    cross_rels=self.get_normalized_cross_rels(images)
    return cross_rels_to_lapeig_pos(cross_rels)

  #########################################################################
  # Load/save only the cached matchings (they require expensive computations)

  def save_to(self,fname):
    f=open(fname,"w")
    if self.matchings_key!=None:
      print "key=%x"%self.matchings_key
    cPickle.dump(self.matchings_key,f)    
    if self.matchings_key!=None:
      n=len(self.cached_matchings)
      cPickle.dump(n,f)
      for i in range(n):
        mat=self.cached_matchings[i]
        if self.matchings_key & (1L<<i):        
          assert mat!=None
          cPickle.dump(mat,f)
        else:
          assert mat==None
    f.close()
        
  def load_from(self,fname):
    self.reset()
    try:
      f=open(fname,"r")
    except IOError,e:      
      if e.errno==2: return
      raise e
    self.matchings_key=cPickle.load(f)
    if self.matchings_key!=None:
      print "key=%x"%self.matchings_key
      n=cPickle.load(f)
      self.cached_matchings=[None for i in range(n)]        
      for i in range(n):
        if self.matchings_key & (1L<<i):        
          mat=cPickle.load(f)
          self.cached_matchings[i]=mat
    # print "read ",f.tell()
    f.close()

class DummyCrossRelations:
  def get_2image_matches(self,im1,im2,x):
    return []
  def save_to(self,fname):
    pass
  def load_from(self,fname):
    pass
    

crossRelations=have_photomole and CrossRelations() or DummyCrossRelations()






def desc_to_parallelogram(desc,f=0.5):
  (m11,m12,m21,m22)=desc.imageMapping()
  corns=[]
  (x0,y0)=(desc.getx(),desc.gety())
  for (x,y) in [(-f,-f),(-f,f),(f,f),(f,-f)]:
    x2=x*m11+y*m12
    y2=x*m21+y*m22
    corns.append((x2+x0,y2+y0))
  return corns


if __name__=='__main__':

  os.system("mkdir /tmp/zz.yorg || rm -rf /tmp/zz.yorg/*")

  # pdb.set_trace()
  
  for fno in [3493,3283,2532,3113]:
    fname="test_img/sz_subset/PIC_%04d.JPG"%fno
    im=gtk.gdk.pixbuf_new_from_file(fname)
    # geomIndex.mf.verbose=10
    geomIndex.add_image(im,fname)
    # geomIndex.recompile()
    # describeFeaturesComputer(geomIndex.fc)
  imdesc=geomIndex.mf.getDescriptor(0)
  print "# points: ",imdesc.getNDescs()

  mat=geomIndex.find_matches_raw(imdesc,False)
  # mat.describe_tab()
  # print "votes: ",mat.getWeightedVotes(0),mat.getWeightedVotes(1)

  im1=Object(); im1.geom_imno=0
  im2=Object(); im2.geom_imno=1

  n=len(crossRelations.get_2image_matches(im1,im2))
  print "# 2 img matches",n
    
  dbfn="/tmp/zz.yorg/tt.db"
  geomIndex.save_to(dbfn)

    # pdb.set_trace()
  
  geomIndex.load_from(dbfn)
  
  imdesc=geomIndex.mf.getDescriptor(0)
  print "# points: ",imdesc.getNDescs()

  # describeFeaturesComputer(gi2.fc)
  
  #print gi2.find_matches(gi2.mf.getDescriptors(0))
  # geomIndex.save_to("/tmp/zz.yorg/tt2.db")
  
  im1=Object(); im1.geom_imno=0
  im2=Object(); im2.geom_imno=1
  
  n=len(crossRelations.get_2image_matches(im1,im2))
  print "# 2 img matches",n
