

from math import *

import gtk,gobject
import cPickle,os,gc

import libexif,color,person_detector
from geometry import *

from extra_functions import *

from texture import *


class Imagecache:

  err_pat=('0000000'+
           '0011100'+
           '0100010'+
           '0111100'+
           '0100000'+
           '0011110'+
           '0000000')
  err_data=''.join(map(lambda x:x=='1' and '\xff\x00\x00' or '\x00\x00\x00',err_pat))
  
  error_im=gtk.gdk.pixbuf_new_from_data(
    err_data,gtk.gdk.COLORSPACE_RGB,False,8,7,7,21)
  
  def __init__(self,maxnim):
    self.maxnim=maxnim
    self.nim=0
    self.sims={}
    self.date=0

  def find_oldest(self):
    oldest_date=self.date    
    for fname,(date,im) in self.sims.iteritems():
      if date<oldest_date:
        oldest_date=date
        oldest=fname
    return oldest
  
  def get_im(self,fname):
    self.date+=1
    if self.sims.has_key(fname):
      (date,im)=self.sims[fname]
      self.sims[fname]=(self.date,im)
      return im
    assert self.nim==len(self.sims.items())    
    if self.nim>=self.maxnim:
      ofname=self.find_oldest()
      del self.sims[ofname]
      gc.collect() # else memory leak
      self.nim-=1
    try:
      im=gtk.gdk.pixbuf_new_from_file(fname)
    except gobject.GError,e:
      print "error loading ",fname
      return None
    self.sims[fname]=(self.date,im)
    self.nim+=1
    return im

imagecache=Imagecache(32)



class Image:
  """
  fname = the filename
  size = (xd,yd), may be None if file was never seen
  ref_thumb = reference thumbnail of <= ref_thumb_npix pixels.
     Set if size is set
  exif_info = ExifInfo structure, may be None
  thumb = current thumbnail (pixmap), may be None
  tags=list of text tags
  geom_imno=if object added to geometry manager, its no
  selected= boolean
  hist= histogram object
  person=output of the person detector (list of tuples)
  tex_imno=same as geom_imno but for texture descriptors
  mask=VectorMask or None
  """

  ref_thumb_npix=240*180
  
  def __init__(self,fname=None):
    self.fname=fname
    if fname!=None:
      self.fstat=os.stat(self.fname)      
    self.size=None
    self.thumb=None
    self.exif_info=None
    self.tags=[]
    self.geom_imno=None
    self.selected=False
    self.hist=None
    self.persons=None
    self.tex_imno=None
    self.mask=None
    self.change=ChangeBroadcaster()
    
  def __repr__(self):
    return self.fname[self.fname.rfind('/')+1:]
 
  __str__=__repr__


  def analyze_image_partial(self):
    """ loads the full-size image, makes the reference thumbnail """

    full_im=imagecache.get_im(self.fname)
    self.valid_im=full_im!=None
    
    if not full_im:
      full_im=Imagecache.error_im      
      # TODO add real handler
    
    xd=full_im.get_width(); yd=full_im.get_height()
    self.size=(xd,yd)
    
    # ref thumbnail
    self.ref_thumb=shrink_big_image(full_im,Image.ref_thumb_npix)

    self.change.broadcast("analyze_partial")

    self.hist=None

  def analyze_image(self):
    """ adds the color histogram and the geometry """
    full_im=imagecache.get_im(self.fname)

    # histogram
    self.hist=color.FixedColorHistogram(10)
    self.hist.computeFromPixbuf(full_im)
    
    # geometry
    (scale,imno)=geomIndex.add_image(full_im,self.fname)
    self.geom_imno=imno   
    self.geom_scale=scale

    self.change.broadcast("analyze_image")


  def analyze_tex(self):
    if not self.size:
      self.analyze_image_partial()
    
    full_im=imagecache.get_im(self.fname)
    
    (scale,pixbuf)=scale_pixbuf_maxpix(full_im,640*480)

    self.tex_imno=textureManager.compute_tex(pixbuf)
    self.tex_scale=scale

    self.change.broadcast("analyze_tex")

  def get_thumb_size(self,w,h):
    """ get scale and size of an image with the same aspect ratio
    that fits in w*h"""
    if not self.size:
      self.analyze_image_partial()      
    (xd,yd)=self.size
    xscale=w/float(xd)
    yscale=h/float(yd)
    if xscale<yscale:
      wr=w
      hr=int(yd*xscale)
      scale=xscale
    else:
      hr=h
      wr=int(xd*yscale)
      scale=yscale
    return (scale,wr,hr)

  def get_thumb(self,w,h,win,gc,quality=False):
    (scale,wr,hr)=self.get_thumb_size(w,h)      
    # print "sizes= ",(w,h),self.size,(wr,hr)
    if not quality:
      if not self.thumb or self.thumb.get_size()!=(wr,hr):
        pixb=self.ref_thumb.scale_simple(wr,hr,gtk.gdk.INTERP_NEAREST)   
        self.thumb=gtk.gdk.Pixmap(win,wr,hr)
        self.thumb.draw_pixbuf(gc,pixb,0,0,0,0)
      return (scale,self.thumb)
    else:
      full_im=imagecache.get_im(self.fname)
      if not full_im: return
      pixb=full_im.scale_simple(wr,hr,gtk.gdk.INTERP_BILINEAR)
      thumb=gtk.gdk.Pixmap(win,wr,hr)
      thumb.draw_pixbuf(gc,pixb,0,0,0,0)
      return (scale,thumb)

  # what has been computed
  def what_computed(self):
    """ returns a string with letters:
    s = size + reference thumbnail is ready (the file was opened)
    e = exif info ready
    h = color histogram
    g = geometric descriptors computed
    p = persons
    t =textures """
    accu=""
    if self.size:            accu+='s'
    if self.exif_info:       accu+='e'
    if self.hist:            accu+='h'
    if self.geom_imno!=None: accu+='g'
    if self.persons!=None:   accu+='p'
    if self.tex_imno!=None:  accu+='t'
    return accu
  
  def all_computed(self,what):
    """ checks if all components in what have been computed """
    if 's' in what and not self.size:            return False
    if 'e' in what and not self.exif_info:       return False
    if 'h' in what and not self.hist:            return False
    if 'g' in what and not self.geom_imno!=None: return False
    if 'p' in what and not self.persons!=None:   return False
    if 't' in what and not self.tex_imno!=None:  return False
    return True
  

  ## image functions

  def is_valid(self):
    if not self.size:
      self.analyze_image_partial()
    return self.valid_im

  def get_npix(self):
    if not self.size:
      self.analyze_image_partial()
    return self.size[0]*self.size[1]

  def get_size(self):
    if not self.size:
      self.analyze_image_partial()
    return self.size

  def get_ref_thumb(self):
    if not self.size:
      self.analyze_image_partial()
    return self.ref_thumb

  def get_full_image(self):
    if not self.size:
      self.analyze_image_partial()      
    return imagecache.get_im(self.fname)

  # exif info
       
  def get_exif_info(self):
    if not self.exif_info:
      ei=libexif.ExifInfo(self.fname)
      self.exif_info=eip=Object()
      eip.load_ok=ei.load_ok and ei.a>0
      # copy to python object (easier to pickle)        
      if eip.load_ok:
        eip.a=ei.a
        eip.t=ei.t
        eip.f=ei.f
        eip.iso=ei.iso
        eip.orientation=ei.orientation
        eip.date=ei.date
        eip.model=ei.model
      self.change.broadcast("analyze_exif")
    if self.exif_info.load_ok:
      return self.exif_info

  # color histogram

  def get_hist(self):
    if not self.hist:
      self.analyze_image()      
    return self.hist

  def dominant_color(self):
    try:
      return self.dom_color
    except AttributeError:
      pass    
    if not self.hist:
      self.analyze_image()
    self.dom_color=self.hist.typicalColor()
    return self.dom_color

  # geometry

  def get_imdesc(self):
    if self.geom_imno==None:
      self.analyze_image()    
    return geomIndex.mf.getDescriptor(self.geom_imno)

  # person detector

  def get_persons(self):
    if self.persons==None:
      im=imagecache.get_im(self.fname)
      (scale,im)=scale_pixbuf_maxpix(im,640*480)
      self.persons_scale=scale
      print "detecting persons..."
      self.persons=person_detector.persondet_w(im)
      self.change.broadcast("analyze_persons")
    return self.persons

  # texture

  def get_tex_npatch(self):
    if self.tex_imno==None:
      self.analyze_tex()
    return textureManager.get_npatch(self.tex_imno)
   
  def get_tex_imno(self):
    if self.tex_imno==None:
      self.analyze_tex()
    return self.tex_imno

  def get_tex_scale(self):
    if self.tex_imno==None:
      self.analyze_tex()
    return self.tex_scale    

  def long_str(self):
    txt=""
    txt+="%d bytes\n"%self.fstat.st_size
    txt+="\n"
    if not self.size:
      txt+="not examined\n"
    elif not self.is_valid():
      txt+="invalid image\n"
    else:
      txt+="%d*%d\n"%(self.size[0],self.size[1])
      if self.hist:
        txt+="dominant RGB=%d %d %d\n"%self.dominant_color()
        txt+="%d keypoints\n"%self.get_imdesc().getNDescs()
      else:
        txt+="color & geometry not computed\n"
    txt+="\n"
    if not self.exif_info:
      txt+="EXIF info not loaded\n"
    elif not self.exif_info.load_ok:
      txt+="no EXIF info\n"
    else:
      txt+="%s\n"%self.exif_info.model
      txt+="on %s\n"%self.exif_info.date
      txt+="F/%g, %g s\n"%(self.exif_info.a,self.exif_info.t)
      txt+="focal length %g mm\n"%self.exif_info.f
      txt+="pixel (0,0) is %s\n"%libexif.ExifInfo.getOrientationName(
        self.exif_info.orientation)

    if self.persons==None:
      txt+="persons not detected\n"
    else:
      txt+="%d persons found\n"%len(self.persons)
    if self.tex_imno==None:
      txt+="textures not detected\n"
    else:
      txt+="%d texture patches\n"%self.get_tex_npatch()
      
    if self.mask==None:
      txt+="no mask defined\n"
    else:
      txt+="mask: %s"%self.mask
     
    return txt
  

  ##################################################################
  # file I/O for db serialization
  
  def save_to(self,f):
    cPickle.dump(obj_to_dict(
      self,
      ('fname','size','exif_info','tags','persons','tex_imno','tex_scale',
       'valid_im','geom_imno','geom_scale','persons_scale')),
                 f)
    
    if self.size:
      savePixbuf(self.ref_thumb,f)    
    if self.hist!=None:
      f.write("H")
      self.hist.saveTo(f)
    else:
      f.write("0")
      
  def load_from(self,f,version):
    dict_to_obj(cPickle.load(f),self)
    self.mask=ReadOnlyVectorMask # old versions saved mask
    try:
      self.fstat=os.stat(self.fname)
      exists=True
    except OSError,e:
      exists=False
      print "failed to read ",self.fname
    if self.size:
      self.ref_thumb=loadPixbuf(f)
    histv=f.read(1)
    if histv=='H':
      self.hist=color.FixedColorHistogram(10)        
      self.hist.loadFrom(f)
    elif histv=='0':
      pass
    else:
      raise SyntaxError("histv has wrong value")
    return exists


class TagManager(ChangeBroadcaster):

  def __init__(self):
    ChangeBroadcaster.__init__(self)
    self.all_tags=[]

  def add_tag_to(self,ims,tag):
    assert tag in self.all_tags
    for im in ims:
      if tag not in im.tags:
        im.tags.append(tag)
    self.broadcast("add")

  def has_tag(self,tag):
    return tag in self.all_tags

  def new_tag(self,tag):
    assert tag not in self.all_tags
    self.all_tags.append(tag)
    self.broadcast("new")

  def remove_tag_from(self,ims,tag):
    assert tag in self.all_tags
    for im in ims:
      try:
        im.tags.remove(tag)
      except ValueError:
        pass
    self.broadcast("rm")
   
  def load_from(self,f):
    self.all_tags=cPickle.load(f)
    self.broadcast("replace")

  def save_to(self,f):
    cPickle.dump(self.all_tags,f)



    
tagManager=TagManager()

