
import os,sys,time


import gtk


from image import *

from  extra_widgets import *
from extra_functions import *

    
class Database:
  def __init__(self):
    self.prev_dirname=None
    self.clear()
    self.change_mask=ChangeBroadcaster()
    
  def clear(self):
    self.images=[]
    self.imageset=None # set if only a subset is to be shown
    self.all_filenames={}
    
  accepted_extensions=['.jpg','.jpeg','.pgm','.ppm','.png','.tif','.tiff']

  def add_from_dir(self,dirname,recurse=True):
    for name in os.listdir(dirname):
      filename=dirname+'/'+name
      if name[name.rfind('.'):].lower() in Database.accepted_extensions:
        self.add_file(filename)
      elif recurse and os.path.isdir(filename):
        self.add_from_dir(filename,True)
      else:
        print "ignored ",filename
     
  def add_file(self,filename):
    if self.all_filenames.has_key(filename):
      print "duplicate",filename
    else:
      im=Image(filename)
      self.images.append(im)
      self.all_filenames[filename]=None
  
  def analyze_all(self,what='segpt',images_to_do=None):
    """ see Image.what_computed for explanation about segpt"""
    if images_to_do==None:
      images_to_do=self.images
    images_to_do=[im for im in images_to_do if not im.all_computed(what)]
    if images_to_do==[]: return
    pbwidget=gtk.Label("")
    pb=ProgressBarWindow("Analyzing images...",len(images_to_do),
                         None,pbwidget)
    t0=time.time()
    n=0
    for im in images_to_do:
      pbwidget.set_label(im.fname)
      if 's' in what:
        im.get_ref_thumb()
      if 'e' in what:
        im.get_exif_info()
      if 'g' in what:
        im.get_imdesc()
      if 'p' in what:
        im.get_persons()
      if 't' in what:
        im.get_tex_npatch()
      n+=1
      if not pb.one_step(): break
    t1=time.time()
    print "analysis time: %.3f s for %d images"%(t1-t0,n)
    pb.destroy()
    

  def clear_exif(self):
    for im in self.images:
      im.exif_info=None

  def clamp_vmasks(self):
    for im in self.images:
      if im.mask!=None:
        (xd,yd)=im.get_size()
        im.mask.clamp_curves(xd,yd)
        
    
  ######################################################################
  # Handle subset of images

  def images_to_display(self):
    # maybe we'll organize only a subset of images some time
    if self.imageset==None:
      return self.images
    else:
      return self.imageset

  def set_imageset(self,images):
    self.imageset=images[:]

  def reset_imageset(self):
    self.imageset=None

  ######################################################################
  # File I/O

  file_magic='Image Organizer 3.0'
  
  def save_as(self,dirname):
    self.prev_dirname=dirname
    f=open(dirname+'/image.dat','w')
    cPickle.dump(Database.file_magic,f)
    tagManager.save_to(f)
    cPickle.dump(len(self.images),f)
    for i in self.images:
      i.save_to(f)
    f.close()
    geomIndex.save_to(dirname+'/geom_index.phmdb')
    crossRelations.save_to(dirname+'/cross_relations.raw')
    textureManager.save_to(dirname+'/texture.raw')
    allTexlas.save_to(dirname+'/texlas')
    
  def getversion(self,magic):
    t=magic.split()
    assert (t[0],t[1])==('Image','Organizer')
    assert t[2] in ('3.0',)
    return t[2]    
    
  def load_from(self,dirname):
    self.clear()
    self.prev_dirname=dirname
    f=open(dirname+'/image.dat','r')
    magic=cPickle.load(f)
    v=self.getversion(magic)
    print "reading version ",v
    tagManager.load_from(f)
    nim=cPickle.load(f)
    for i in range(nim):
      im=Image()
      ok=im.load_from(f,v)
      if not ok: continue
      self.images.append(im)
      self.all_filenames[im.fname]=None
    f.close()
    geomIndex.load_from(dirname+'/geom_index.phmdb')
    crossRelations.load_from(dirname+'/cross_relations.raw')
    textureManager.load_from(dirname+'/texture.raw')
    allTexlas.load_from(dirname+'/texlas')

  def previous_name(self):
    return self.prev_dirname

  def save_tags(self,filename):
    d={}
    for im in self.images:
      d[im.fname]=im.tags
    ll=(tagManager.all_tags,d)
    cPickle.dump(ll,open(filename,"w"))

  def load_tags(self,filename):
    (add_tags,d)=cPickle.load(open(filename,"r"))
    for tag in add_tags:
      if tagManager.has_tag(tag): continue
      tagManager.new_tag(tag)
    for im in self.images:
      if not d.has_key(im.fname): continue
      for tag in d[im.fname]:
        if tag not in im.tags:
          im.tags.append(tag)
    tagManager.broadcast("replace")

  def save_masks(self,filename):
    maskdict={}
    for im in self.images:
      maskdict[im.fname]=im.mask
    cPickle.dump(maskdict,open(filename,"w"))
  
  # def on_mask_change(self,im):
  #  self.change_mask.broadcast(None,im)

def usage():
  sys.stderr.write("usage: %s [-add directory] [-compile (s|e|g|p|t)] [dir.db] [-selwin]\n"%sys.argv[0])
  sys.exit(1)

open_db=None
preload_dirs=[]
pre_compile=''
n_open_selected=0


args=sys.argv[1:]
while args:
  arg=args[0]; del args[0]
  if arg=='-add':
    preload_dirs.append(args[0])
    del args[0]
  elif arg=='-compile':
    pre_compile=args[0]
    del args[0]    
  elif arg=='-selwin':
    n_open_selected+=1
  else:
    open_db=arg



database=Database()

if open_db:
  database.load_from(open_db)

for d in preload_dirs:
  database.add_from_dir(d)

if pre_compile:
  database.analyze_all(pre_compile)



class ImageSelection:

  def __init__(self):
    self.selected_ims=[]
    self.sel_change_cb=[]

  def add_callback(self,cb):    
    self.sel_change_cb.append(cb)

  def remove_callback(self,cb):    
    self.sel_change_cb.remove(cb)
    
  def set(self,sel):
    for im in self.selected_ims:
      im.selected=False
    print "selection set to ",sel
    for im in sel:
      im.selected=True
    # make list with only only 1 instance of each image
    self.selected_ims=[]
    for im in database.images:
      if im.selected:
        self.selected_ims.append(im)
    for f in self.sel_change_cb:      
      f(sel)
      
  def get(self):
    return self.selected_ims

  def invert(self):
    sel=[]
    for im in database.images_to_display():
      if not im.selected:
        sel.append(im)
    self.set(sel)


imageSelection=ImageSelection()


