
#ifndef COLOR_HISTOGRAM_H_INCLUDED
#define COLOR_HISTOGRAM_H_INCLUDED

#include <stdio.h>
#include <vector>


extern "C" {
  // used in color_emd
typedef struct signature_t signature_t;
}

struct ColorHistogram {
  int totpix;

  virtual ~ColorHistogram() {};

  virtual void typicalColor(double &r_out,double &g_out,double &b_out)=0;

  virtual void computeFrom(int xd,int yd,const unsigned char *data,int bpl)=0;
  virtual void saveTo(FILE *f)=0;
  virtual void loadFrom(FILE *f)=0;

  // used by EMD
  virtual void toSignature(signature_t &sig) const=0;
  
};

struct FixedColorHistogram:ColorHistogram {
  int nbin,totsz;
  int *bins;
  
  FixedColorHistogram(int nbin=10);
  ~FixedColorHistogram();

  // assumes  r g b r g b... layout
  void computeFrom(int xd,int yd,const unsigned char *data,int bpl);
  
  void typicalColor(double &r_out,double &g_out,double &b_out);

  // the *_curve should be nbin long
  void getCurves(int *r_curve,int *g_curve,int *b_curve) const;
  
  
/*
  void getWeightedAvg(double *weights,
                      double &r_out,double &g_out,double &b_out);
*/
  
  void saveTo(FILE *f);
  void loadFrom(FILE *f);

  void toSignature(signature_t &sig) const;

};





struct ColorBin {
  unsigned char r,g,b;
  int w;
};

struct ClusteredColorHistogram:ColorHistogram {

  // parameters to make the histogram.
  
  bool use_lab;
  int ncluster;

  bool use_hash;
  int param; 
  // if use_hash, param=nb of bits to ignore initially 
  // else param=nb of bins in the regular 3D color histogram
  
  bool use_mean_shift;
  // if not, use KD-tree


  // result
  std::vector<ColorBin> clusters;

  
  ClusteredColorHistogram(int ncluster);

  ~ClusteredColorHistogram();

  void typicalColor(double &r_out,double &g_out,double &b_out);

  void computeFrom(int xd,int yd,const unsigned char *data,int bpl);
  
  void saveTo(FILE *f);
  void loadFrom(FILE *f);
  void toSignature(signature_t &sig) const;

};


void rgb_to_labs(double r,double g,double b,
                double &ls,double &as,double &bs); // 's' like 'star'

void labs_to_rgb(double ls,double as,double bs,
                 double &r_out,double &g_out,double &b_out);





// some functions & constants that can be used with ColorHistogram
// ColorHistogramCommon doesn't need to be instanciated more than once 
struct ColorHistogramCommon {
  int nbin,totsz;

  ColorHistogramCommon(int nbin=10);
  ~ColorHistogramCommon();


  double distance(const FixedColorHistogram &h1,
                  const FixedColorHistogram &h2);

  double weightedDistance(const FixedColorHistogram &h1,
                          const FixedColorHistogram &h2,
                          const double *weights);

  double earthMoversDistance(const ColorHistogram &h1,
                             const ColorHistogram &h2);
  

                                      
  double *saturated_weights;
};  

#endif


