# -*- makefile -*-
###########################################################
# Platform specific configuration for YORG
#
# YORG was developed on GNU/linux i386 and x86_64 platforms and it has
# only been tested on these. There are quite a few dependencies. At a
# minimum you will need to install your distribution's standard
# packages for the development versions of python, gtk, pygtk,
# libexif, boost, blitz, and ImageMagick C and C++ libraries. You will
# need working C++ and Fortran compilers (we use gcc and g77 with
# libg2c). 
# You will also need the ARPACK library libarpack.a and the python
# bindings for libsvm. These are not usually available as standard
# packages but they are easy to compile from source. For convenience
# we have included precompiled versions of them for the i386 and
# x86_64 platforms in native/lib.$platform.
#
# Flags: for package XXXX you usually need
# XXXXCFLAGS: C/C++ flags needed to compile with the package 
#             (may be empty if the package is in a standard include path)
# XXXLDFLAGS: link flags and libraries to link in
#
###################################################################
# Common flags

PLATFORM=$(shell uname -i)

ifeq ($(PLATFORM),x86_64)
   PICFLAGS =-fPIC
endif

CFLAGS=-g -Wall -Wno-sign-compare $(PICFLAGS)
LDFLAGS=-g $(PICFLAGS)

###################################################################
# Python and GTK flags

PYTHONCFLAGS=-I /usr/include/python2.5/

PYGTKCFLAGS= \
   -I /usr/include/pygtk-2.0 \
   -I /usr/include/glib-2.0 \
   -I /usr/include/gtk-2.0 \
   -I /usr/lib/glib-2.0/include

PYGTKLDFLAGS=-lgdk_pixbuf-2.0

###################################################################
# The python bindings to libsvm are used to learn classifiers. Before
# running YORG you will need to compile them separately (just run make
# in the python subdirectory of the libsvm distribution) and copy
# svm.py and svmc.so to native/lib.$(PLATFORM) or somewhere else in
# your PYTHONPATH. There are no compilation flags here for libsvm.

###################################################################
# libexif - used to read image EXIF tags

EXIFCFLAGS=
EXIFLDFLAGS= -lexif
# EXIFDIR=<path to libexif installation>
# EXIFCFLAGS=-I $(EXIFDIR)/include
# EXIFLDFLAGS= -L $(EXIFDIR)/lib -lexif

###################################################################
# OLT (Navneet Dalal's Object Localization Toolkit), used as the
# person detector. Requires Blitz and Boost.

BOOSTCFLAGS= 
BOOSTLDFLAGS= -lboost_filesystem

BLITZCFLAGS=
BLITZLDFLAGS= -lblitz
#BLITZDIR=<path to blitz>
#BLITZCFLAGS=-I $(BLITZDIR)/include
#BLITZLDFLAGS=-L $(BLITZDIR)/lib -lblitz

OLTDIR=./olt
OLTCFLAGS=-I $(OLTDIR)/include $(BLITZCFLAGS)
OLTLDFLAGS= \
    -L $(OLTDIR)/lib -L $(OLTDIR)/lib/classifier -lutil -lclassifier -lutil \
    $(BLITZLDFLAGS) $(BOOSTLDFLAGS)

###################################################################
# Extreme Multidimensional Scaling package, used for scatter plots of
# images based on image similarity. This is in fortran and it needs
# ARPACK, which you will need to compile separately. There is an
# example ARPACK makefile for linux in ARmake.inc.linux

# fortran compiler
FC=g77
FFLAGS=-g -O3 $(PICFLAGS)
FLDFLAGS=-lg2c $(PICFLAGS)
# macro to convert fortran to C function names
FCMACRO=-D "CNAME(fname)=fname\#\#__"

# We provide precompiled ARPACK libraries for a few platforms in
# native/lib.$(PLATFORM)

ARPACKDIR=../lib.$(PLATFORM)
ARPACKLDFLAGS=-L$(ARPACKDIR) -larpack

###################################################################
# ImageMagick - needed for FRIMCLA texture classifier

MAGICKCFLAGS=`Magick++-config --cppflags`
MAGICKLDFLAGS=`Magick++-config --ldflags --libs`

###################################################################
# Java. This is only needed if you are a partner of the European Union
# research project aceMedia and need to make the aceMedia Java
# wrappers.

ACEMEDIA_CODE_ROOT=<path to aceMedia installation>/buildtree/trunk/src
JAVADIR=/usr/java/jdk1.5.0_02
JAVACFLAGS=-I $(JAVADIR)/include -I $(JAVADIR)/include/linux
# no JAVALDFLAGS as JNI libraries do not depend on Java

###################################################################
# LAVA - this is an image feature library only available within the
# LEAR research team. YORG uses it to detect interest points for the
# texture classifier but it can be built without it, in which case it
# uses random rather than interest point features for the texture
# classifier. LAVA can be compiled without image i/o support.

USELAVA=no

ifeq "$(USELAVA)" "yes"
  LAVA2DIR=<path to lava2 installation>
  LAVA2LIB=$(LAVA2DIR)/lib/$(PLATFORM)
  LAVA2CFLAGS=-I$(LAVA2DIR)/include -DHAVE_LAVA
  LAVA2LDFLAGS=-L$(LAVA2LIB) -llava2
else
  LAVA2CFLAGS=-DNOLAVA
  LAVA2LDFLAGS=
endif

###########################################################
# Example of automatic selection of your linux distribution

# ISSUE=$(shell head -1 /etc/issue)
# ifneq ($(findstring Fedora release 8,$(ISSUE)),)
# os=fc8
# endif

ifdef os
  # name of this makefile.inc (with dirname)
  myname=$(word $(words $(MAKEFILE_LIST)),$(MAKEFILE_LIST))
  include $(myname)_$(os)
endif
###########################################################