#include "mds.h"

#include <stdio.h>
#include <stdlib.h>



extern void CNAME(pca_gma1)(int *natoms,int *ndim,int *mgutt,
                       double *tau,double *x,double *stress,int *info);


int multi_dimensional_scaling(int natom,int ndim,int niter,
                              double *tau,double *x,
                              double *stress_out) {
  int info;
  double dummy;
  if(!stress_out) stress_out=&dummy;
  CNAME(pca_gma1)(&natom,&ndim,&niter,tau,x,stress_out,&info);
  return info;
}
                             
static __thread int *global_n_non0;

int dec_n_non0(const void *va, const void *vb) {
  int a=*(int*)va;
  int b=*(int*)vb;

  return global_n_non0[b]-global_n_non0[a];
}

void one_dim_matrix_permutation(int n,double *w,int *perm) {
  
  { // make symmetric
    int i,j;
    for(i=0;i<n;i++) 
      for(j=0;j<i;j++) 
        w[i+j*n]=w[j+i*n]=0.5*(w[i+j*n]+w[j+i*n]);
  }

  { // initial permutation given by nb of non-0 elts
    int i;
    int n_non0[n];
    for(i=0;i<n;i++) {
      int k=0,j;
      for(j=0;j<n;j++) if(w[i+j*n]!=0) k++;
      n_non0[i]=k;
    }
    for(i=0;i<n;i++) perm[i]=i;
    global_n_non0=n_non0;
    qsort(perm,n,sizeof(int),&dec_n_non0);
  }

  {
    int d=0,i;
#define W(i,j) w[perm[i]+n*perm[j]]
    for(i=0;i<n;i++) {
      int r=n-1;
      for(;;) {        
        while(d<r && W(i,d)!=0) d++;
        if(d==r) break;
        while(d<r && W(i,r)==0) r--;
        if(d==r) break;
        { // swap d & r
          int tmp=perm[d]; 
          perm[d]=perm[r]; perm[r]=tmp;
        }
      }
    }
  }

  {
    int i,j;
    for(i=0;i<n;i++) {
      for(j=0;j<n;j++) {
        if(W(i,j)) printf("#");
        else printf(" ");
      } 
      printf("\n");
    }
  }
  
#undef W
}




void laplacian_eigenmap(int natom,int ndim,int niter,
                        double *tau,double *x) {
  int i,j;    
  {
    FILE *f=fopen("mds/cross_rels.dat","w");
    fprintf(f,"%d %d \n",natom,ndim);
    
    for(i=0;i<natom;i++) {
      for(j=0;j<natom;j++) {
        fprintf(f,"%g ",tau[i+j*natom]);
      }
      fprintf(f,"\n");
    }
    fclose(f);
  }
  printf("run octave....\n");
  getchar();

  {
    FILE *f=fopen("mds/positions.dat","r");    
    for(i=0;i<natom;i++) {
      for(j=0;j<ndim;j++) {
        fscanf(f," %lg",&x[i*ndim+j]);
        printf("%g ",x[i*ndim+j]);
      }
      printf("\n");
    }
    fclose(f);
  }

}
