


#ifndef MDS_H_INCLUDED
#define MDS_H_INCLUDED

/**
 * reduce object-o-object distances to a set of low-dimensional coordinates 
 * \param natoms nb of objects 
 * \param ndim target nb of dimensions
 * \param niter maximum nb of iterations
 * \param tau input, dimension [natoms*natoms], 
 *        tau[i+j*natoms] is the distance between objects i and j 
 * \param x output, dimension [natoms*ndim], (x[i+j*natom])_{j=0..ndim-1} is
 *        the coordinate vector of object i
 * \param stress_out if not NULL, the final stress criterion 
 *        is written to *stress_out
 * \return info returned by pca_gma1, 1=ok, 
 *        else = error from ARPACK's dsaupd or dsaepd
 */
 
int multi_dimensional_scaling(int natoms,int ndim,int niter,
                              double *tau,double *x,
                              double *stress_out);


void one_dim_matrix_permutation(int n,double *w,int *perm);



void laplacian_eigenmap(int natom,int ndim,int niter,
                        double *tau,double *x);

#endif


/*

from LavaAndGtk import *
import random



n=4
m=DoubleArray(n*n)
for i in range(n):
  m[i+n*i]=0
  for j in range(i):
    m[i+j*n]=m[j+i*n]=random.random()

x=DoubleArray(2*n)

multi_dimensional_scaling(n,2,10,m,x,None)



*/
