/***************************************************************************
                                ClassLabelHistogram.cpp
                             ------------------
    begin                : 16.11.2005
    copyright            : (C) 2005 by Frank Moosmann
    email                : frank.moosmann@inrialpes.fr
 ***************************************************************************/

#include <iostream>
#include "ClassLabelHistogram.h"
#include "FeatureVector.h"

using namespace std;

 
//**************************************************************************
//                         Constructors / Destructors
//**************************************************************************
ClassLabelHistogram::ClassLabelHistogram()
{
  memberCount = 0;
}


ClassLabelHistogram::ClassLabelHistogram(std::list<FeatureVector*>* sampleList)
{
  memberCount = 0;
  for(list<FeatureVector*>::iterator p=sampleList->begin(); (p!=sampleList->end()); ++p)  {
    increaseCountOfLabel((*p)->classLabel);
  }
}


ClassLabelHistogram::~ClassLabelHistogram()
{
}


//**************************************************************************
//                         public Methods
//**************************************************************************
void ClassLabelHistogram::reset ()
{
  countHistogram.clear();
  weightsumHistogram.clear();
  memberCount = 0;
}

unsigned int ClassLabelHistogram::getTotalCount()
{
  return memberCount;
}

void ClassLabelHistogram::getListOfClassLabels (list<ClassLabel>* classLabelList)
{
  classLabelList->clear();
  for (map<ClassLabel, int>::iterator p=countHistogram.begin(); (p!=countHistogram.end()); ++p)  {
    ClassLabel l = (*p).first;
    classLabelList->push_back(l);
  }
}


int ClassLabelHistogram::getCountOfLabel (ClassLabel label)
{
  return countHistogram[label];
}

double ClassLabelHistogram::getWeightOfLabel (ClassLabel label)
{
  return weightsumHistogram[label];
}

double ClassLabelHistogram::getAvgWeightOfLabel (ClassLabel label)
{
  cout << endl << "weightsum " << weightsumHistogram[label] << " div count " << countHistogram[label];
  double result = weightsumHistogram[label]/((double)countHistogram[label]);
  cout << " return " << result;
  return result;
}

void ClassLabelHistogram::increaseCountOfLabel (ClassLabel label, double weight)
{
  memberCount++;
  if (countHistogram.find(label) != countHistogram.end()) {
    countHistogram[label] = countHistogram[label] + 1;
    weightsumHistogram[label] = weightsumHistogram[label] + weight;
  } else {
    countHistogram[label] = 1;
    weightsumHistogram[label] = weight;
  }
}

  
ClassLabel ClassLabelHistogram::getLabelWithMaxOccurance ()
{
  int max = 0;
  ClassLabel maxLabel = 0;
  for(map<ClassLabel, int>::iterator p=countHistogram.begin(); (p!=countHistogram.end()); ++p)
  {
    if ((*p).second > max)
    {
      max = (*p).second;
      maxLabel = (*p).first;
    }
  }
  return maxLabel;
}


ClassLabel ClassLabelHistogram::getRealLabelWithMaxOccurance ()
{
  int max = 0;
  ClassLabel maxLabel = 0;
  for(map<ClassLabel, int>::iterator p=countHistogram.begin(); (p!=countHistogram.end()); ++p)
  {
    if ((*p).first > 0) { // only process the classes different from 0
      if ((*p).second > max)
      {
        max = (*p).second;
        maxLabel = (*p).first;
      }
    }
  }
  return maxLabel;
}


ClassLabel ClassLabelHistogram::getRealLabelWithMaxWeight ()
{
  double max = 0;
  ClassLabel maxLabel = 0;
  for(map<ClassLabel, double>::iterator p=weightsumHistogram.begin(); (p!=weightsumHistogram.end()); ++p)
  {
    if ((*p).first > 0) { // only process the classes different from 0
      if ((*p).second > max)
      {
        max = (*p).second;
        maxLabel = (*p).first;
      }
    }
  }
  return maxLabel;
}


//**************************************************************************
//                         private Methods
//**************************************************************************

// --- none ---
