/***************************************************************************
                                ClassLabelHistogram.h
                             ------------------
    begin                : 16.11.2005
    copyright            : (C) 2005 by Frank Moosmann
    email                : frank.moosmann@inrialpes.fr
 ***************************************************************************/

#ifndef CLASSLABELHISTOGRAM_H
#define CLASSLABELHISTOGRAM_H

#include "ClassLabel.h"
// #include "FeatureVector.h"
#include <map>
#include <list>

class FeatureVector;

class ClassLabelHistogram {

public:
  
  ClassLabelHistogram();
  ClassLabelHistogram(std::list<FeatureVector*>* sampleList);
  ~ClassLabelHistogram();
  void reset ();
  unsigned int getTotalCount();
  void getListOfClassLabels (std::list<ClassLabel>* classLabelList);
  int getCountOfLabel (ClassLabel label);
  double getWeightOfLabel (ClassLabel label);
  double getAvgWeightOfLabel (ClassLabel label);
  void increaseCountOfLabel (ClassLabel label, double weight = 1.0);
  ClassLabel getLabelWithMaxOccurance (); // should not be used!!!
  ClassLabel getRealLabelWithMaxOccurance (); // get second best label, if best class was class_0
  ClassLabel getRealLabelWithMaxWeight (); // get second best label, if best class was class_0

private:
  unsigned int memberCount;
  std::map<ClassLabel, int> countHistogram;
  std::map<ClassLabel, double> weightsumHistogram;
};

class BinClassLabelHistogram {
 public:
  static const int POS=103;
  static const int NEG=102;

  BinClassLabelHistogram() { clear(); }
  void clear() {
    n_pos=n_neg=0;
    w_pos=w_neg=0;    
  }
  void increaseCountOfLabel (ClassLabel label, double weight = 1.0) {
    if(label==POS) {n_pos++; w_pos+=weight; }
    else           {n_neg++; w_neg+=weight; }
  }
  ClassLabel getRealLabelWithMaxWeight () {
    if(w_pos>w_neg) return POS;
    else return NEG;
  }
  double getWeightOfLabel (ClassLabel label) {
    if(label==POS) return w_pos;
    else return w_neg;
  }
 private:
  int n_pos,n_neg;
  double w_pos,w_neg;
};

#endif //CLASSLABELHISTOGRAM_H
