/***************************************************************************
                            DecisionNode.cpp
                             ------------------
    begin                : 07.11.2005
    copyright            : (C) 2005 by Frank Moosmann
    email                : frank.moosmann@inrialpes.fr
 ***************************************************************************/

#include "DecisionNode.h"
#include "FMUtils.h"
#include <cmath>
#include <iostream>

unsigned int DecisionNode::totalDecisionNodeCount = 0;

using namespace std;

/***************************************************************************
                         Constructors / Destructors
 ***************************************************************************/
DecisionNode::DecisionNode(int attributeNo, int attributeNo2, FeatureData splitThreshold, TreeNode* nodeAttributeSmaller, TreeNode* nodeAttributeGE, double scoreVal, unsigned int treeLevel) {
  parent = 0;
  attrNo = attributeNo;
  attrNo2 = attributeNo2;
  threshold = splitThreshold;
  score = scoreVal;
  level = treeLevel;
  nodeAttSmaller = nodeAttributeSmaller;
  nodeAttGE = nodeAttributeGE;
  totalDecisionNodeCount++;
}

DecisionNode::~DecisionNode ( ) {
  if (nodeAttSmaller != 0)
    delete nodeAttSmaller;
  if (nodeAttGE != 0)
    delete nodeAttGE;
  totalDecisionNodeCount--;
}



/***************************************************************************
                              public Methods
 ***************************************************************************/
TerminalNode* DecisionNode::classify (FeatureVector* featureVec, bool addFVecInfo) {
  if (attrNo2 == -1) { // compare with threshold
    if (featureVec->getElement(attrNo) < threshold)
      return nodeAttSmaller->classify(featureVec, addFVecInfo);
    else
      return nodeAttGE->classify(featureVec, addFVecInfo);
  } else { // compare with second attribute
    if (featureVec->getElement(attrNo) < featureVec->getElement(attrNo2))
      return nodeAttSmaller->classify(featureVec, addFVecInfo);
    else
      return nodeAttGE->classify(featureVec, addFVecInfo);
  }
}


void DecisionNode::writeToFile(ostream *out) {
  //TreeNode *parent;
  //unsigned int level;
  //double score;
  //int attrNo;
  //int attrNo2;
  //FeatureData threshold;
  //TreeNode* nodeAttSmaller;
  //TreeNode* nodeAttGE;
  (*out) << "DecisionNode;" << level << ";"
                            << score << ";"
                            << attrNo << ";"
                            << attrNo2 << ";"
                            << threshold << ";"
									 << endl;
  nodeAttSmaller->writeToFile(out);
  nodeAttGE->writeToFile(out);
}

/***************************************************************************
                              private Methods
 ***************************************************************************/
