/*!
    \file ExtraTreeClassifier.cpp


*/
 

#include "ExtraTreeClassifier.h"
#include "TreeNode.h"
#include "TerminalNode.h"
#include "DecisionNode.h"
#include "ClassLabel.h"

#include "FMUtils.h"
#include <iostream>
#include <istream>
#include <fstream>
#include <cmath>
#include <map>
#include <list>

#define PI 3.14159265

using namespace std;


#if 0

//***************************************************************************
//                        Utility functions
//***************************************************************************
extern "C" int compare(DecisionNode*, DecisionNode*);

void printDecisionPatterns(list<TerminalNode*> *tnList) {
  unsigned int listsize = tnList->size();
  //cout << "hello " << listsize << endl;
  unsigned int i = 0;
  for (list<TerminalNode*>::iterator p=tnList->begin(); p!=tnList->end(); p++) {
    if (i%(listsize/20) == 0) {
       //cout << "node accepted" << endl;
       Image image( "16x16", "white" );
       DecisionNode* currnode = dynamic_cast<DecisionNode*>((*p)->parent);
       cout << i << "    ";
       while (currnode != 0) {
         int pix = currnode->attrNo;
         cout << pix << " ";
         unsigned int x = (unsigned int)(pix/16);
         unsigned int y = (unsigned int)(pix%16);
         cout << pix << " (" << x << "," << y << ") ";
         image.pixelColor( x, y, "black" );
         currnode = dynamic_cast<DecisionNode*>(currnode->parent);
       }
       cout << endl;
       string fileout = string("auswertung/");
       fileout += to_String(i);
       fileout += string(".pgm");
       image.write(fileout.c_str());
    }
    i++;
  }
}

void ensureHeight(Magick::Image *img, int height) {
  int oldheight = img->rows();
  if (height > oldheight) {
    //cout << endl << "new height: " << height << flush;
    int width = img->columns();
    int diff = height-oldheight;
/*    string borderString = "0x"+to_String<int>(diff)+"+0+0";
    img->border(Geometry(borderString.c_str()));
    string cropString = to_String<int>(width)+"x"+to_String<int>(height)+"+0+"+to_String<int>(diff);
    img->crop(Geometry(cropString.c_str()));*/
    img->border(Geometry(0,diff,0,0));
    img->crop(Geometry(width,height,0,diff));
  }
}

void ensureWidth(Magick::Image *img, int width) {
  int oldwidth = img->columns();
  if (width > oldwidth) {
    //cout << endl << "new width: " << width << flush;
    int height = img->rows();
    int diff = width-oldwidth;
/*    string borderString = to_String<int>(diff)+"x0+0+0";
    img->border(Geometry(borderString.c_str()));
    string cropString = to_String<int>(width)+"x"+to_String<int>(height)+"+"+to_String<int>(diff)+"+0";
    img->crop(Geometry(cropString.c_str()));*/
    img->border(Geometry(diff,0,0,0));
    img->crop(Geometry(width,height,diff,0));
  }
}

void pixCompose(Magick::Image *master, Magick::Image *patch, unsigned int posX, unsigned int posY) {
  unsigned int width = patch->columns();
  unsigned int height = patch->rows();
  //cout << endl << "composing " << width*height << " pixels" << flush;
  for (unsigned int x=0; x<width; x++) {
    for (unsigned int y=0; y<height; y++) {
      master->pixelColor(posX+x,posY+y,patch->pixelColor(x,y));
    }
  }
}

void visualizeFeatures(Magick::Image *img, list<FeatureVector*> *fvList) {
  int listsize = fvList->size();
  unsigned int patchsize = 32;
  unsigned int maxpatchperline = 20;
  unsigned int emptypatches = 2;
  unsigned int distance = 5;
  unsigned int oldheight = img->rows();
  //unsigned int newwidth = listsize*(patchsize+distance)+3*distance;
  unsigned int newlines = 1;
  listsize -= maxpatchperline;
  //cout << endl << "entering loop with listsize " << listsize << flush;
  while (listsize > 0) {
    newlines++;
    listsize -= (maxpatchperline-emptypatches);
  }
  //cout << ", finished with newlines " << newlines << flush;
  unsigned int newheight = oldheight + distance + newlines*(patchsize+distance);
  unsigned int newwidth = maxpatchperline*(patchsize+distance)+3*distance;
  ensureHeight(img, newheight);
  ensureWidth(img, newwidth);
  unsigned int posX = 2*distance;
  unsigned int posY = oldheight;
    
  for (list<FeatureVector*>::iterator p=fvList->begin(); p!=fvList->end(); p++) {
    FeatureVector *currFV = *p;
    if (fileExists(currFV->imageName)) {
      Image imageFV(currFV->imageName.c_str());
      imageFV.crop(Geometry(currFV->windowSize, currFV->windowSize, currFV->windowPosX, currFV->windowPosY));
      imageFV.scale(Geometry(patchsize,patchsize));
      //img->composite(imageFV, posX, posY);   // <--- does not work!!
      pixCompose(img, &imageFV, posX, posY);
      posX += patchsize+distance;
      if (posX+patchsize > newwidth) {
        posX = 2*distance + emptypatches*(distance+patchsize);
        posY += patchsize+distance;
      }
    } // image file exists
  } // for each feature vector
}

void visualizeFeatures(Magick::Image *img, list<FeatureDetails> &fdList) {
  int listsize = fdList.size();
  unsigned int patchsize = 32;
  unsigned int maxpatchperline = 20;
  unsigned int emptypatches = 2;
  unsigned int distance = 5;
  unsigned int oldheight = img->rows();
  //unsigned int newwidth = listsize*(patchsize+distance)+3*distance;
  unsigned int newlines = 1;
  listsize -= maxpatchperline;
  //cout << endl << "entering loop with listsize " << listsize << flush;
  while (listsize > 0) {
    newlines++;
    listsize -= (maxpatchperline-emptypatches);
  }
  //cout << ", finished with newlines " << newlines << flush;
  unsigned int newheight = oldheight + distance + newlines*(patchsize+distance);
  unsigned int newwidth = maxpatchperline*(patchsize+distance)+3*distance;
  ensureHeight(img, newheight);
  ensureWidth(img, newwidth);
  unsigned int posX = 2*distance;
  unsigned int posY = oldheight;
    
  for (list<FeatureDetails>::iterator p=fdList.begin(); p!=fdList.end(); p++) {
    FeatureDetails currFD = *p;
    if (fileExists(currFD.imageName)) {
      string boxcolor;
      switch (currFD.label) {
        case 0: boxcolor = "black"; break;
        case 1: boxcolor = "yellow"; break;
        case 2: boxcolor = "red"; break;
        case 3: boxcolor = "blue"; break;
        case 4: boxcolor = "green"; break;
        default: boxcolor = "pink";
      }
      img->strokeWidth(2);
      img->strokeColor(boxcolor);
      img->draw(DrawableRectangle(posX-1, posY-1, posX+patchsize+1, posY+patchsize+1));
      Image imageFV(currFD.imageName.c_str());
      imageFV.crop(Geometry(currFD.size, currFD.size, currFD.posX, currFD.posY));
      imageFV.scale(Geometry(patchsize,patchsize));
      //img->composite(imageFV, posX, posY);   // <--- does not work!!
      pixCompose(img, &imageFV, posX, posY);
      posX += patchsize+distance;
      if (posX+patchsize > newwidth) {
        posX = 2*distance + emptypatches*(distance+patchsize);
        posY += patchsize+distance;
      }
    } // image file exists
  } // for each feature vector
}
#endif
 
//***************************************************************************
//                         Constructors / Destructors
//***************************************************************************

/*!
  \brief Constructor to build decision trees from the examples

  This method creates the decision trees based on the given examples
  \param featVecList the list with FeatureVectors and ClassLabels
  \param params structure containing the parameters
*/
ExtraTreeClassifier::ExtraTreeClassifier (list<FeatureVector*> &featVecList, ExtraTreeParams &params)
{
  // fully grow the trees
  ClassLabelHistogram totalHistogram = ClassLabelHistogram(&featVecList);
  totalHistogram.getListOfClassLabels(&classLabelList);                 // set up classLabelList
  for (int i=0; i<params.noOfTrees; i++)  {
    cout << "." << flush; 
    rootNodeList.push_back(constructSubTree(featVecList, totalHistogram, 0, params));
    //cout << "tree built" << endl << flush; 
  }
  
  // do pruning until specified number of leaf nodes are reached:
  //cout << "start pruning" << endl << flush; 
  pruneTrees(params);
  
  // reset IDs of terminal nodes:
  unsigned int currID = 0;
  for (list<TerminalNode*>::iterator p=tnlist.begin(); p!=tnlist.end(); ++p)  {
    currID++;
    (*p)->setID(currID);
  }
  maxTnID=currID;
}


ExtraTreeClassifier::ExtraTreeClassifier (std::string filename)
{
  ifstream inFile (filename.c_str());
  loadFromifstream(inFile);
  inFile.close();
}

ExtraTreeClassifier::ExtraTreeClassifier (std::ifstream &inFile) {
  loadFromifstream(inFile);
}
void ExtraTreeClassifier::loadFromifstream(std::ifstream &inFile) {
  // read treecount
  if (! inFile.eof() )
  {
    string s;
    getline(inFile,s);
    //cout << s << endl;
    int treecount = 0;
    char classlabels [255];
    if (sscanf(s.c_str(), "ExtraTreeClassifier;%d%s", &treecount, classlabels) == 2) {
      string remainingLabels = classlabels;
      while (remainingLabels.length() > 0) {
        int label;
        char remainder [255];
        int count = sscanf(remainingLabels.c_str(), ";%d%s", &label, remainder);
        if (count >= 1)
          classLabelList.push_back(label);
        if (count == 2)
          remainingLabels = remainder;
        else
          remainingLabels = "";
      }
      for (int i=0; i<treecount; i++) {
      TreeNode* tn = readSubTree(&inFile);
      rootNodeList.push_back(tn);
      }
    } // scanf first line
  } // !eof

}

ExtraTreeClassifier::~ExtraTreeClassifier ( )
{
  for(list<TreeNode*>::iterator p=rootNodeList.begin(); p!=rootNodeList.end(); ++p)
  {
    TreeNode* rootNode = *p;
    if (rootNode != 0) {
      delete rootNode; // this will delete the whole decision tree;
    }
  }
}



//***************************************************************************
//                         public Methods
//***************************************************************************

void ExtraTreeClassifier::writeToFile(std::string filename) {
  ofstream outfile;
  outfile.open(filename.c_str(), ios::out);
  writeToFile(outfile);  
  outfile.close();
}


void ExtraTreeClassifier::writeToFile(std::ofstream &outfile)
{
  outfile << "ExtraTreeClassifier;" << rootNodeList.size();
  for(list<ClassLabel>::iterator p=classLabelList.begin(); p!=classLabelList.end(); ++p)
    outfile << ";" << *p;
  outfile << endl;
  for(list<TreeNode*>::iterator p=rootNodeList.begin(); p!=rootNodeList.end(); ++p)
  {
    TreeNode* rootNode = *p;
    if (rootNode != 0) {
      rootNode->writeToFile(&outfile);
    }
  }
}

/*!
  \brief Method for classifying a new example

  This method classifies the new example based on the decision trees
  created with the method learnFromExamples().
  \param featureVec the list with FeatureVectors of the example
  \param totalVotes pointer to the histrogram where the votes for all windows will be stored (optional)
  \param windowVotes pointer to the histrogram where the votes for the class will be stored (optional)
  \return the estimated class
*/
ClassLabel ExtraTreeClassifier::classifyBin (list<FeatureVector*> *featureVec, BinClassLabelHistogram *totalVotes, BinClassLabelHistogram *windowVotes,bool saveAllClassHistrogams)
{
  return classifyBin (featureVec, 0, false, totalVotes, windowVotes, saveAllClassHistrogams);
}

ClassLabel ExtraTreeClassifier::classifyBin (list<FeatureVector*> *featureVec, std::list<TerminalNode*> *tnList, bool addFVecInfo, BinClassLabelHistogram *totalVotes, BinClassLabelHistogram *windowVotes,bool saveAllClassHistrogams)
{
  //debug_out(1, "starting to classify...");  
  BinClassLabelHistogram featureVotes; // used to classify the list of feature vectors

  // loop over features
  for(list<FeatureVector*>::const_iterator featI=featureVec->begin(); (featI!=featureVec->end()); ++featI)
  {
    FeatureVector* fVec = *featI;
    BinClassLabelHistogram currFeatureVotes;
    // loop over trees
    for(list<TreeNode*>::const_iterator treeI=rootNodeList.begin(); (treeI!=rootNodeList.end()); ++treeI)
    {
      TerminalNode* tn = (*treeI)->classify(fVec, addFVecInfo);
      ClassLabel classLabel = tn->classLabel;
      double labelWeight = tn->weight;
      if (tnList != 0) {
        tnList->push_back(tn);
      }
      //debug_out(6, "voting for class %i", classLabel);
      featureVotes.increaseCountOfLabel(classLabel, labelWeight);
      currFeatureVotes.increaseCountOfLabel(classLabel, labelWeight);
      if (totalVotes != 0)
        totalVotes->increaseCountOfLabel(classLabel, labelWeight);
    }
    //fVec->classLabel = currFeatureVotes->getRealLabelWithMaxOccurance();
    fVec->classLabel = currFeatureVotes.getRealLabelWithMaxWeight();
    if(saveAllClassHistrogams) {
      fVec->classLabelHist=currFeatureVotes;
    }
    if (windowVotes != 0)
      windowVotes->increaseCountOfLabel(fVec->classLabel);
  }
  //ClassLabel classifiedLabel = featureVotes->getRealLabelWithMaxOccurance();
  ClassLabel classifiedLabel = featureVotes.getRealLabelWithMaxWeight();
  if ((classLabelList.size() == 2) && (totalVotes != 0)) { // in two-class-case create statistic about confidence of decision
    //cout << endl;
    list<ClassLabel>::iterator p=classLabelList.begin();
    //confidenceValue = -totalVotes->getCountOfLabel(*p);
    confidenceValue = -totalVotes->getWeightOfLabel(*p);
    //confidenceValue = -totalVotes->getAvgWeightOfLabel(*p);
    //cout << endl << " --> " << confidenceValue;
    ++p;
    //confidenceValue += totalVotes->getCountOfLabel(*p);
    confidenceValue += totalVotes->getWeightOfLabel(*p);
    //confidenceValue += totalVotes->getAvgWeightOfLabel(*p);
    //cout << endl << " --> " << confidenceValue;
  }

  return classifiedLabel;
}


void ExtraTreeClassifier::visualizeLeafNodes(string outputImage)
{
  cout << "visualizing leaf nodes..." << flush;
  //printDecisionPatterns(&tnList);
#if 0
  for (list<TerminalNode*>::iterator p=tnlist.begin(); p!=tnlist.end(); ++p)  {
    Image *img = new Image( "10x10", "white" );
    img->borderColor(ColorRGB("white"));
    visualizeFeatures(img, (*p)->trainingFeatures);
    if ((*p)->testingFeatures.size() > 0)
      visualizeFeatures(img, (*p)->testingFeatures);
    string fn = outputImage + "." + to_String((*p)->getID()) + ".jpg";
    img->write(fn.c_str());
  }
#endif
  cout << endl;
}


//***************************************************************************
//                         private Methods
//***************************************************************************

/*!
  \brief Method to read a subtree out of a file

  \param in input stream (already opened file)
  \return the pointer to root node of the constructed subtree
*/
TreeNode* ExtraTreeClassifier::readSubTree (istream *in)
{
  if (! (*in).eof() )
  {
    string s;
    getline((*in),s);
    // split lines into values
    if (s.find("DecisionNode;") == 0)
    {
      unsigned int level;
      double score;
      int attrNo;
      int attrNo2;
      FeatureData threshold;
      //(*out) << "DecisionNode;" << level << ";" << score << ";" << attrNo << ";" << attrNo2 << ";" << threshold << ";" << endl;
      if (sscanf(s.c_str(), "DecisionNode;%u;%lf;%d;%d;%f", &level, &score, &attrNo, &attrNo2, &threshold ) == 5)
			{
        TreeNode *nodeAttSmaller = readSubTree(in);
        TreeNode *nodeAttGE = readSubTree(in);
        DecisionNode *dnode = new DecisionNode(attrNo, attrNo2, threshold, nodeAttSmaller, nodeAttGE, score, level);
        dnlist.push_back(dnode);
        nodeAttSmaller->parent = dnode;
        nodeAttGE->parent = dnode;
        return dnode;
      }
    } 
    if (s.find("TerminalNode;") == 0)
    {
      unsigned int id;
      unsigned int level;
      double score;
      ClassLabel classLabel;
      double weight;
      double windowPosXRatio;
      double windowPosYRatio;
      double windowSizeRatio;
      double windowToObjectRatio;
      //(*out) << "TerminalNode;" << id << ";" << level << ";" << score << ";" << classLabel << ";" << weight << ";" 
      //                          << windowPosXRatio << ";" << windowPosYRatio << ";" << windowSizeRatio << ";" << windowToObjectRatio << ";" << endl;
      if (sscanf(s.c_str(), "TerminalNode;%u;%u;%lf;%d;%lf;%lf;%lf;%lf;%lf", &id, &level, &score, &classLabel, &weight, &windowPosXRatio, &windowPosYRatio, &windowSizeRatio, &windowToObjectRatio ) == 9)
			{
        TerminalNode *tnode = new TerminalNode(id, level, score, classLabel, weight, windowPosXRatio, windowPosYRatio, windowSizeRatio, windowToObjectRatio);
        tnlist.push_back(tnode);
        return tnode;
      }
    }
  } // !eof
  return 0;
}

 
/*!
  \brief Method to construct a subtree for the decision tree

  Method with wich the decision tree is built recursively. Based on the 
  feature vectors it is decided if a decision node or a leaf is needed.
  If a decision node was needed the set of feature vecors are split and
  the subtrees are constructed recursively with these subsets.
  \param featVecList the list of featureVectors with ClassLabels
  \param size number of elements in the featVecList
  \param histogram disitribution of the elements in featVecList concerning class labels
  \param dnlist list with DecisionNodes where an entry will be created in case a DecisionNode is created
  \param level current level of the tree. Will be stored as information in the node
  \param params structure containing the parameters
  \return the pointer to root node of the constructed subtree
*/
TreeNode* ExtraTreeClassifier::constructSubTree(list<FeatureVector*> &featVecList, ClassLabelHistogram &histogram, unsigned int level, ExtraTreeParams &params)
{
  //cout << endl << "building subtree in level " << level << " from " << featVecList->size() << " features";
  if (stopSplitting(featVecList))
  {
    // **** create terminal node ****
    ClassLabel maxLabel = histogram.getLabelWithMaxOccurance();
    //cout << endl << "constructing terminal node with classLabel " << maxLabel << flush;
    TerminalNode* n = new TerminalNode(maxLabel, level, &featVecList);
    tnlist.push_back(n);
    return n;
  }
  else
  {
    // **** create decision node ****
    int attrNo;
    int attrNo2 = -1;
    int attrMax = NULL_CLASS_LABEL;
    int attr2Max = -1;
    FeatureData mean;
    FeatureData deviation;
    FeatureData threshold;
    FeatureData thresholdMax = 0;
    FeatureData deviationMax = 0;
    int noTries = 0;
    double scoreVal = 0.0;
    double scoreMax = 0.0;
    list<FeatureVector*> featListLe;
    list<FeatureVector*> featListGE;
    ClassLabelHistogram histogramLe;
    ClassLabelHistogram histogramGE;
    ClassLabelHistogram histogramHSLe;
    ClassLabelHistogram histogramHSGE;
    unsigned int availableAttrSize = (*(featVecList.begin()))->getSize();
    list<int> availableAttr;
    for (unsigned int i=0; i<availableAttrSize; i++) {
      availableAttr.push_back(i);
    }
    int listEntryNo;
    bool continueCriteria;
    do {
      noTries++;
      //cout << endl << "   trying ";
      // 1) Select an attribute randomly
      listEntryNo = randomInRange((unsigned int)0, availableAttrSize-1);
      list<int>::iterator it = availableAttr.begin();
      for (int j=0; j<listEntryNo; j++) {it++;}
      attrNo = *it;
      availableAttr.erase(it);
      availableAttrSize--;
      //debug_out(5, "attribute number %i selected", attrNo);
      // 2a) Select a second attribute randomly
      if (params.useCompareAtt) {
        attrNo2 = randomInRange((unsigned int)0, (*(featVecList.begin()))->getSize());
        //debug_out(5, "attribute number %i selected as second attribute", attrNo2);
      } else {
        // 2a) Select a threshold randomly according to normal distribution N(mean,deviation)
        mean = FeatureVector::compEmpMean(featVecList, attrNo);
        deviation = FeatureVector::compStdDeviation (featVecList, attrNo, mean);
        threshold = randomStdDistrib (mean, deviation);
/*        cout << "  mean " << mean;
        cout << "  deviation " << deviation;
        cout << "  threshold " << threshold;*/
      }
      if (deviation > 0) {
        featListLe.clear();
        featListGE.clear();
        histogramLe.reset();
        histogramGE.reset();
        histogramHSLe.reset();
        histogramHSGE.reset();
        splitSamples(featVecList, attrNo, attrNo2, threshold, deviation, params.useSoftThreshold,
                     featListLe, histogramLe, histogramHSLe,
                     featListGE, histogramGE, histogramHSGE);
        //cout << "  splitted with " << countListLe << "/" << countListGE;
        if (params.useSimpleScore)
          scoreVal = score2(histogramHSLe, histogramHSGE);
        else
          scoreVal = score1(histogram, histogramHSLe, histogramHSGE);
        //cout << endl << " score " << scoreVal << " for split ratio " << histogramHSLe.getTotalCount() << "/" << histogramHSGE.getTotalCount() << " or " << histogramLe.getTotalCount() << "/" << histogramGE.getTotalCount() << flush;
        if (scoreVal > scoreMax) {
          scoreMax = scoreVal;
          attrMax = attrNo;
          attr2Max = attrNo2;
          thresholdMax = threshold;
          deviationMax = deviation;
        }
      } else {
        //cout << "x";
        scoreVal = 0;
      }
      continueCriteria = ((scoreVal < params.scoreThreshold) && (noTries < params.noScoreTries)) || (scoreMax == 0.0);
    } while ((continueCriteria) && (availableAttrSize > 0));
    if ((availableAttrSize == 0) && (scoreMax == 0.0)) {
      //cout << endl << " --> moehb!!! --> Terminal Node" << flush;
      //cout << "o";
      ClassLabel classLabel = histogram.getLabelWithMaxOccurance();
      //ClassLabel classLabel = 0;
      TerminalNode* n = new TerminalNode(classLabel, level, &featVecList);
      tnlist.push_back(n);
      return n;
    } else {
      if (attrNo != attrMax) {
        // resplit the samples with attribut that achieved the highest score
        attrNo = attrMax;
        attrNo2 = attr2Max;
        threshold = thresholdMax;
        deviation = deviationMax;
        featListLe.clear();
        featListGE.clear();
        histogramLe.reset();
        histogramGE.reset();
        histogramHSLe.reset();
        histogramHSGE.reset();
        splitSamples(featVecList, attrNo, attrNo2, threshold, deviation, params.useSoftThreshold,
                     featListLe, histogramLe, histogramHSLe,
                     featListGE, histogramGE, histogramHSGE);
      }
      // 6) create decision node
      TreeNode *nodeLe = constructSubTree(featListLe, histogramLe, level+1, params);
      TreeNode *nodeGE = constructSubTree(featListGE, histogramGE, level+1, params);
      double assignedScore = scoreMax;
      DecisionNode *dnode = new DecisionNode(attrNo, attrNo2, threshold, nodeLe, nodeGE, assignedScore, level);
      dnlist.push_back(dnode);
      nodeLe->parent = dnode;
      nodeGE->parent = dnode;
      return dnode;
    }
  } // if not stopsplitting
  return 0;
}

/*!
  \brief comparison function to sort DecisionNodes
*/
extern "C"
{
  int compare(DecisionNode *x, DecisionNode *y)
  {
    double fx = ((double)(max(max(x->score, x->nodeAttSmaller->score), x->nodeAttGE->score))) / ((double)x->level+1.0);
    double fy = ((double)(max(max(y->score, y->nodeAttSmaller->score), y->nodeAttGE->score))) / ((double)y->level+1.0);
    if (fx < fy) return 1;
    return 0;
  }
}

/*!
  \brief Method to start pruning on the tree

  This method prunes the tree according to the parameters set in params.
  \param dnlist list with pointers to all the decision nodes of the tree
  \param params parameter settings for tree learning
*/
void ExtraTreeClassifier::pruneTrees(ExtraTreeParams &params)
{
  unsigned int maxLevel = 0;
  unsigned int dncount = 0;
  for (list<DecisionNode*>::iterator p=dnlist.begin(); p!=dnlist.end(); p++) {
    maxLevel = max(maxLevel, (*p)->level);
    dncount++;
  }
  //cout <<endl<< "doing pruning on " << dncount << " decision nodes"<< flush;
  unsigned int tncount = dncount+1; //this is true for binary trees
  
  dnlist.sort(compare);
  
  unsigned int maxNoToPrune = 0;
  if (params.maxNoTerminalNodes > 0)	// according to parameter --maxTerminalNodes
    maxNoToPrune = max(0, (int)tncount - (int)(params.maxNoTerminalNodes));
  if (params.maxNoPrunedNodes > 0)		// according to parameter --maxPruningPercentage
    maxNoToPrune = params.maxNoPrunedNodes * dncount / 100;
  
  unsigned int noPrunedNodes = 0;
  list<DecisionNode*>::iterator p=dnlist.begin();
  list<DecisionNode*> dnToDelete;
  while ((noPrunedNodes < maxNoToPrune) && (p != dnlist.end())) {
    // select a node
    DecisionNode* currDN = *p;
    // decide weather to prune
    bool doPruning = (((currDN->score < params.pruningThreshold) && (currDN->level > maxLevel/4))
                      || ((params.maxNoTerminalNodes > 0) && (currDN->level > 0)));
    if (doPruning) {
      //cout << endl << "decided to prune node with score " << currDN->score << " in level " << currDN->level << flush;
      DecisionNode *nodeP = dynamic_cast <DecisionNode *> (currDN->parent);
      TerminalNode *tn = new TerminalNode(0, currDN->level);;
      tnlist.push_back(tn);
      if (nodeP) {
        tn->parent = nodeP;
        if (nodeP->nodeAttSmaller == currDN)
          nodeP->nodeAttSmaller = tn;
        else
          nodeP->nodeAttGE = tn;
      }
      deleteOutOfLists(currDN, false, &noPrunedNodes, &(tn->trainingFeatures));
      dnToDelete.push_back(currDN);
      delete currDN; // free memory of whole pruned subtree
      noPrunedNodes++;
      //cout << endl << "pruned nodes (total): " << noPrunedNodes << flush;
    }
    p++;
  }
  for (list<DecisionNode*>::iterator p=dnToDelete.begin(); p!=dnToDelete.end(); p++) {
    dnlist.remove(*p);
  }
  cout << params.noOfTrees << " trees built with " << dncount << " decision nodes and maxLevel " << maxLevel << ", afterwards pruned: " << noPrunedNodes << flush;
}

/*!
  \brief Method delete all pointers to any node in a tree out of a list

  This method deletes all pointers in the list "dnlist" which point to any node
  in the tree starting with node "node".
  \param node the root node of the tree
  \param dnlist the list to delete pointers from
  \param deleteThis weather to delete also the pointer to the root node
  \param noDeletedNodes possibility to keep track of how many pointers were deleted
*/
void ExtraTreeClassifier::deleteOutOfLists(DecisionNode *node, bool deleteThis, unsigned int *noDeletedNodes, std::list<FeatureDetails> *featureDetails)
{
  // delete out of list all decision nodes of Le-subtree
  DecisionNode *dnode;
  TerminalNode *tnode;
  dnode = dynamic_cast <DecisionNode *> (node->nodeAttSmaller);
  if (dnode) {
    deleteOutOfLists(dnode, true, noDeletedNodes, featureDetails);
  } else {
    tnode = dynamic_cast <TerminalNode *> (node->nodeAttSmaller);
    if (tnode) {
      appendList(&(tnode->trainingFeatures), featureDetails);
      tnlist.remove(tnode);
    }
  }
  // delete out of list all decision nodes of GE-subtree
  dnode = dynamic_cast <DecisionNode *> (node->nodeAttGE);
  if (dnode) {
    deleteOutOfLists(dnode, true, noDeletedNodes, featureDetails);
  } else {
    tnode = dynamic_cast <TerminalNode *> (node->nodeAttGE);
    if (tnode) {
      appendList(&(tnode->trainingFeatures), featureDetails);
      tnlist.remove(tnode);
    }
  }
  // delete decision node itself
  //if (typeid(*node)==typeid(DecisionNode)) // see rtti - run time type information
  if (deleteThis) {
    dnlist.remove(node);
    if (noDeletedNodes != 0)
      (*noDeletedNodes)++;
  }
}

/*!
  \brief Method that decides weather splitting should be stopped

  This method decides weather the given featureVectorList should be further
  splitted in the decision tree.
  \param featVecList the list with FeatureVectors and ClassLabels
  \return decides weather splitting should be stopped
*/
bool ExtraTreeClassifier::stopSplitting (list<FeatureVector*> &featVecList)
{
  //debug_out(4, "checking 'stopSplitting' criteria...");
  // compare the attributes and classLabels of the feature vectors
  bool attributesAllSame = true;
  bool classLabelAllSame = true;
  int count = 0;
  FeatureVector* lastSample;
  FeatureVector* currSample = 0;
  //debug_out(5, "iterating through feature list...");
  for(list<FeatureVector*>::iterator p=featVecList.begin(); ((p!=featVecList.end())&&(attributesAllSame||classLabelAllSame)); ++p)
  { 
    //debug_out(6, "iteration...");
    count++;
    lastSample = currSample;
    currSample = *p;
    if (lastSample != 0)
    {
      if (lastSample->classLabel != currSample->classLabel)  {
        classLabelAllSame = false;
      }
      unsigned int availableAttrSize = currSample->getSize();
      for (unsigned int i=0; (i<availableAttrSize) && (attributesAllSame); i++) {
        //debug_out(7, "getting element...");
        if (currSample->getElement(i) != lastSample->getElement(i)) {
          attributesAllSame = false;
        }
      }
    }
  }
  // debug output
/*  if (classLabelAllSame)
    cout << "  classLabels all the same  ";
  if (attributesAllSame)
    cout << "  attributes all the same  ";*/
  
  // if there are < 2 featureVectors left then stop splitting
  if (count < 2)
    return true;
  // if either attributes or classLabels are all equal then stop splitting, otherwise continue splitting
  else
    return (attributesAllSame || classLabelAllSame);
}


/*!
  \brief Method to get a random number which has normal distribution

  \param mean the empirical mean
  \param deviation the standard deviation
  \return a random number
*/
double ExtraTreeClassifier::randomStdDistrib (double mean, double deviation)
{
  // calclulation method of Box-Mueller
  double y1 = randomInRange(0.0,0.99999999);
  double y2 = randomInRange(0.0,0.99999999);
  return (sqrt( -2.0 * log(y1)) * cos(2.0 * PI * y2)) * deviation + mean;
  //debug_out(0, "!!!!ExtraTreeClassifier::randomStdDistrib still fake-random-implementation!!!!");
  //return randomInRange(mean-deviation, mean+deviation);
}

 
/*!
  \brief Method to split the set of samples

  This method splits the given set of samples into two new sets where 
  the resListLe contains all samples where the value of attribute number attrNo
  is less than threshold and resListGE contains all samples where the value of
  attribute number attrNo is greater or equal than threshold 
  \param featVecList the list of featureVectors with ClassLabels
  \param attrNo the number of the attribute to decide on
  \param attrNo2 the number of the attribute 2 to compare with (-1 if threshold should be used)
  \param threshold the decision threshold
  \param resListLe pointer to the resulting list with attribute < threshold
  \param countListLe number of items of in list resListLe
  \param resHistogramLe histogram of the distribution of classes in resListLe
  \param resListGE pointer to the resulting list with attribute >= threshold
  \param countListGE number of items of in list resListLe
  \param resHistogramGE histogram of the distribution of classes in resListGE
*/
void ExtraTreeClassifier::splitSamples (list<FeatureVector*> &featVecList, int attrNo, int attrNo2, FeatureData threshold, FeatureData deviation, bool softThreshold,
                                        list<FeatureVector*> &resListLe, ClassLabelHistogram &resHistogramLe, ClassLabelHistogram &hardSplitHistLe,
                                        list<FeatureVector*> &resListGE, ClassLabelHistogram &resHistogramGE, ClassLabelHistogram &hardSplitHistGE)
{
  //debug_out(5, "splitting samples...");
  FeatureVector* csample;
  for(list<FeatureVector*>::iterator p=featVecList.begin(); p!=featVecList.end(); ++p)
  {
    csample = *p;
    double attrVal = csample->getElement(attrNo);
    //debug_out(6, "deciding on attribute value %lf with threshold %lf",attrVal,threshold);
    bool goesLe;
    bool goesRi;
    bool goesHardLe;
    if (attrNo2 == -1) {
      goesHardLe = (attrVal < threshold);
      if (softThreshold) {
        goesLe = (attrVal <= threshold+(deviation/6.0));
        goesRi = (attrVal >= threshold-(deviation/6.0));
      } else {
        goesLe = goesHardLe;
        goesRi = !goesLe;
      }
    } else {
      FeatureData attrVal2 = csample->getElement(attrNo2);
      goesHardLe = (attrVal < attrVal2);
      goesLe = goesHardLe;
      goesRi = !goesLe;
    }
    if (goesHardLe)
      hardSplitHistLe.increaseCountOfLabel(csample->classLabel);
    else
      hardSplitHistGE.increaseCountOfLabel(csample->classLabel);
    if (goesLe) {
      resListLe.push_back(csample);
      resHistogramLe.increaseCountOfLabel(csample->classLabel);
      //debug_out(7, "goes left");
    }
    if (goesRi) {
      resListGE.push_back(csample);
      resHistogramGE.increaseCountOfLabel(csample->classLabel);
      //debug_out(7, "goes right");
    }
  }
}


/*!
  \brief Method to calculate the score of a split

  \param size total number of elements
  \param histogram histogram containing the distribution of the elements concerning the class labels
  \param countListLe number of elements of the first partition
  \param histogramLe histogram containing the distribution of the elements in the first partition
  \param countListGE number of elements of the second partition
  \param histogramGE histogram containing the distribution of the elements in the second partition
  \return the score of the set in the range 0.0 - 1.0
*/
double ExtraTreeClassifier::score1 (ClassLabelHistogram &histogram, ClassLabelHistogram &histogramLe, ClassLabelHistogram &histogramGE)
{
  /*
  if ((countListLe == 0) || (countListGE == 0))
    return 0.0;
  else
    return 1.0;
  */
  double size = histogram.getTotalCount();
  double countListLe = histogramLe.getTotalCount();
  double countListGE = histogramGE.getTotalCount();
  double result;
  if ((countListLe == 0) || (countListGE == 0))
  {
    result = 0.0;
  }
  else
  {
    result = 2 * i_t_c( (double)size, histogram,
                        (double)countListLe, histogramLe,
                        (double)countListGE, histogramGE)
               / ( h_t((double)size, (double)countListLe, (double)countListGE)
                  +h_c((double)size, histogram) );
  }//
  //debug_out(2, "score is %lf", result);
  return result;
}

/*!
  \brief Method to calculate the score of a split

  \param countListLe number of elements of the first partition
  \param countListGE number of elements of the second partition
  \return the score of the set in the range 0.0 - 1.0
*/
double ExtraTreeClassifier::score2 (ClassLabelHistogram &histogramLe, ClassLabelHistogram &histogramGE)
{
  int countListLe = histogramLe.getTotalCount();
  int countListGE = histogramGE.getTotalCount();
  if ((countListLe == 0) || (countListGE == 0))
    return 0.0;
  else
  {
    double result;
    if (countListLe > countListGE)
      result = (double)countListGE/(double)countListLe;
    else
      result = (double)countListLe/(double)countListGE;
    return result*result;
  }
}

/*!
  \brief calculate the entropy of the current set
*/
double ExtraTreeClassifier::h_c (double size, ClassLabelHistogram &histogram)
{
  // for all classes calculate the formula Nc/N * log Nc/N and sum it up
  double sum = 0;
  for (list<ClassLabel>::iterator p=classLabelList.begin(); (p!=classLabelList.end()); ++p)  {
    double nc = (double)(histogram.getCountOfLabel(*p));
    if (nc != 0)
      sum -= (nc/size) * log2(nc/size);
  }
  return sum;
}


/*!
  \brief calculate the entropy of the test outcome
*/
double ExtraTreeClassifier::h_t (double totalSize, double partitionOneSize, double partitionTwoSize)
{
  double sum = 0;
  sum -= (partitionOneSize/totalSize) * log2 (partitionOneSize/totalSize);
  sum -= (partitionTwoSize/totalSize) * log2 (partitionTwoSize/totalSize);
  return sum;
}


/*!
  \brief calculate the information gain / reduction of impurity of a test
*/
double ExtraTreeClassifier::i_t_c (double size, ClassLabelHistogram &histogram,
                                   double partitionOneSize, ClassLabelHistogram &partitionOneHist,
                                   double partitionTwoSize, ClassLabelHistogram &partitionTwoHist)
{
  double sum = h_c (size, histogram); // entropy of the current set
  sum -= (partitionOneSize/size) * h_c (partitionOneSize, partitionOneHist); // entropy of the test outcome
  sum -= (partitionTwoSize/size) * h_c (partitionTwoSize, partitionTwoHist); // entropy of the test outcome
  return sum;
}
   

