/***************************************************************************
                  FMUtils.h  -  some small helper methods
                             -------------------
    begin                : 04.11.2005
    copyright            : (C) 2005 by Frank Moosmann
    email                : frank.moosmann@inrialpes.fr
 ***************************************************************************/

#ifndef FMUTILS_H
#define FMUTILS_H

#include "FeatureVector.h"
#include <string>
#include <list>
#include <vector>
#include <iostream>
#include <sstream>

/***************************************************************************
   String utility functions
 ***************************************************************************/
// to convert a string into another type
template<class T> bool from_string(T& t, const std::string& s, std::ios_base& (*f)(std::ios_base&)) { 
  std::istringstream iss(s);
  return !(iss >> f >> t).fail();
}
// to convert another type into a string
template<typename T> std::string to_String(const T& x) {
  std::ostringstream oss;
  oss << x;
  return oss.str();
}
// decompose a string into substrings
void explode(std::string s, std::string e, std::vector<std::string>* ret);

/***************************************************************************
   Functions for the list template
 ***************************************************************************/
template<class T> void appendList(std::list<T> *origin, std::list<T> *target) { 
  typename std::list<T>::iterator p;
  for (p=origin->begin(); p!=origin->end(); p++) {
    target->push_back(*p);
  }
}

/***************************************************************************
   Function for easier output of debug information
   set DEBUG_LVL to control amount of output
 ***************************************************************************/
void debug_out(const int debug_level, const char* message);
void debug_out(const int debug_level, const char* message, double value1, double value2=0.0);
void debug_out(const int debug_level, const char* message, int value1, int value2=0);

/***************************************************************************
   Initialisation routine for the random generator
 ***************************************************************************/
void randomInit();

/***************************************************************************
   Function to get a random number within a specific range
 ***************************************************************************/
unsigned int randomInRange(const unsigned int lowest, const unsigned int highest);
double randomInRange(const double lowest, const double highest);

/***************************************************************************
   File utility functions
 ***************************************************************************/
bool fileExists(std::string filename);
int getFiles(std::list<std::string>* fileList, std::string filename, bool recursive=true, std::list<std::string>* extList = 0);
bool hasExtension(std::string filename, std::list<std::string>* extList);
int getSubdirs(std::list<std::string>* fileList, std::string dirName);
std::string extractFileName(std::string file, bool withExtension = true);
std::string extractPath(std::string file);

#endif /*FMUTILS_H*/
