/*!
    \file FeatureExtractor.h

    $Author: douze $ ( < moosmann@mrt.uka.de>)
    $Date: 2007/10/30 08:04:04 $
*/


#ifndef FEATURE_EXTRACTOR_H
#define FEATURE_EXTRACTOR_H

#include "FeatureVector.h"
#include "IntegralImage.h"

#include <list>
#include <string>
#include <Magick++.h>

#define WINDOW_SIZE 16              // subwindow is 16x16pixel

class FeatureExtractor {

public:

  /*! \brief extraction method */
  enum ExtractionMethod {
    ColorHSL,    /*!< pixels in HSL color space */
    ColorRGB,    /*!< pixels in RGB color space */
    Greyscale,   /*!< pixels in greyscale */ 
    WaveletHSL,  /*!< wavelet transform of pixels in HSL color space */ 
    WaveletRGB,  /*!< wavelet transform of pixels in RGB color space */ 
    WaveletGrey, /*!< wavelet transform of greyscale pixels */ 
    Gradient,    /*!< gradient, pixelwise */
    SIFT,        /*!< SIFT-Descriptor */
    normSIFT ,   /*!< normalized SIFT-Descriptor (rotational invariant)*/
    SIFTHSL,     /*!< SIFT-Descriptor per Color-Chanel */
    HSLplusSIFT  /*!< SIFT-Descriptor plus concatenated 4x4 bin with Color Value */
  };

  /*! \brief For printout purposes, returns the string of the extraction method */
  static std::string asText(ExtractionMethod method);
  static ExtractionMethod asExtractionMethod (std::string method);
        
  FeatureExtractor(Magick::Image img, std::string imgName, ExtractionMethod method);
  ~FeatureExtractor();
  
  void computeFeatures (std::list<FeatureVector*>& featureList);

private:
  std::string imageName;
  Magick::Image masterImg;
  unsigned int im_width;
  unsigned int im_height;
  IntegralImage *integImg;
  ExtractionMethod extractionMethod;
  int maxSubwinSize;
  double relObjSize;
  void extractHSLFeatures(Magick::Image window, unsigned int windowSize, FeatureVector* featureVec);
  void extractRGBFeatures(Magick::Image window, unsigned int windowSize, FeatureVector* featureVec);
  void extractGreyFeatures(Magick::Image window, unsigned int windowSize, FeatureVector* featureVec);
  void extractHSLBins(unsigned int subwindowSize, unsigned int xOffset, unsigned int yOffset, unsigned int nbBins, FeatureVector* featureVec);
  void extractWaveletFeatures(unsigned int subwindowSize, unsigned int xOffset, unsigned int yOffset, unsigned int targetSize, FeatureVector* featureVec);
  void extractWaveletCoefficients(unsigned int sizeX, unsigned int sizeY, unsigned int posX, unsigned int posY, int level, FeatureVector* featureVec);
};


#endif /*FEATURE_EXTRACTOR_H*/

