/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    \file FeatureVector.cpp

    $Author: douze $ ( <frank.moosmann@inrialpes.fr>)
    $Date: 2007/10/30 08:04:04 $
*///////////////////////////////////////////////////////////////////////////
 

#include "FeatureVector.h"
#include "ClassLabelHistogram.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>

using namespace std;


FeatureVector::FeatureVector() {
  pyrlevel=-1;
  flabel=gtlabel=-1;
};
FeatureVector::~FeatureVector() {
  
};
  
void FeatureVector::addElement(FeatureData elem) {data.push_back(elem); };
 FeatureData FeatureVector::getElement(int index) {return data[index];};
        void FeatureVector::setElement(int index, FeatureData elem) {data[index] = elem;};
unsigned int FeatureVector::getSize() {return data.size();};
//       void  FeatureVector::setElement(int index, FeatureData elem) {data[index] = elem;};
        
/*!
  \brief Method to compute the empirical mean

  This method computes the empirical mean of a specific attribute 
  of the feature vectors.
  \param featVecList the list of FeatureVectors
  \param attrNo the attribute number
  \return the empirical mean
*/
FeatureData FeatureVector::compEmpMean (list<FeatureVector*> &featVecList, int attrNo)
{
  FeatureData sum = 0;
  FeatureData count = 0;
  FeatureVector* csample;
  for(list<FeatureVector*>::const_iterator p=featVecList.begin(); p!=featVecList.end(); p++)
  {
    csample = *p;
    FeatureData fVal = csample->getElement(attrNo);
    sum += fVal;
    count++;
  }
  //cout << "  attrNo: " << attrNo << "  sum: " << sum << "  count: " << count;
  if (count > 0)
    return sum/count;
  else
    return 0.0;
}


/*!
  \brief Method to compute the standard deviation

  This method computes the standard deviation of a specific attribute 
  of the feature vectors.
  \param featVecList the list of FeatureVectors
  \param attrNo the attribute number
  \param mean the empirical mean
  \return the standard deviation
*/
FeatureData FeatureVector::compStdDeviation (list<FeatureVector*> &featVecList, int attrNo, FeatureData mean)
{
  FeatureData sum = 0.0;
  FeatureData count = 0;
  FeatureVector* csample;
  for(list<FeatureVector*>::iterator p=featVecList.begin(); p!=featVecList.end(); ++p)
  {
    csample = *p;
    FeatureData fVal = csample->getElement(attrNo);
    FeatureData difference = abs(fVal - mean);
    sum += difference;
    count++;
  }
  if (count > 0)
    return sum/count;
  else
    return 0.0;
}


void FeatureVector::writeIntoStream(list<FeatureVector*> &featureList, ostream &data, ostream &info, int maxcount)
{
	int count = 0;
  for(list<FeatureVector*>::const_iterator q=featureList.begin(); (q!=featureList.end()); ++q)
  {
    FeatureVector *featI = *q;
    //featurelist << featI->windowPosX << " " << featI->windowPosY << " " << featI->windowSize << " 0 ";
    for(unsigned int i = 0; i<featI->getSize(); i++)
      data << featI->getElement(i) << " ";
    data << endl;
    info << featI->classLabel << " ";
    // info << featI->windowToObjectRatio << " ";
    info << 0 << " ";
    info << featI->windowPosX << " ";
    info << featI->windowPosY << " ";
    info << featI->windowSize << " ";
    // info << featI->imageWidth << " ";
    // info << featI->imageHeight << " ";
    info << 0 << " ";
    info << 0 << " ";
    info << endl;
    //cout << "." << flush;
		count++;
		if ((maxcount > 0) && (count >= maxcount))
			break;
  }
}

int FeatureVector::readFromStream(istream &data, istream &info, list<FeatureVector*> &featureList, int noToRead)
{
  string sdata;
  string sinfo;
  int noRead = 0;
  //cout << endl << "reading...";
  while ( ((noRead < noToRead) || (noToRead == 0)) && (!data.eof()) && (!info.eof())) {
    getline(data,sdata);
    getline(info,sinfo);
    FeatureVector *fv = new FeatureVector();
    
    istringstream ssdata(sdata);
    FeatureData value;
    while (!ssdata.eof()) {
      ssdata >> value;
      if (ssdata.fail())
        break;
      else
        fv->addElement(value);
    }
    unsigned int dummy;
    double ddummy;
    if (sscanf(sinfo.c_str(), "%d %lf %u %u %u %u %u ",
               &(fv->classLabel),
               &ddummy, // &(fv->windowToObjectRatio),
               &(fv->windowPosX),
               &(fv->windowPosY),
               &(fv->windowSize),
               &dummy, // &(fv->imageWidth),
               &dummy) //&(fv->imageHeight))
        >= 1) {  //at least classLabel must have been read successfully
      featureList.push_back(fv);
      noRead++;
    } else {
      delete fv;
    }
  }
  //cout << noRead;
  return noRead;
}

