/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    \file FeatureVector.h

    $Author: douze $ ( <frank.moosmann@inrialpes.fr>)
    $Date: 2007/10/30 08:04:04 $
*///////////////////////////////////////////////////////////////////////////

 
#ifndef FEATURE_VECTOR_H
#define FEATURE_VECTOR_H

#include <vector>
#include <list>
#include <ostream>
#include <istream>
#include "ClassLabel.h"
#include "ClassLabelHistogram.h"

typedef float FeatureData;
#define FEATURE_NULL 0.0;



class FeatureVector {
  public:
    FeatureVector();
    ~FeatureVector();
  
    void addElement(FeatureData elem);
    FeatureData getElement(int index); //!< index in range 0..getSize()-1
    unsigned int getSize();
    void setElement(int index, FeatureData elem); //!< index in range 0..getSize()-1

/*
    std::string imageName;
    unsigned int imageWidth;    //!< in pixel
    unsigned int imageHeight;   //!< in pixel
    */
    unsigned int windowPosX;    //!< in pixel
    unsigned int windowPosY;    //!< in pixel
    unsigned int windowSize;    //!< scale in pixel
    /*
    double windowToObjectRatio; //!< with respect to area
    */
    // ClassLabelHistogram* classLabelHist;
    BinClassLabelHistogram classLabelHist;
    ClassLabel classLabel;      //! dominant label

    float flabel;               //! 1=texture, 0=background + shades in-between
    float gtlabel;              //! ground truth label
    short pyrlevel;             // pyramid level
    
    static FeatureData compEmpMean (std::list<FeatureVector*> &featVecList, int attrNo);
    static FeatureData compStdDeviation (std::list<FeatureVector*> &featVecList, int attrNo, FeatureData mean);
    
    static void writeIntoStream(std::list<FeatureVector*> &featureList, std::ostream &data, std::ostream &info, int maxcount = 0);
    static int  readFromStream(std::istream &data, std::istream &info, std::list<FeatureVector*> &featureList, int noToRead = 0);
    
  private:
    std::vector<FeatureData> data;

};



#endif /*FEATURE_VECTOR_H*/
