/***************************************************************************
                            SvmClassifier.h
                             ------------------
    begin                : 14.01.2006
    copyright            : (C) 2006 by Frank Moosmann
    email                : frank.moosmann@inrialpes.fr
 ***************************************************************************/

#ifndef SVMCLASSIFIER_H
#define SVMCLASSIFIER_H

#include "Classifier.h"
#include "libsvm/svm.h"
#include "FeatureVector.h"
#include "ClassLabel.h"
#include "ClassLabelHistogram.h"
#include <list>

class SvmClassifier : public virtual Classifier {

public:
  SvmClassifier(std::list<FeatureVector*> *featVecList, svm_parameter *params);
  SvmClassifier(struct svm_problem *problem, ClassLabelHistogram *classHistogram, svm_parameter *params);
  SvmClassifier(std::string filename);
  virtual ~SvmClassifier();
  virtual void writeToFile(std::string filename);

  virtual ClassLabel classify (std::list<FeatureVector*> *featureVec, ClassLabelHistogram *totalVotes = 0, ClassLabelHistogram *windowVotes = 0, bool saveAllClassHistograms=false);
  void getWeights(double** w, int dim);

private:

  struct svm_model *svm;     // pointer to support vector machine
  struct svm_problem prob;   // contains pointer to training data etc. Not possible to free memory if svm_model produced by svm_train() still in use !!!

  void convertFeatureVectors (std::list<FeatureVector*>* featureVec);
  void convertFeatureVector (FeatureVector* fVec, struct svm_node *x);

};
#endif //SVMCLASSIFIER_H
