/***************************************************************************
                            TerminalNode.cpp
                             ------------------
    begin                : 07.11.2005
    copyright            : (C) 2005 by Frank Moosmann
    email                : frank.moosmann@inrialpes.fr
 ***************************************************************************/

#include "TerminalNode.h"
#include "FMUtils.h"
#include <iostream>

unsigned int TerminalNode::totalTerminalNodeCount = 0;
unsigned int TerminalNode::idCounter = 0;
ClassLabelHistogram* TerminalNode::classDistribution = 0;

using namespace std;

//**************************************************************************
//                       Constructors / Destructors
//**************************************************************************
/*!
  \brief Method to construct a decision node

  \param ClassLabel the desired class label of the node
  \param treeLevel the current tree level, used for pruning purposes
  \param featureVecList list of features associated with this node. windowToObjectRatio is averaged from that
*/
TerminalNode::TerminalNode(ClassLabel nodeClassLabel, unsigned int treeLevel, list<FeatureVector*> *featureVecList) {
  if (totalTerminalNodeCount == 0)
    TerminalNode::classDistribution = new ClassLabelHistogram();
  parent = 0;
  classLabel = nodeClassLabel;
  classDistribution->increaseCountOfLabel(classLabel);
  score = 1.0;
  weight = 1.0;
  level = treeLevel;
  windowPosXRatio = 0.0;
  windowPosYRatio = 0.0;
  windowSizeRatio = 0.0;
  windowToObjectRatio = 0.0;
  unsigned int fcount = 0;
  if (featureVecList != 0) {
    for(list<FeatureVector*>::iterator p=featureVecList->begin(); p!=featureVecList->end(); p++) {
      FeatureVector* currFV = *p;
      FeatureDetails fd;
      fd.posX = currFV->windowPosX;
      fd.posY = currFV->windowPosY;
      fd.size = currFV->windowSize;
      fd.label = currFV->classLabel;
      trainingFeatures.push_back(fd);
      if (currFV->classLabel == nodeClassLabel) {
        fcount++;
        /*
        windowPosXRatio += (double)(currFV->windowPosX) / (double)(currFV->imageWidth);
        windowPosYRatio += (double)(currFV->windowPosY) / (double)(currFV->imageHeight);
        windowSizeRatio += currFV->windowPosY / (double)(min(currFV->imageWidth, currFV->imageHeight));
        
        windowToObjectRatio += currFV->windowToObjectRatio;
        */
      }
    }
    windowPosXRatio /= fcount;
    windowPosYRatio /= fcount;
    windowSizeRatio /= fcount;
    windowToObjectRatio /= fcount;
  }
  idCounter++;
  id = idCounter;
  totalTerminalNodeCount++;
  //if (classLabel == 0)
  //  std::cout << "Create Terminal Node with class label " << classLabel << std::endl;
}

TerminalNode::TerminalNode(unsigned int id, unsigned int level, double score, ClassLabel classLabel, double weight, double windowPosXRatio, double windowPosYRatio, double windowSizeRatio, double windowToObjectRatio) {
  this->id = id;
  this->parent = 0;
  this->level = level;
  this->score = score;
  this->classLabel = classLabel;
  this->weight = weight;
  this->windowPosXRatio = windowPosXRatio;
  this->windowPosYRatio = windowPosYRatio;
  this->windowSizeRatio = windowSizeRatio;
  this->windowToObjectRatio = windowToObjectRatio;
  if (totalTerminalNodeCount == 0)
    TerminalNode::classDistribution = new ClassLabelHistogram();
  classDistribution->increaseCountOfLabel(classLabel);
  idCounter = max(idCounter, id);
  totalTerminalNodeCount++;
}

TerminalNode::~TerminalNode ( ) {
  totalTerminalNodeCount--;
  if (totalTerminalNodeCount == 0) {
    delete classDistribution;
    classDistribution = 0;
  }
}


//**************************************************************************
//                              public Methods
//**************************************************************************
TerminalNode* TerminalNode::classify (FeatureVector* featureVec, bool addFVecInfo) {
  if (addFVecInfo) {
    FeatureDetails fd;
    fd.posX = featureVec->windowPosX;
    fd.posY = featureVec->windowPosY;
    fd.size = featureVec->windowSize;
    fd.label = featureVec->classLabel;
    testingFeatures.push_back(fd);
  }
  return this;
}

void TerminalNode::writeToFile(ostream *out) {
  //unsigned int id;  private
  //TreeNode *parent;
  //unsigned int level;
  //double score;
  //ClassLabel classLabel;
  //double weight;              //!< in range 0..1 states the importance of this node for classification (default 1)
  //double windowPosXRatio;     //!< with respect to im_width
  //double windowPosYRatio;     //!< with respect to im_height
  //double windowSizeRatio;     //!< with respect to min(im_height,im_width)
  //double windowToObjectRatio; //!< with respect to area
  //static unsigned int totalTerminalNodeCount;
  //static ClassLabelHistogram* classDistribution;
  //static unsigned int idCounter;  private 
  (*out) << "TerminalNode;" << id << ";"
                            << level << ";"
                            << score << ";"
                            << classLabel << ";"
                            << weight << ";"
                            << windowPosXRatio << ";"
                            << windowPosYRatio << ";"
                            << windowSizeRatio << ";"
                            << windowToObjectRatio << ";"
                            << endl;
}

