/***************************************************************************
                            TerminalNode.h
                             ------------------
    begin                : 07.11.2005
    copyright            : (C) 2005 by Frank Moosmann
    email                : frank.moosmann@inrialpes.fr
 ***************************************************************************/
 
#ifndef TERMINALNODE_H
#define TERMINALNODE_H

#include "TreeNode.h"
#include "ClassLabelHistogram.h"
#include <list>
struct FeatureDetails { 
  std::string imageName;
  unsigned int posX;
  unsigned int posY;
  unsigned int size;
  ClassLabel label;
};

class TerminalNode : public virtual TreeNode {

public:
  TerminalNode(ClassLabel nodeClassLabel, unsigned int treeLevel, std::list<FeatureVector*> *featureVecList = 0);
  TerminalNode(unsigned int id, unsigned int level, double score, ClassLabel classLabel, double weight, double windowPosXRatio, double windowPosYRatio, double windowSizeRatio, double windowToObjectRatio);
  virtual ~TerminalNode();

  virtual TerminalNode* classify (FeatureVector* featureVec, bool addFVecInfo);
  void virtual writeToFile(std::ostream *out);

  ClassLabel classLabel;
  double weight;              //!< in range 0..1 states the importance of this node for classification (default 1)
  unsigned int getID() {return id;};
  void setID(unsigned int newID) {id=newID;};
  double windowPosXRatio;     //!< with respect to im_width
  double windowPosYRatio;     //!< with respect to im_height
  double windowSizeRatio;     //!< with respect to min(im_height,im_width)
  double windowToObjectRatio; //!< with respect to area
  std::list<FeatureDetails> trainingFeatures;
  std::list<FeatureDetails> testingFeatures;
  
  static unsigned int getMaxID() {return idCounter;};
  static unsigned int totalTerminalNodeCount;
  static ClassLabelHistogram* classDistribution;
  
private:
  
  unsigned int id;
  
  static unsigned int idCounter;
  
};
#endif //TERMINALNODE_H
