
import os,sys,math
import gtk  # for image i/o

from TextureClassifier import *


FeaturesContainer.ref_init=FeaturesContainer.__init__
def fcinit(self,*args):  
  if len(args)==1 and type(args[0])==type([]):
    self.ref_init()
    self.addFromList(args[0])
  else:
    self.ref_init(*args)

FeaturesContainer.__init__=fcinit


FeatureVector.gt_tex=lambda self:self.gtlabel>0.5

FeatureVector.f_tex=lambda self:self.flabel>0.5

def describe_PatchParams(self):
  pp=self
  return "SIFT scale one=%d detector:%s %s (aff,ang,haar,hue)=(%d,%d,%d,%d)"%(
    pp.scone,
    (pp.dets & FeaturesComputer.LOG) and "LoG" or "",
    (pp.dets & FeaturesComputer.HL) and "H-L" or "",
    pp.doAffNorm,pp.doAngNorm,pp.doHaar,
    pp.addhue)  

FeaturesComputer.__str__=describe_PatchParams



# retrieved=[(v1,ok1),...,(v_n,ok_n)]
# where v_i is a value and ok_i is the ground truth about it
# v1 < v2 < ... < vn
def make_pr(retrieved,totpos=-1,plotf=None):

  if totpos==-1:
    totpos=0
    for (score,pos) in retrieved:
      if pos:
        totpos+=1

  nretrieved=0
  npos=0

  prev_pr=(0.0,1.0)

  auc=0.0

  for (score,pos) in retrieved:
    nretrieved+=1
    if pos: npos+=1
    precision=npos/float(nretrieved)
    recall=npos/float(totpos)
    if plotf:
      plotf.write("%g %g\n"%(recall,precision))

    (prev_recall,prev_precision)=prev_pr
    auc+=(recall-prev_recall)*.5*(precision+prev_precision)
    prev_pr=(recall,precision)
  return auc

def compute_features(fc,imname,maskname):
  print "compute_features on %s (mask %s)"%(imname,maskname)
  print "  loading"
  # im=PhotoMole.RGBImage(imname)
  im=gtk.gdk.pixbuf_new_from_file(imname)
  if maskname:
    # mask=PhotoMole.GrayImage(maskname)
    mim=gtk.gdk.pixbuf_new_from_file(maskname)
    mask=colorImageToMaskImage(mim); mask.thisown=True;    
  else:
    mask=None
  if maskname:
    imsize=(im.get_width(),im.get_height())
    masksize=(mask.xd,mask.yd)
    if imsize==masksize:
      pass
    elif imsize==(masksize[1],masksize[1]):
      print "  transpose ? checking exif info"
      raise "not implemented"
    else:
      raise "size mismatch"  
  print "    size %d*%d"%(im.get_width(),im.get_height())
  print "  preprocess"
  pi=PreprocessedImage(im,mask)  
  print "  detect"
  pl=fc.computeFeatures(pi)
  print "    %d features"%len(pl)
  return (pi,pl)


def extractLearningSet(pll):
  if type(pll)!=type([]):
    pll=pll.toList()

  posex=[patch for patch in pll if patch.gtlabel==1.0]
  negex=[patch for patch in pll if patch.gtlabel==0.0]
  
  for patch in posex:
    patch.classLabel=LABEL_FG
  for patch in negex:
    patch.classLabel=LABEL_BG

  return posex+negex

# learn_set = (im_1,...,im_n)]
# where im_1=imname | (imname,maskname)
# where maskname=None (whole image is positive)
# or "!" (whole image is neg)
# or a mask file name

def learn_from_set(pp,ep,learn_set,outdir=None):
  
  print "="*70
  print "output dir ",outdir
  print "learn from", learn_set
  sys.stdout.flush()
           
  print pp

  fc=pp

  #########################
  # learn

  allpatches=[]
  
  for i in range(len(learn_set)):
    print "learn image %d "%i

    if type(learn_set[i])==type(""):
      imname=learn_set[i]
      maskname=None
    else:
      (imname,maskname)=learn_set[i]      
      
    (pi,pl)=compute_features(fc,imname,maskname!='!' and maskname or None)
    
    if not maskname:
      for patch in pl.toList():
        patch.gtlabel=1.0
    elif maskname=='!':
      for patch in pl.toList():
        patch.gtlabel=0.0
    else:
      pl.setGroundTruth(pi)

    allpatches.append(pl)

    if outdir:
      f=open(outdir+"/learn_%02d.features"%i,"w")
      pl.writeTo(f)
      
    sys.stdout.flush()
    
  pll=FeaturesContainer()
  npos=nneg=0
  for pl in allpatches:
    pl=extractLearningSet(pl)
    npos+=len([patch for patch in pl if patch.classLabel==LABEL_FG])
    nneg+=len(pl)-nneg             
    # print "add %dp+%dn=%d features"%(npos,nneg,len(pl))
    pll.addFromList(pl)
  print "total %dp+%dn=%d features"%(npos,nneg,len(pll))
  sys.stdout.flush()

  cl=TextureClassifier(pll,ep)
  print
  
  if outdir:
    cl.writeToFile(outdir+"/interest_learn.frimcla")
 
  # dealloc
  for pl in allpatches:
    pl.clear()

  return cl


def test_on_image(fc,cl,imname,maskname,outdir=None,key=''):
  strong_lim=0.2

  (pi,pl)=compute_features(fc,imname,maskname)

  if maskname:
    print "GT"
    pl.setGroundTruth(pi)

  if outdir:
    f=open(outdir+"/test%s.features"%key,"w")
    pl.writeTo(f)

  print "classifying"
  sys.stdout.flush()

  cl.classify(pl)

  pll=pl.toList()

  if outdir:  
    gp=open(outdir+"/gt_vs_f%s.dat"%key,"w")
    for patch in pll:
      gp.write("%g %g\n"%(patch.flabel,patch.gtlabel))
    gp.close()

  pll_classified=[patch for patch in pll if patch.flabel!=-1]

  pll_strong=[patch for patch in pll_classified if abs(patch.flabel-0.5)>strong_lim]
  print "nb strong patches: %dp+%dn=%d/%d"%(
    len([patch for patch in pll_classified if patch.f_tex()]),
    len([patch for patch in pll_classified if not patch.f_tex()]),    
    len(pll_strong),len(pll))
  sys.stdout.flush()

  if maskname:  
    retrieved=[(patch.flabel,patch.gt_tex()==patch.f_tex()) for patch in
              pll_strong]
    auc=make_pr(retrieved)
    print "auc=%.4f"%auc


  f_prop=reduce(lambda accu,patch:accu+patch.flabel,pll_classified,0.0)/len(pll_classified)

  if maskname:
    gt_prop=pi.getTexProportion()  
    print "tex proportion gt=%.4f found: %.4f"%(gt_prop,f_prop)
  else:
    print "tex proportion gt=?? found: %.4f"%f_prop

  if outdir:
    (xd,yd)=(pi.getxsize(),pi.getysize())
    if xd*yd>800*600:
      scale=math.sqrt(800*600/float(xd*yd))
      (xd,yd)=(int(xd*scale),int(yd*scale))
    # im=PhotoMole.RGBImage(xd,yd)
    im=gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB,False,8,xd,yd)
    FeaturesContainer(pll_strong).showVotes(scale,im)
    im.save(outdir+"/found_strong%s.png"%key,"png")      

  pl.clear()

  return f_prop



# test_set = (im_1,...,im_n)]
# where im_1=imname | (imname,maskname)
# where maskname=None (mask unknown)
# or a mask file name

def test_with_set(pp,cl,test_set,outdir=None):

  fc=InterestFeaturesComputer(pp)

  ################################
  # test
  
  print "="*70
  print "output dir ",outdir
  print "test on", test_set
  sys.stdout.flush()

  res=[]

  for i in range(len(test_set)):
    
    print "test image %d"%i

    if type(test_set[i])==type(""):
      imname=test_set[i]
      maskname=None
    else:
      (imname,maskname)=test_set[i]      

    score=test_on_image(fc,cl,imname,maskname,outdir,"_%02d"%i)
    res.append((imname,score))
  return res

  
def mkdir_or_clear(dirpref):
  try:
    os.mkdir(dirpref)
  except OSError,e:
    if e.errno==17:
      print "clearing "+dirpref
      os.system("rm -rf "+dirpref+"/*")
    else:
      raise e

def get_files_in(dirname):
  return [dirname+'/'+f for f in os.listdir(dirname)]


def getline(f):
  l=f.readline()
  # print l
  if not l: return None
  i=l.find('#')
  if i>=0: l=l[:i]
  l=l.strip()
  if l=='': return getline(f)
  return l


def load_index_dat(fname):
  basename=fname[:fname.rfind("/")+1]
  f=open(fname,"r")
  if getline(f)!="texture classifier 102":
    raise "wrong header"
  tcsets=[]
  while True:
    l=getline(f)
    if not l: break
    if l!="BeginSet":
      raise "parse error 5"
    pp=FeaturesComputer()
    pp.readFrom(f)
    tcs=[]
    while True:
      l=getline(f)
      if not l: break
      if l=="EndSet":
        break
      try:
        [name,fname]=l.split()
      except ValueError:
        raise "parse error 2"
      fname=basename+fname
      print "loading class %s from %s"%(name,fname)      
      if os.access(fname,os.R_OK):
        tc=TextureClassifier(fname)
      else:
        print "  could not load"
        tc=None
      tcs.append((name,tc))
    tcsets.append((pp,tcs))
  return tcsets


  

