#include <cstdio>
#include <cerrno>
#include <cstring>
#include <cassert>


#include "org_acemedia_am_textureclassifier_TextureClassifierImpl.h"

#include "TextureClassifier.h"

using namespace lava_ns;
using namespace std;


struct TextureClassifierSet {

  struct NamedTC {
    string name;
    TextureClassifier *tc;
  };

  FeaturesComputer pp;
  vector<NamedTC> tcs;
  
};

static vector<TextureClassifierSet> tcSets;



static char *strip(char *b) {
  for(;*b && isspace(*b);b++);
  char *c;
  for(c=b;*c;c++); // find eos
  for(c--;c>=b && isspace(*c);c--) *c=0;
  return b;
}

static char* getline(char *buf,int sz,FILE *f) {
  if(!fgets(buf,sz,f)) return NULL;
  // printf("getline %s\n",buf);

  { // remove comment introduced by "#"
    char *c;
    for(c=buf;*c && *c!='#';c++);
    if(*c=='#') *c=0;
  }
  { // strip begin & end spaces
    char *begin=strip(buf);
    if(*begin==0) // empty line
      return getline(buf,sz,f);
    else 
      return begin;
  }
}

JNIEXPORT jint JNICALL Java_org_acemedia_am_textureclassifier_TextureClassifierImpl_loadClassifier
(JNIEnv *env, jclass, jstring sfname)
{
  const char *fname;
  
  fname=env->GetStringUTFChars(sfname,NULL);
  
  char basename[500];
  {
    strcpy(basename,fname);
    char *c;
    for(c=basename+strlen(basename)-1;c>=basename;c--) 
      if(*c=='/') break;
    if(*c=='/') {
      *c=0;
    } else {
      basename[0]='.';
      basename[1]=0;
    } 
  }  

  {
    FILE *f=fopen(fname,"r");
    if(!f) return 3;

    char buf[1024],*l;
    l=getline(buf,1024,f);
    int version;
    if(sscanf(l,"texture classifier %d",&version)!=1 ||
       version!=102) {      
      return 1;
    }

    for(;;) {

      l=getline(buf,1024,f);
      
      if(!l) 
        break;
      if(strcmp(l,"BeginSet")) {
        return 5;
      }
      TextureClassifierSet tcSet;
            
      tcSet.pp.readFrom(f);

      for(;;) {
        char name[512];
        char fname[512];
        l=getline(buf,1024,f);
        if(!l) break;
        
        if(!l || !strcmp(l,"EndSet")) {
          break;
        }      

        if(sscanf(l," %512s %512s ",name,fname)!=2) {
          return 2;
        }
        
        // load classifier from file
        
        int m=strlen(basename);
        basename[m]='/';
        strcpy(basename+m+1,fname);
        TextureClassifierSet::NamedTC ntc;
        printf("class %s from %s\n",name,basename);
        ntc.name=name;
      
        if(access(basename,R_OK)==0) {
          ntc.tc=new TextureClassifier(basename);          
        } else {
          printf("cannot access %s!!\n",basename);
          ntc.tc=NULL;
        }
        tcSet.tcs.push_back(ntc);
        basename[m]=0;

      }

      tcSets.push_back(tcSet);
    }    
    fclose(f); 
  }

  env->ReleaseStringUTFChars(sfname,fname);
  
  return 0;
}


JNIEXPORT void JNICALL Java_org_acemedia_am_textureclassifier_TextureClassifierImpl_unloadClassifier
(JNIEnv *, jclass) {

  for(int j=0;j<tcSets.size();j++) {
    TextureClassifierSet &tcSet=tcSets[j];
    for(int i=0;i<tcSet.tcs.size();i++) 
      delete tcSet.tcs[i].tc;
  }
  tcSets.clear();

}

static PreprocessedImage* dataToPreprocessedImage(JNIEnv *env, jintArray data, jint xd, jint yd) {
  
  if(env->GetArrayLength(data)!=xd*yd) {
    fprintf(stderr,"TextureClassifier dataToRGBImage:"
            " data size = %d != xd*yd = %d*%d\n",
            env->GetArrayLength(data),xd,yd);
    return NULL;
  }  

  jboolean isCopy;
  jint *dt=env->GetIntArrayElements(data,&isCopy);
  // it *is* a copy ! why ?

  // convert input to RGBImage
  // pixel layout: 0xFFRRGGBB

  ColorImage cim;
  cim.xd=xd; cim.yd=yd; cim.bpl=3*xd;
  cim.data=new unsigned char[xd*yd];

  int j=0;
  for(int i=0;i<xd*yd;i++) {
    jint v=dt[i];
    cim.data[j++]=v>>16;
    cim.data[j++]=v>>8;
    cim.data[j++]=v;
  }

  env->ReleaseIntArrayElements(data,dt,JNI_ABORT);

  PreprocessedImage *pi=new PreprocessedImage(cim);
  
  delete [] cim.data;
  
  return pi;
}


JNIEXPORT jstring JNICALL Java_org_acemedia_am_textureclassifier_TextureClassifierImpl_detectTexture
(JNIEnv *env, jobject zhis, jintArray data, jint xd, jint yd) {

  PreprocessedImage *pi=dataToPreprocessedImage(env,data,xd,yd);

  if(!pi) 
    return env->NewStringUTF("could not convert data to image");  
  

  jmethodID mid=env->GetMethodID(
          env->GetObjectClass(zhis),
          "setTextureInfo","(Ljava/lang/String;D)I");
  assert(mid);

  { // detect all textures

    for(int j=0;j<tcSets.size();j++) {
      TextureClassifierSet &tcSet=tcSets[j];
   
      FeaturesContainer *pl=tcSet.pp.computeFeatures(*pi);
    
      for(int i=0;i<tcSet.tcs.size();i++) {
        
        TextureClassifierSet::NamedTC &ntc=tcSet.tcs[i];
      
        if(ntc.tc==NULL) continue;
      
        ntc.tc->classify(pl);

        double f_prop=0;
        int n=0;
        for(FeaturesContainer::const_iterator it=pl->begin();
            it!=pl->end();++it) if((**it).flabel>=0) {
          f_prop+=(**it).flabel;
          n++;
        }
        f_prop/=n;
        env->CallIntMethod(zhis,mid,
                           env->NewStringUTF(ntc.name.c_str()),f_prop);
        
      }
      pl->clear();
      delete pl;
    }
  }

/*
  for(int i=0;i<10;i++) {
    char buf[50];
    jint k=dt[i+i*xd];
    jdouble f=i/double(10);
    sprintf(buf,"type %d",i);
    jint no=env->CallIntMethod(
           zhis,mid,
           env->NewStringUTF(buf),f);
    
    //  printf("no=%d\n",no);
    
  }
*/
  delete pi;

  return env->NewStringUTF("OK");  
}





///////////////////////////////////////////////////////////////////////
// IC classifier


static struct {
  TextureClassifier *tc;
  FeaturesComputer pp;
  int nneg,npos;
  int verbose;
} ic={0};


#define CE if (env->ExceptionCheck ()) {                        \
  fprintf(stderr,"%s:%d\n",__FILE__,__LINE__);         \
  env->ExceptionDescribe ();                                    \
  exit(1);                                                      \
}

static int getVerbose(JNIEnv *env, jclass clazz) {
  jfieldID verbose_field=env->GetStaticFieldID(clazz,"icVerbose","I");
  CE;
  ic.verbose=env->GetStaticIntField(clazz,verbose_field);  
}


#define SV getVerbose(env,clazz)

JNIEXPORT jstring JNICALL Java_org_acemedia_am_textureclassifier_TextureClassifierImpl_loadICParams
  (JNIEnv *env, jclass clazz, jstring filename) {

  SV;
  
  const char *fname=env->GetStringUTFChars(filename,NULL);
  
  FILE *f=fopen(fname,"r");
  if(!f) {
    char err[200];
    sprintf(err,"TextureClassifierImpl.loadICParams could not open %s: %s",fname,strerror(errno));
    return env->NewStringUTF(err);    
  }
  ic.pp.readFrom(f);
  fclose(f);  

  ic.tc=0;

  env->ReleaseStringUTFChars(filename,fname);

  return env->NewStringUTF("OK");    
}


JNIEXPORT jstring JNICALL Java_org_acemedia_am_textureclassifier_TextureClassifierImpl_compileForIC
(JNIEnv *env, jclass clazz, jintArray data, jint xd, jint yd, jstring filename) {
  SV;
  PreprocessedImage *pi=dataToPreprocessedImage(env,data,xd,yd);

  if(!pi) 
    return env->NewStringUTF("could not convert data to image");  
  
  FeaturesContainer *pl=ic.pp.computeFeatures(*pi);

  if(pl==NULL) {
    delete pi;
    return env->NewStringUTF("image too small");    
  }

  const char *fname=env->GetStringUTFChars(filename,NULL);
  FILE *f=fopen(fname,"w");
  if(!f) {
    delete pi;
    char err[200];
    sprintf(err,"could not open %s: %s",fname,strerror(errno));
    return env->NewStringUTF(err);    
  }
  fprintf(f,"%d %d ",xd,yd);
  pl->writeTo(f);
  fclose(f);
  
  if(ic.verbose) 
    printf("stored patches from %d*%d image in %s\n",xd,yd,fname);
    
  env->ReleaseStringUTFChars(filename,fname);
  delete pi;
  return env->NewStringUTF("OK");    
}


static void curve_to_double(
    JNIEnv *env,jobjectArray ja,double **pts_out,int *npt_out) {
  int npt=env->GetArrayLength(ja);
  CE;
  *npt_out=npt;
  
  double *pts=new double[2*npt];
  *pts_out=pts;

  jfieldID x_field,y_field;
  
  for(int i=0;i<npt;i++) {
    jobject jo=env->GetObjectArrayElement(ja,i);
    CE;
    if(i==0) {
      jclass cl=env->GetObjectClass(jo);
      x_field=env->GetFieldID(cl,"x","I");
      y_field=env->GetFieldID(cl,"y","I");  
      CE;
    }
    
    pts[i*2  ]=env->GetIntField(jo,x_field);
    pts[i*2+1]=env->GetIntField(jo,y_field);

  }
  
}


struct ImWithMask {
  
  enum Combination {
    ZERO,
    POS_AND_NEG,
    ALLPOS,
    ALLNEG,
    POS_AND_INV,
    NEG_AND_INV
  };
  
  int no;
  int xd,yd;  
  MaskImage *pos_mask,*neg_mask;
  FeaturesContainer *pl;
  
  Combination comb;


  ImWithMask():no(-1),pos_mask(0),neg_mask(0),comb(ZERO),pl(0) {
  }
  
  ~ImWithMask() {
    delete pos_mask;
    delete neg_mask;
    if(pl) {
      pl->clear();
      delete pl;
    }
  }

  const char* initFeaturesContainer(const char *fname) {
    FILE *f=fopen(fname,"r");
    if(!f) {
      return "could not open features file";  
    }
    
    pl=new FeaturesContainer();
    
    char c;
    if(fscanf(f,"%d %d%c",&xd,&yd,&c)!=3 || c!=' ')
      return "bad header in features file";
    pl->readFrom(f);
    fclose(f);

    return NULL;
  }

  // cf. VectorMask.get_as_image in Python version
  void initMaskImage(bool init_pos,JNIEnv *env,jobjectArray curves) {

    MaskImage * &gi=init_pos ? pos_mask : neg_mask;

    gi=new MaskImage(xd,yd);
    gi->clear();

    for(int j=0;j<env->GetArrayLength(curves);j++) {
      double *pts;
      int npt;
      jobjectArray ja2=(jobjectArray)env->GetObjectArrayElement(curves,j);
      CE;          
      curve_to_double(env,ja2,&pts,&npt);
      
      fillPolygon(*gi,255,pts,npt);

      free(pts);
    }

    if(ic.verbose>5) {
      char fname[512];
      sprintf(fname,"/tmp/mask_%d_%s.pgm",no,init_pos ? "pos" : "neg");
      MonoImage<unsigned char> gmask(xd,yd);
      memcpy(gmask[0],gi->data,xd*yd);      
      gmask.store(fname);
    }

  }
  
  // cf. TextureManager.learn in Python version
  void extractLearningPatches(FeaturesContainer *pll) {
    switch(comb) {
    case ALLPOS: case ALLNEG: {
      int label=comb==ALLPOS ? LABEL_FG : LABEL_BG;
      for(FeaturesContainer::iterator it=pl->begin();it!=pl->end();++it) {
        (**it).classLabel=label;
        pll->push_back(*it);
      }
    } break;
    case POS_AND_INV: case NEG_AND_INV: {
      int label,otherlabel;
      MaskImage *mask;
      if(comb==POS_AND_INV) {
        label=LABEL_FG;
        otherlabel=LABEL_BG;
        mask=pos_mask;
      } else {
        label=LABEL_BG;
        otherlabel=LABEL_FG;
        mask=neg_mask;
      }
      pl->setGroundTruth(*mask);
      for(FeaturesContainer::iterator it=pl->begin();it!=pl->end();++it) {
        if((**it).gtlabel==1.0) 
          (**it).classLabel=label;
        else
          (**it).classLabel=otherlabel;
        pll->push_back(*it);
      }
    } break;
    case POS_AND_NEG: {
      int label=LABEL_FG;
      MaskImage*mask=pos_mask;
      for(int ll=0;ll<2;ll++,label=LABEL_BG,mask=neg_mask) {
        if(!mask) continue;
        pl->setGroundTruth(*mask);
        for(FeaturesContainer::iterator it=pl->begin();it!=pl->end();++it) 
          if((**it).gtlabel==1.0) {
            (**it).classLabel=label;
            pll->push_back(*it);
          }            
      }      
    } break;
    case ZERO: assert(0);
    }        
  }


};

#define POS 1
#define NEG 2


JNIEXPORT jstring JNICALL Java_org_acemedia_am_textureclassifier_TextureClassifierImpl_learnIC
(JNIEnv *env, jclass clazz, jobjectArray jims) {
  
  SV;  
  int n=env->GetArrayLength(jims);
  CE;
  ImWithMask ims[n];

  jfieldID filename_field;
  jfieldID pos_curve_field;
  jfieldID neg_curve_field;
  jfieldID copied_field;

  FeaturesContainer *pll=new FeaturesContainer();
  
  for(int i=0;i<n;i++) {
    jobject jo=env->GetObjectArrayElement(jims,i);

    /********************
     * demangle java structures 
     */

    if(i==0) {
      jclass cl=env->GetObjectClass(jo);
      filename_field=env->GetFieldID(cl,"patchfile","Ljava/lang/String;");
      CE;
      
      pos_curve_field=env->GetFieldID(
          cl,"pos_curves",
          "[[Lorg/acemedia/am/textureclassifier/Point;");
      CE;
      neg_curve_field=env->GetFieldID(
          cl,"neg_curves",
          "[[Lorg/acemedia/am/textureclassifier/Point;");
      copied_field=env->GetFieldID(cl,"copied","I");
      CE;
  
    }

    jstring filename=(jstring)env->GetObjectField(jo,filename_field);
    CE;
    const char *fname=env->GetStringUTFChars(filename,NULL);
    CE;

    jobjectArray pos_curves=
      (jobjectArray)env->GetObjectField(jo,pos_curve_field);
    CE;
    int n_pos_curve=pos_curves ? env->GetArrayLength(pos_curves) : 0;
    CE;

    jobjectArray neg_curves=
      (jobjectArray)env->GetObjectField(jo,neg_curve_field);
    int n_neg_curve=neg_curves ? env->GetArrayLength(neg_curves) : 0;
    CE;

    int copied=env->GetIntField(jo,copied_field);
    CE;

    /******************
     * end demangle
     */

    if(ic.verbose)
      printf("im %s n_pos_curve:%d n_neg_curve:%d copied=%d\n",
             fname,n_pos_curve,n_neg_curve,copied);  

    const char *err=ims[i].initFeaturesContainer(fname);
    if(err) 
      return env->NewStringUTF(err);

    env->ReleaseStringUTFChars(filename,fname);
 

    ims[i].no=i;

    // cf. to_bm_masks in Python version

    if(copied==0) {
      if(n_pos_curve!=0) 
        ims[i].initMaskImage(true,env,pos_curves);
      if(n_neg_curve!=0) 
        ims[i].initMaskImage(false,env,neg_curves);
      ims[i].comb=ImWithMask::POS_AND_NEG;
    } else if(copied==POS) {
      if(n_pos_curve==0) 
        ims[i].comb=ImWithMask::ALLNEG;
      else {
        ims[i].initMaskImage(true,env,pos_curves);
        ims[i].comb=ImWithMask::POS_AND_INV;
      }
    } else if(copied==NEG) {
      if(n_neg_curve==0) 
        ims[i].comb=ImWithMask::ALLPOS;
      else {
        ims[i].initMaskImage(true,env,neg_curves);
        ims[i].comb=ImWithMask::NEG_AND_INV;
      }      
    }

    if(ic.verbose)
      printf("comb=%s masks=%p,%p\n",
             ims[i].comb==ImWithMask::POS_AND_NEG ? "POS_AND_NEG" :
             ims[i].comb==ImWithMask::ALLPOS ? "ALLPOS" :
             ims[i].comb==ImWithMask::ALLNEG ? "ALLNEG" :
             ims[i].comb==ImWithMask::POS_AND_INV ? "POS_AND_INV" :
             ims[i].comb==ImWithMask::NEG_AND_INV ? "NEG_AND_INV" :
             "invalid",
             ims[i].pos_mask,ims[i].neg_mask);
    
    ims[i].extractLearningPatches(pll);

      

/*
    jobjectArray ja=pos_curves;
    for(int ll=0;ll!=2;ll++,ja=neg_curves) {
      
      printf("%s curves\n",ja==pos_curves ? "pos" : "neg");
      if(ja==NULL) continue;
      for(int j=0;j<env->GetArrayLength(ja);j++) {
        printf("  curve %d: ",i);
        double *pts;
        int npt;
        jobjectArray ja2=(jobjectArray)env->GetObjectArrayElement(ja,j);
        CE;          
        curve_to_double(env,ja2,&pts,&npt);

        for(int k=0;k<npt;k++) 
          printf("(%g,%g) ",pts[2*k],pts[2*k+1]);
        
        printf("\n");
        free(pts);
      }
    }       
*/
  }

  int npos=0,nneg=0;
  for(FeaturesContainer::iterator it=pll->begin();it!=pll->end();++it) 
    if((**it).classLabel==LABEL_FG) npos++;
    else if((**it).classLabel==LABEL_BG) nneg++;
    else assert(0);
  
  ic.nneg=nneg; ic.npos=npos;

  if(ic.verbose)
    printf("learn from %dp+%dn patches\n",npos,nneg);

  ExtraTreeClassifier::ExtraTreeParams ep;
  ep.noOfTrees=15;

  delete ic.tc; 
  ic.tc=new TextureClassifier(pll,ep);

  if(ic.verbose>5) {
    static int count=0;
    char fname[512];

    sprintf(fname,"/tmp/learning_patches_%d.patchlist",count);
    FILE *f=fopen(fname,"w");
    if(!f) {
      fprintf(stderr,"could not open %s for w",fname);
      perror("");
      exit(1);
    }
    pll->writeTo(f);
    fclose(f);

    sprintf(fname,"/tmp/classifier_%d.frimcla",count++);
    ic.tc->writeToFile(fname);

    double fpos=ic.npos/float(ic.nneg+ic.npos);
    double strong_lim_inf,strong_lim_sup;
    strong_lim_inf=0.4*fpos;
    strong_lim_sup=1-0.4*(1-fpos);
    
    int n_classified,n_strong_pos,n_strong_neg;
    double f_prop;
    
    // printf("classifying, strong lims=%g,%g\n",strong_lim_inf,strong_lim_sup);
    
    
    ic.tc->classifyAndStats(pll,strong_lim_inf,strong_lim_sup,
                            n_classified,n_strong_pos,n_strong_neg,f_prop);


    int nfp=0,nfn=0;
    for(FeaturesContainer::iterator it=pll->begin();it!=pll->end();++it) {
      FeatureVector *rf=*it;
      if(rf->classLabel==LABEL_FG) 
        if(rf->flabel!=1.0) 
          nfn++;
      if(rf->classLabel==LABEL_BG) 
        if(rf->flabel!=0.0) 
          nfp++;
    }
    printf("FP,FN: %d,%d\n",nfp,nfn);
 

  }
   
  // individual patches will be deleted by ims[] destructor
  delete pll;
  
  return env->NewStringUTF("OK");
}


JNIEXPORT jdouble JNICALL Java_org_acemedia_am_textureclassifier_TextureClassifierImpl_classifyIC
(JNIEnv *env, jclass clazz, jstring filename) {
  SV;

  const char *fname=env->GetStringUTFChars(filename,NULL);
  CE;

  ImWithMask im;
  const char *err=im.initFeaturesContainer(fname);

  if(err) {
    fprintf(stderr,"learning not done or failed!");
    return -1;
  }

  double strong_lim_inf,strong_lim_sup;
  
  double fpos=ic.npos/float(ic.nneg+ic.npos);
  strong_lim_inf=0.4*fpos;
  strong_lim_sup=1-0.4*(1-fpos);

  int n_classified,n_strong_pos,n_strong_neg;
  double f_prop;

  if(ic.verbose) 
    printf("classifying, strong lims=%g,%g\n",strong_lim_inf,strong_lim_sup);


  ic.tc->classifyAndStats(im.pl,strong_lim_inf,strong_lim_sup,
                          n_classified,n_strong_pos,n_strong_neg,f_prop);

  if(ic.verbose>5) {
    static int count=0;
    lava_ns::CRGBImage voteIm(im.xd,im.yd);
    ColorImage cim;
    cim.xd=voteIm.getxsize();
    cim.yd=voteIm.getysize();
    cim.bpl=voteIm.getxsize()*3;
    cim.data=(unsigned char*)voteIm[0];
    im.pl->showVotes(1.0,cim);
    char fname[512];
    sprintf(fname,"/tmp/show_votes_%d.ppm",count++);
    voteIm.store(fname);
  }

  if(ic.verbose) 
    printf("strong patches: %dp+%dn\n",n_strong_pos,n_strong_neg);
  
  if(n_strong_neg+n_strong_pos<5) {
    if(ic.verbose) printf("too few strong patches\n");
    f_prop=0.0/0.0;
  } else {
    
  }

  env->ReleaseStringUTFChars(filename,fname);
  CE;

  return f_prop;
}
