

import random,os,gc,traceback,cPickle,sys
from math import *

import gobject,gtk,pango

import libexif,texture

from  extra_widgets import *
from extra_functions import *

from image import *
from image_db import *

from organizers import *
from selected_window import *


class DBWidgetCommon(gtk.DrawingArea):
  """
  parent of the 2 DB widget layouts
  self.mainwidget: access to the images and selection
  """
  def __init__(self,mainwidget):
    gtk.DrawingArea.__init__(self)
    self.connect('size-allocate', self.on_size_allocate)
    self.connect('size-request',self.on_size_request)
    self.connect('expose-event',self.on_expose_event)
    self.connect('realize',self.on_realize)
    self.mainwidget=mainwidget
    self.init_interactive()
    self.second_coords=None
    
  def on_realize(self,widget):
    self.gc = widget.window.new_gc()
    pcont=gtk.Widget.create_pango_context(widget)
    self.pl=pango.Layout(pcont)
    self.sel_gc = widget.window.new_gc()
    #self.sel_gc.set_line_attributes(2,gtk.gdk.LINE_SOLID,
    #                                gtk.gdk.CAP_BUTT,gtk.gdk.JOIN_ROUND)
    self.sel_gc.set_rgb_fg_color(gtk.gdk.color_parse('Red'))    
    self.presel_gc = widget.window.new_gc()
    self.presel_gc.set_rgb_fg_color(gtk.gdk.color_parse('Blue'))    
    self.unsel_gc= widget.window.new_gc()
    self.unsel_gc.set_fill(gtk.gdk.STIPPLED)
    self.mask_gc=texture.make_mask_gc(widget)
    # self.unsel_gc.set_attribute

  def draw_image(self,widget,im,x,y,centered=False):
    """
    im is either
      an Image instance to draw
      a (im,text) where the text is to be displayed at top left of the image
    """        
    if type(im)==type(()):
      add_text=im[1]
      im=im[0]
    else:
      add_text=None
    (scale,thumb)=im.get_thumb(
      self.mainwidget.im_width-5,self.mainwidget.im_height-5,
      widget.window,self.gc)
    (xd,yd)=thumb.get_size()
    if centered:
      x-=xd/2; y-=yd/2
    widget.window.draw_drawable(self.gc,thumb,0,0,int(x),int(y),-1,-1)

    if add_text:
      self.pl.set_text(add_text)
      ps=self.pl.get_pixel_size()
      if ps[0]<self.mainwidget.im_width:
        widget.window.draw_layout(self.gc,x,y,self.pl,
                                  gtk_color_black,gtk_color_white)

    if im.mask!=None:      
      im.mask.draw_on_window(widget,self.mask_gc,scale,(x,y,xd,yd))
   
    if im.selected:
      gc=self.sel_gc
    elif hasattr(im,'preselected') and im.preselected:
      gc=self.presel_gc
    else:
      gc=None
    if gc:
      widget.window.draw_rectangle(gc,False,int(x-2),int(y-2),xd+3,yd+3)

  def init_interactive(self):
    self.connect('button-press-event', self.on_button_press)
    self.connect('button-release-event', self.on_button_release)
    self.connect('motion-notify-event', self.on_motion_notify)
    self.set_property("events",
                      gtk.gdk.BUTTON_PRESS_MASK|
                      gtk.gdk.BUTTON_RELEASE_MASK |
                      gtk.gdk.BUTTON_MOTION_MASK)
    
  def on_button_press(self,widget,ev):
    self.modify_sel=(ev.get_state()&gtk.gdk.CONTROL_MASK) != 0
    # is the selection to be replaced or modified ?
    self.mainwidget.clear_preselected()
    self.init_coords=ev.get_coords()
    self.second_coords=None
    self.update_selection(self.init_coords,self.init_coords)

  def draw_selection_rect(self,widget):
    if self.second_coords==None: return
    (x0,y0)=self.init_coords
    (x1,y1)=self.second_coords
    if x1<x0: (x0,x1)=(x1,x0)
    if y1<y0: (y0,y1)=(y1,y0)    
    widget.window.draw_rectangle(self.sel_gc,False,
                                 int(x0),int(y0),int(x1-x0),int(y1-y0))
    
  def on_motion_notify(self,widget,ev):
    self.second_coords=ev.get_coords()
    self.update_selection(self.init_coords,self.second_coords)    

  def update_selection(self,p0,p1):
    self.preselect(self.init_coords,p1)
    self.queue_draw()

  def on_button_release(self,widget,ev):
    self.second_coords=None
    self.mainwidget.preselection_to_selection(self.modify_sel)

  #def on_mask_change(self,im):
  #  self.queue_draw()
  



class DBWidgetLines(DBWidgetCommon):
  """ images are organized in "paragraphs" introduced by titles"""


  def __init__(self,mainwidget):
    DBWidgetCommon.__init__(self,mainwidget)
    self.lines=[]
    self.h=0; self.tried_h=-1    

  def on_size_allocate(self, widget, allocation):
    self.width = allocation.width
    h=self.calc_layout()
    if allocation.height!=h and not (allocation.height>h and h==self.tried_h):
      self.h=h
      self.queue_resize()
    
  def on_size_request(self,widget,requisition):
    requisition.width=0
    requisition.height=self.h
    self.tried_h=self.h

  def calc_layout(self):
    """ transform self.timages into self.lines
    self.lines = list of tuples
       (y_min,content)
      where:
      y_min is the upper limit of the line
      content is either
        a string: it's a group title
        a list of (x,im): draw im at coordinate (x,y_min)
        nothing (the tuple has only 1 elt) for the end of the list
    """
    # round down
    ncol=self.width/self.mainwidget.im_width
    if ncol==0: ncol=1
    self.col_w=self.width/ncol
    cur_y=0
    self.lines=[]
    for title,ims in self.mainwidget.timages:
      self.lines.append((cur_y,title))
      cur_y+=20
      for iml in packs(ims,ncol):
        l=[]
        cur_x=0
        for im in iml:
          l.append((cur_x,im))
          cur_x+=self.col_w
        self.lines.append((cur_y,l))
        cur_y+=self.mainwidget.im_height
    self.lines.append((cur_y,))
    return cur_y
  
  def on_expose_event(self, widget, event):
    for lno,line in enumerate(self.lines):
      y_min=line[0]
      if len(line)==1:
        continue # end sentinel
      y_max=self.lines[lno+1][0]-1
      if y_max<event.area.y or y_min>=event.area.y+event.area.height:
        continue
      elif type(line[1])==type(''):
        self.pl.set_text(line[1])
        widget.window.draw_layout(self.gc,2,y_min+2,self.pl)
      else:
        for x,im in line[1]:
          self.draw_image(widget,im,x,y_min)
    self.draw_selection_rect(widget)


  def preselect(self,(x0,y0),(x1,y1)):
    # feels like text selection
    if x1<x0: (x0,x1)=(x1,x0)
    if y1<y0: (y0,y1)=(y1,y0)
    self.mainwidget.clear_preselected()
    for lno,line in enumerate(self.lines):
      y_min=line[0]
      if len(line)==1:
        continue # end sentinel
      y_max=self.lines[lno+1][0]-1
      if y_min>y1 or y_max<y0:
        continue
      elif type(line[1])==type(''):
        if y_min<y0 and y1<y_max:
          # select whole group
          for l2 in self.lines[lno+1:-1]:
            if type(l2[1])==type(''): break
            for x,im in l2[1]:
              if type(im)==type(()): im=im[0]
              im.preselected=True
      else:
        sel_begin=y0<y_min
        sel_end=y1>y_max
        for x_min,im in line[1]:
          x_max=x_min+self.mainwidget.im_width-5
          if (sel_begin or x0<x_max) and (sel_end or x1>x_min):
            if type(im)==type(()): im=im[0]            
            im.preselected=True


class DBWidgetCanvas(DBWidgetCommon):
  """
  images are displayed at arbitrary coordinates
  """

  def __init__(self,mainwidget):
    DBWidgetCommon.__init__(self,mainwidget)
    self.cimages=[]
    self.autoscale=False

  def on_size_allocate(self, widget, allocation):
    # print "on_size_allocate",allocation.width,allocation.height
    if self.autoscale:
      self.width=allocation.width
      self.height=allocation.height
      self.layout_autoscaled()
      
  def on_size_request(self,widget,requisition):
    if not self.autoscale:
      (w,h)=self.layout_unconstrained()
      requisition.width=w
      requisition.height=h
    # else we have nothing to say, but it's ok
    
  def find_bbox(self):
    " transform self.timages into self.cimages "
    # round down
    self.cimages=[]

    xmin=ymin=10000; xmax=ymax=-10000;
    for x,y,im in self.mainwidget.timages:
      xi=int(x); yi=int(y)
      self.cimages.append((xi,yi,im))
      if xi<xmin: xmin=x
      if yi<ymin: ymin=y
      if xi>xmax: xmax=x
      if yi>ymax: ymax=y      

    return (xmin,ymin,xmax,ymax)

  def layout_unconstrained(self):
    (xmin,ymin,xmax,ymax)=self.find_bbox()
    iw_2=self.mainwidget.im_width/2
    ih_2=self.mainwidget.im_height/2
    
    xmin-=iw_2; xmax+=iw_2
    ymin-=ih_2; ymax+=ih_2
    self.shift_x=-xmin
    self.shift_y=-ymin
    return (xmax-xmin,ymax-ymin)    

  def layout_autoscaled(self):
    (xmin,ymin,xmax,ymax)=self.find_bbox()
    if xmax==xmin: xmax=xmin+1
    if ymax==ymin: ymax=ymin+1
    iw=self.mainwidget.im_width
    ih=self.mainwidget.im_height
    fx=(self.width-iw)/float(xmax-xmin)
    fy=(self.height-ih)/float(ymax-ymin)
    self.cimages=[(x*fx,y*fy,im) for (x,y,im) in self.cimages]    
    self.shift_x=-xmin*fx+iw/2.0
    self.shift_y=-ymin*fy+ih/2.0
    
    
  def on_expose_event(self, widget, event):
    for x,y,im in self.cimages:
      self.draw_image(widget,im,self.shift_x+x,self.shift_y+y,True)
    self.draw_selection_rect(widget)

  def preselect(self,(x0,y0),(x1,y1)):
    if x1<x0: (x0,x1)=(x1,x0)
    if y1<y0: (y0,y1)=(y1,y0)
    self.mainwidget.clear_preselected()
    iw_2=self.mainwidget.im_width/2
    ih_2=self.mainwidget.im_height/2    
    for x,y,im in self.cimages:
      x+=self.shift_x
      y+=self.shift_y      
      if rectangles_inter(
        x0,y0,x1,y1,
        x-iw_2,y-ih_2,x+iw_2,y+ih_2):
        if type(im)==type(()): im=im[0]
        im.preselected=True
        

class DBWidget(gtk.Frame):
  """ self.lt: current widget type """

  def __init__(self):
    gtk.Frame.__init__(self)

    # prepare both types of widget 
    
    self.lines_s=gtk.ScrolledWindow()    
    self.lines_s.set_policy(gtk.POLICY_NEVER,gtk.POLICY_ALWAYS)    
    self.lines_w=DBWidgetLines(self)
    self.lines_s.add_with_viewport(self.lines_w)
    
    self.canvas_s=gtk.ScrolledWindow()    
    self.canvas_s.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)    
    self.canvas_w=DBWidgetCanvas(self)    
    self.canvas_s.add_with_viewport(self.canvas_w)
    
    self.lt=-1 
    self.set_lt(0)
    self.set_im_scale(0.9)
    imageSelection.add_callback(self.on_selection_changed)

    database.change_mask.register(lambda im: self.on_selection_changed())
    
  def set_lt(self,no):
    if self.lt==no: return
    self.lt=no
    if self.get_child():
      self.remove(self.get_child())
    if no==0:
      self.add(self.lines_s)
      self.current_widget=self.lines_w
    else: # 1 or 2
      self.add(self.canvas_s)
      self.current_widget=self.canvas_w
      self.canvas_w.autoscale=no==2
    self.show_all() # force redraw  

  def set_im_scale(self,scale):
    self.im_width=int(160*scale)
    self.im_height=int(120*scale)
    self.current_widget.queue_resize() # forces redraw    

  def set_timages(self,timages,*args):
    if timages==[]:
      self.set_lt(0)
      self.timages=[]
    else:
      self.set_lt(timages[0])
      self.timages=timages[1:]
    self.current_widget.queue_resize() # forces redraw

  def on_selection_changed(self,*arg):
    self.current_widget.queue_draw()

  def clear_preselected(self):
    if self.lt==0:
      for (t,ims) in self.timages:
        for im in ims:
          if type(im)==type(()): im=im[0]
          im.preselected=False
    else:
      for x,y,im in self.timages:
        if type(im)==type(()): im=im[0]
        im.preselected=False

  def preselection_to_selection(self,modify_sel):
    selection=[]
    for im in database.images:
      # don't use timages because we don't want duplicates
      if hasattr(im,'preselected') and im.preselected:
        selection.append(im)
    self.clear_preselected()
    if not modify_sel: # ctrl not pressed
      imageSelection.set(selection)
    elif len(selection)==0:
      pass
    elif len(selection)==1: # toggle selection
      im=selection[0]
      sel=imageSelection.get()[:]
      if im.selected:
        sel.remove(im)
      else:
        sel.append(im)
      imageSelection.set(sel)
    else: # add selection
      imageSelection.set(imageSelection.get()+selection)


class MainWidget(gtk.VBox):

  def __init__(self):
    gtk.VBox.__init__(self)       
    # menu (up)
    mb=build_menubar([
        ("_File",None,[
          ("_New",'<control>N',self.new),
          ("_Open...",None,self.open),
          ("_Save",None,self.save),
          ("_Save as...",None,self.save_as),
          "_",
          ("_Quit",None,self.quit),    
        ]),
        ("Data_base",None,[
          ("_Add images...",None,self.add_images),
          ("Anal_yze basic ",None,lambda x: database.analyze_all('se')),
          ("Analyze +color +geometry",None,lambda x: database.analyze_all('seg')),
          ("Analyze +persons",None,lambda x: database.analyze_all('sep')),
          ("Analyze +texture patches",None,lambda x: database.analyze_all('set')),
          "_",
          ("_Save tags...",None,self.save_tags),
          ("_Load tags...",None,self.load_tags),
        ]),
        ("_Selection",None,[
          ("_Invert",None,self.on_invert),
          ("_Selection window",None,self.open_selected_window),    
        ]),
        ("Developer",None,[
          ("Clear exif info",None,self.clear_exif),
          ("debugger",None,self.run_debugger),
        ])

      ])            
    self.pack_start(mb,False)

    # split horizontally    
    paned=gtk.HPaned()
    self.add(paned)
    # left part
    left=DBWidget()
    paned.pack1(left,True)
    self.dbwidget=left
    # right part
    right=gtk.VBox()
    # navigation
    self.make_nav(right)
    # organizers
    self.make_organizer()
    right.add(self.organizer)
    # right.add(gtk.Frame()) # spacer
    self.make_scale_widget(right)    
    paned.pack2(right,False)    

  ######################################################################
  # Widgets
         
  def make_nav(self,parent):
    hb=gtk.HBox()
    parent.pack_start(hb,False)
    hb.add(gtk.Label("work on"))
    self.but_wo_sel=make_button("selection",self.on_wo_sel)
    self.but_wo_sel.set_sensitive(False)
    hb.add(self.but_wo_sel)
    self.but_wo_all=make_button("all",self.on_wo_all)
    self.but_wo_all.set_sensitive(False)
    hb.add(self.but_wo_all)
    imageSelection.add_callback(self.on_change_selection)
    
  def make_organizer(self):
    self.organizer=AltFrames("by:")
    self.organizer.set_size_request(164,-1)
    self.organizer.add_widget("file",PlainOrganizer(self.dbwidget.set_timages))
    self.organizer.add_widget("exif",ExifOrganizer(self.dbwidget.set_timages))
    self.organizer.add_widget("tags",TagOrganizer(self.dbwidget.set_timages))
    self.organizer.add_widget("color",ColorOrganizer(self.dbwidget.set_timages))
    self.organizer.add_widget("geometry",GeometryOrganizer(self.dbwidget.set_timages))
    self.organizer.add_widget("persons",PersonOrganizer(self.dbwidget.set_timages))
    self.organizer.add_widget("texture",TextureOrganizer(self.dbwidget.set_timages))
    self.organizer.add_widget("refine",UserOrganizer(self.dbwidget.set_timages))
    self.organizer.connect("switch",self.on_switch_organizer)
    self.organize()

  def make_scale_widget(self,parent):
    l=gtk.HBox()
    parent.pack_end(l,False)
    l.pack_start(gtk.Label("size:"),False)
    sb=gtk.HScale()
    sb.set_range(0.2,2.0)
    sb.set_value(0.9)
    sb.connect("value-changed",self.on_scale_changed)
    l.add(sb)    

  ######################################################################
  # Handlers
         
  def on_scale_changed(self,source):
    val=source.get_value()
    self.dbwidget.set_im_scale(val)

  def on_switch_organizer(self,org,new_n):
    self.organizer.get_widget().get_focus()
    
  def organize(self):
    # print "organize db images= ",database.images
    organizer=self.organizer.get_widget()
    organizer.organize()
    
  def dummy_op(self,arg):
    print "sorry not implemented"

  def quit(self,quit):
    gtk.main_quit()    
    
  def open(self,args):
    dirname=file_chooser("Load DB from",gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
    if not dirname:
      return  
    print 'load from',dirname
    database.clear()
    database.load_from(dirname)
    self.organize()
    
  def save_as(self,args):
    dirname=file_chooser("Save DB as",gtk.FILE_CHOOSER_ACTION_SAVE)
    if not dirname:
      return
    while dirname[-1]=='/': dirname=dirname[:-1]
    if dirname[-5:]!='.yorg':
      dirname+='.yorg'
    print "making dir ",dirname
    try:
      os.mkdir(dirname)
    except OSError,e: # ignore exists
      if e.errno!=17:
        raise e      
    database.save_as(dirname)

  def save(self,sender):
    name=database.previous_name()
    if name==None:
      self.save_as(None)
    else:
      database.save_as(name)      

  def new(self,args):
    database.clear()

  def add_images(self,args):
    dirname=file_chooser("Add directory",gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
    if not dirname:
      return  
    database.add_from_dir(dirname)
    self.organize()

  def load_tags(self,args):
    fname=file_chooser("Load tags from",gtk.FILE_CHOOSER_ACTION_OPEN)
    if not fname:
      return  
    database.load_tags(fname)
    
  def save_tags(self,args):
    fname=file_chooser("Save tags to",gtk.FILE_CHOOSER_ACTION_SAVE)
    if not fname:
      return
    database.save_tags(fname)
    
  def on_invert(self,sender):
    imageSelection.invert()

  def clear_exif(self,sender):
    database.clear_exif()
    
  def run_debugger(self,sender):
    pdb.set_trace()
    
  def on_change_selection(self,new_sel):
    self.but_wo_sel.set_sensitive(new_sel!=[])     

  def on_wo_sel(self,sender):
    database.set_imageset(imageSelection.get())
    self.but_wo_all.set_sensitive(True)    
    self.organize()

  def on_wo_all(self,sender):
    database.reset_imageset()
    self.but_wo_all.set_sensitive(False)   
    self.organize()

  def open_selected_window(self,arg):
    """make a new selected window"""
    SelectedWindow()
    

for i in range(n_open_selected):
  SelectedWindow()


win = gtk.Window()

win.add(MainWidget())

win.set_geometry_hints(min_width=640,min_height=480)
win.set_title("image organizer")

win.show_all()
win.connect("destroy", lambda w: gtk.main_quit())

SelectedWindow()
gtk.main()

    
