
import traceback

import gtk,pdb
import libexif

from image import *
from image_db import *
from extra_functions import *
from extra_widgets import *


from math import *
import random
import attribute_classifier

class Organizer(gtk.Frame):
  def __init__(self,set_timages):
    gtk.Frame.__init__(self)
    self.set_timages=set_timages
    
  def unorganized(self,images):
    return [("en vrac",images[:])]

  def filter_errors(self,images):
    # filter out images for which we have no image data
    return [im for im in images if im.is_valid()]

  def organize(self):
    if database.images_to_display()==[]:
      print "nothing to organize"
      self.set_timages([0])
    else:
      print "Organizer.organize"
      timages=self.organize_images(database.images_to_display(),
                                   imageSelection.get())
      if timages==None: return # cancel display
      if timages!=[] and type(timages[0])!=type(1):
        timages=[0]+timages
      self.set_timages(timages)

  def get_focus(self):
    # by default, reorganize. Override if it is too expensive
    # or makes no sense
    self.organize()


class PlainOrganizer(Organizer):
  def __init__(self,set_timages):
    Organizer.__init__(self,set_timages)
    buttons=RadioButtons(
      ['vrac','filename','dirname','filetype','file size','random'],
      self.change_order)
    self.add(buttons)
    self.otype=0

  def change_order(self,no):
    self.otype=no
    self.organize()    

  def organize_images(self,images,sel):
    if self.otype==0:
      return self.unorganized(images)
    elif self.otype==1: # filename
      im2=[(i.fname[i.fname.rfind('/')+1:],i) for i in images]           
      im2.sort()
      return [("by filename",[i for key,i in im2])]
    elif self.otype==2:
      im2=[(i.fname[:i.fname.rfind('/')],i) for i in images]      
      return clusterize(im2)
    elif self.otype==3:
      im2=[(i.fname[i.fname.rfind('.'):].lower(),i) for i in images]
      return clusterize(im2)
    elif self.otype==4:
      im2=[(i.fstat.st_size,i) for i in images]
      im2.sort()
      im2=[(sz/(100*1024),i) for (sz,i) in im2]
      cl=clusterize(im2)
      return [("> "+lab+"00 kB",ims) for (lab,ims) in cl]
    elif self.otype==5:
      ims=images[:]
      random.shuffle(ims)
      return [("random",ims)]
      
class ExifOrganizer(Organizer):
  def __init__(self,set_timages):
    Organizer.__init__(self,set_timages)
    buttons=RadioButtons(
      ['date','camera','F-number','exposure time','orientation'],
      self.change_order)
    self.add(buttons)
    self.otype=0

  def change_order(self,no):
    self.otype=no
    self.organize()    

  def organize_images(self,images,sel):
    images2=[]
    no_ei=[]
    for i in images:
      if i.get_exif_info():
        images2.append(i)
      else:
        no_ei.append(i)        
    if self.otype==2:      
      im2=[(i.get_exif_info().a,i) for i in images2]      
    elif self.otype==3:      
      im2=[(i.get_exif_info().t,i) for i in images2]      
    elif self.otype==0:
      im2=[(robust_date(i.get_exif_info()),i) for i in images2]      
    elif self.otype==1:
      im2=[(i.get_exif_info().model,i) for i in images2]      
    elif self.otype==4:
      im2=[(i.get_exif_info().orientation,i) for i in images2]
      im2=[(libexif.ExifInfo.getOrientationName(name),ims) for (name,ims) in im2]
    return clusterize(im2)+[('no exif info',no_ei)]
    
class PersonOrganizer(Organizer):
  
  def __init__(self,set_timages):
    Organizer.__init__(self,set_timages)
    vb=gtk.VBox()
    self.add(vb)
    vb.add(gtk.Label(''))
    b=make_button('organize',self.on_organize)
    vb.pack_start(b,False)
    
  def on_organize(self,unused):
    self.organize()    
    
  def organize_images(self,images,sel):
    database.analyze_all('g',images)
    pi=[(len(im.get_persons()),im) for im in images]
    pi=[(name+(name>"1" and " persons" or " person"),l)
        for (name,l) in clusterize(pi)]
    pi.reverse()
    return pi

  def get_focus(self):
    pass

 
class TagOrganizer(Organizer):
  
  def __init__(self,set_timages):
    Organizer.__init__(self,set_timages)
    tagManager.register(self.on_tag_change)
    vbox=gtk.VBox()
    self.add(vbox)
    hbox=gtk.HBox()
    vbox.pack_start(hbox,False)
    hbox.add(make_button('and',self.on_organize_and_or))
    hbox.add(make_button('or',self.on_organize_and_or))
    self.tag_list=TagList(['tags'])
    self.tag_list.double_click_callback(self.on_tag_double_click)
    vbox.add(self.tag_list)
    self.otype='and'
    self.update_tag_list()
    
  def update_tag_list(self,keep_selection=False):
    self.tag_list.set_tags(tagManager.all_tags)
    # self.organize_callback()
    # not necessarily

  def on_organize_and_or(self,org):
    self.otype=org.get_label()
    self.organize()

  def on_tag_change(self):
    self.update_tag_list(True)
    
  def on_tag_double_click(self,tag):
    self.otype='or' # and works also
    self.organize()
  
  def organize_images(self,images,sel):
    tags=self.tag_list.get_selected_tags()
    if self.otype=='and':      
      if tags==[]:
        return [('all',images[:])]
      title=tags[0]
      for tag in tags[1:]:
        title+=" AND "+tag
      sel=[]
      for im in images:
        for tag in tags:
          if tag not in im.tags: break
        else:
          sel.append(im)
      print (title,sel)
      return [(title,sel)]
    elif self.otype=='or':
      timages=[]
      for tag in tags:
        sel=[]
        for im in images:
          if tag in im.tags:
            sel.append(im)
        timages.append((tag,sel))
      return timages
    assert False

class OrganizerWithGroup:

  def make_group_widgets(self,parent):
    sb=gtk.HScale()
    sb.set_range(0,1)
    sb.set_value(0)
    sb.set_update_policy(gtk.UPDATE_DELAYED)
    sb.set_digits(3)
    sb.set_sensitive(False)
    self.thresh_sb=sb
    self.cross_rel_thresh=0
    sb.set_value(self.cross_rel_thresh)
    sb.connect("value-changed",self.on_thesh_changed)
    parent.pack_start(sb,False)
    
    b=make_button('organize',self.on_organize)
    parent.pack_start(b,False)
    
  def on_organize(self,arg):
    self.organize()

  def on_thesh_changed(self,source):
    self.cross_rel_thresh=source.get_value()
    self.organize()    


class ColorOrganizer(Organizer,OrganizerWithGroup):

  def __init__(self,set_timages):
    Organizer.__init__(self,set_timages)
    vb=gtk.VBox()
    self.add(vb)
    buttons=RadioButtons(
      ['_Ldominant color','luminance','saturation','hue',       
       '_Ldistance to selection','histogram','saturated histogram',
       '_Lauto-organize','group','mds'],
      self.change_order)
    self.buttons=buttons
    vb.add(buttons)

    self.make_group_widgets(vb)
    
    self.otype=0
    self.chc=color.ColorHistogramCommon()
    self.cached_images=None
    self.mds_pos=None

  def change_order(self,no):
    self.otype=no
    self.thresh_sb.set_sensitive(False)
    self.mds_pos=None
    
  def get_focus(self):
    self.buttons.set(0)
    self.organize()

  def organize_images(self,images,sel):
    images=self.filter_errors(images)
    if self.otype in [0,1,2,5,6]:
      return self.organize_absolute(images)
    else:
      if sel==[]:
        message("using first image as reference",False)
        ref=images[0]
      elif len(sel)>1:
        message("using first selected image as reference",False)
        ref=sel[0]
      else:
        ref=sel[0]
      return self.organize_relative(images,ref)

  def organize_relative(self,images,ref_im):
    ref_h=ref_im.get_hist()
    if self.otype==3:
      dfun=self.chc.earthMoversDistance
      im2=[(dfun(ref_h,i.get_hist()),i) for i in images if i!=ref_im]
    else:
      # TODO saturation enhanced version
      dfun=self.chc.earthMoversDistance
      im2=[(dfun(ref_h,i.get_hist()),i)
           for i in images if i!=ref_im]
    print "dists=",im2
    im2.sort()
    return [
      ('reference',[ref_im]),
      ('closest ones',[i for (lum,i) in im2])]

  def get_cross_dists(self,images):
    if self.cached_images!=images:
      n=len(images)
      dfun=self.chc.earthMoversDistance
      dists=[]
      pb=ProgressBarWindow("computing cross-relations",n*(n-1)/2)
      for i,im in enumerate(images):
        h=im.get_hist()
        dline=[]
        for im2 in  images[:i]:
          dline.append(dfun(h,im2.get_hist()))
          pb.one_step()
        dists.append(dline)
      pb.destroy()
      self.cached_cross_rels=complete_sym_mat(dists)
      self.cached_images=images
    return self.cached_cross_rels
  

  def organize_absolute(self,images):      
    if self.otype==0:
      im2=[(rgb_to_luminance(i.dominant_color()),i) for i in images]
      im2.sort()
      return [('by luminance',[i for (lum,i) in im2])]
    if self.otype==1:
      im2=[(rgb_to_hsl(i.dominant_color())[1],i) for i in images]
      im2.sort()
      return [('by saturation',[i for (lum,i) in im2])]
    if self.otype==2:
      im2=[(rgb_to_hsl(i.dominant_color())[0],i) for i in images]
      im2.sort()
      dom=([],[],[])
      for hue,i in im2:
        bin=(int((hue-1)/2.0)+3)%3        
        dom[bin].append(i)
      return [('dominant green',dom[0]),
              ('dominant blue',dom[1]),
              ('dominant red',dom[2])]
    if self.otype in (5,6): # grouping
      self.thresh_sb.set_sensitive(True)
      # compute all cross-distances
      cross_rels=self.get_cross_dists(images)
      if self.otype==5:
        vmax=max([max(l) for l in cross_rels])     
        groups=breakup_cross_rels(cross_rels,self.cross_rel_thresh*vmax,True)
        return [("Group",[images[i] for i in g]) for g in groups]
      else:
        tim=[]
        if self.mds_pos==None:
          self.mds_pos=cross_rels_to_mds_pos(cross_rels,lambda x:x/100.0)
        for i,im in enumerate(images):
          (x,y)=self.mds_pos[i]
          f=(self.cross_rel_thresh+0.3)*5000
          x*=f; y*=f
          tim.append((x,y,im))
        self.thresh_sb.set_sensitive(True)        
        return [1]+tim
        
      
    
class GeometryOrganizer(Organizer,OrganizerWithGroup):

  def __init__(self,set_timages):
    Organizer.__init__(self,set_timages)
    vb=gtk.VBox()
    self.add(vb)
    buttons=RadioButtons(
      ['_Ldescription','nb of keypoints',
       '_Ldistance to selection','no geometry','weak geometry',
       '_LAmazing!','auto-group','circle','mds'],
      self.change_order)
    vb.add(buttons)

    self.make_group_widgets(vb)
    self.otype=0    

  def get_focus(self):
    database.analyze_all('g')
   
  def change_order(self,no):
    self.otype=no
    self.thresh_sb.set_sensitive(False)

  def organize_images(self,images,sel):
    images=self.filter_errors(images)
    if not have_photomole:
      return [('PhotoMole not available!',[])]
    if self.otype in [0,3,4,5,6]:
      return self.organize_absolute(images)
    else:
      if sel==[]:
        message("using first image as reference",False)
        ref=images[0]
      elif len(sel)>1:
        message("using first selected image as reference",False)
        ref=sel[0]
      else:
        ref=sel[0]
      return self.organize_relative(images,ref)


  def organize_absolute(self,images):      
    if self.otype==0:
      im2=[(i.get_imdesc().getNDescs(),i) for i in images if i.is_valid()]
      im2.sort()
      return [('increasing nb of keypoints',[i for (lum,i) in im2])]
    elif self.otype==3:
      # cross_rels=crossRelations.get_cross_rels(images)
      cross_rels=crossRelations.get_normalized_cross_rels(images)
      vmax=max([max(l) for l in cross_rels])
      groups=breakup_cross_rels(cross_rels,(self.cross_rel_thresh**4)*vmax)
      self.thresh_sb.set_sensitive(True)
      return [("Group",[images[i] for i in g]) for g in groups if len(g)>1]
    elif self.otype==4:
      tim=[]
      n=len(images)
      for i,im in enumerate(images):        
        x=self.cross_rel_thresh*1000*cos(i*pi*2/n)
        y=self.cross_rel_thresh*1000*sin(i*pi*2/n)
        tim.append((x,y,im))
      self.thresh_sb.set_sensitive(True)        
      return [1]+tim
    elif self.otype==5:
      tim=[]
      pos=crossRelations.get_mds_pos(images,self.cross_rel_thresh)
      for i,im in enumerate(images):
        (x,y)=pos[i]
        #f=(self.cross_rel_thresh+0.3)*1000
        #x*=f; y*=f
        tim.append((x*1000.0,y*1000.0,im))
      self.thresh_sb.set_sensitive(True)        
      return [2]+tim
    elif self.otype==6:
      self.thresh_sb.set_sensitive(True)        
      oi=crossRelations.get_reasonable_order(images,self.cross_rel_thresh)
      return [('resonable order',oi)]
    elif self.otype==7:
      tim=[]
      pos=crossRelations.get_lapeig_pos(images,self.cross_rel_thresh)
      for i,im in enumerate(images):
        (x,y)=pos[i]
        #f=(self.cross_rel_thresh+0.3)*1000
        #x*=f; y*=f
        tim.append((x*1000.0,y*1000.0,im))
      self.thresh_sb.set_sensitive(True)        
      return [2]+tim
      
      
  def organize_relative(self,images,ref_im):
    if self.otype in [1,2]:
      imdesc=ref_im.get_imdesc()
      mat=geomIndex.find_matches_raw(imdesc,self.otype==2)
      # remap matches to correspond to images
      m2=[(mat.getWeightedVotes(i.geom_imno),i) for i in images]
      m2.sort()
      m2.reverse()
      return [
        ('reference',[ref_im]),
        ('decreasing nb of matches',[i for (n,i) in m2 if n>=5]),
        ('few or no matching keypoints',[i for (n,i) in m2 if n<5])]
    return self.unorganized(images)
  


class PosNegOrganizer(Organizer):

  def __init__(self,set_timages):
    Organizer.__init__(self,set_timages)
    vbox=gtk.VBox()
    self.clear_but=make_button('clear',self.clear_all)
    vbox.pack_start(self.clear_but,False)
    vbox.pack_start(gtk.Label("selection is:"),False)    
    vbox.pack_start(make_button(
      'positive',
      lambda x: self.label_sel(True)),False)
    vbox.pack_start(make_button(
      'negative',
      lambda x: self.label_sel(False)),False)
    self.statsview=gtk.Label()
    vbox.pack_start(self.statsview)
    
    vbox.add(gtk.Label("")) # spacer
    hbox=gtk.HBox()
    self.organize_but=make_button('organize',lambda x:self.organize())
    hbox.add(self.organize_but)
    self.handle_all_cb=gtk.CheckButton('all',False)
    hbox.pack_start(self.handle_all_cb,False)
    vbox.pack_start(hbox,False)
   
    self.add(vbox)
    self.pos_selection=[]
    self.neg_selection=[]
    self.printstats()

  def get_focus(self):
    """ reset selection """
    self.printstats()

  def has_pos_and_neg(self):
    return self.pos_selection!=[] and self.neg_selection!=[]

  def statstring(self):
    return "%dp + %dn"%(len(self.pos_selection),len(self.neg_selection))    

  def printstats(self):
    allclear=self.pos_selection==[] and self.neg_selection==[]
    self.clear_but.set_sensitive(not allclear)
    self.organize_but.set_sensitive(self.has_pos_and_neg())
    
    if allclear:
      self.statsview.set_text("-")      
    else:
      self.statsview.set_text(self.statstring())
      
  def label_sel(self,ispos):
    if ispos:
      (toadd,tosub)=(self.pos_selection,self.neg_selection)
    else:
      (toadd,tosub)=(self.neg_selection,self.pos_selection)
    for im in imageSelection.get():
      if im in toadd:
        pass
      else:
        toadd.append(im)
        try:
          tosub.remove(im)
        except ValueError:
          pass
    self.printstats()

  def clear_all(self,arg):    
    self.pos_selection=[]
    self.neg_selection=[]
    self.printstats()
    
class UserOrganizer(PosNegOrganizer):

  def __init__(self,set_timages):
    PosNegOrganizer.__init__(self,set_timages)
    self.prev_imsubset=(None,None)

  def imageset_key(self,images):
    key=0xeadbeef
    for i in images:
      key=key ^ hash(i)
    return key


  def organize_images(self,images,sel):
    self.printstats()

    # modname="attribute_classifier"
    # if modname in sys.modules.keys():
    #   del sys.modules[modname]
    # import attribute_classifier

    max_class=20

    if len(images)<max_class or self.handle_all_cb.get_active():
      images_to_classify=images
      other_images=[]
      self.prev_imsubset=(None,None)      
    else:
      key=self.imageset_key(images)
      if self.prev_imsubset[0]==key:
        print "re-using previous subset"
        (images_to_classify,other_images)=self.prev_imsubset[1]
      else:      
        (images_to_classify,other_images)=random_split(images,max_class)
        self.prev_imsubset=(key,(images_to_classify[:],other_images[:]))
    
    # add train images if they are in image set
    images_to_classify+=list_sub(
      other_images,self.pos_selection+self.neg_selection)
            
    im_with_flabel=attribute_classifier.classify_with_attrs(
      images_to_classify,
      self.pos_selection,self.neg_selection)

    if im_with_flabel==None: return None

    im_with_flabel.sort()
    im_with_flabel.reverse()

    pos_im=[(im,"%.2f"%flabel) for (flabel,im) in im_with_flabel if flabel>=0]
    neg_im=[(im,"%.2f"%flabel) for (flabel,im) in im_with_flabel if flabel<0]

    return [('positives',pos_im),
            ('negatives',neg_im),
            ('other',other_images)]


class TextureOrganizer(Organizer):


  def __init__(self,set_timages):
    Organizer.__init__(self,set_timages)
  
    vbox=gtk.VBox() 

    self.texlaview=wl=SingleSelectTagList(
      ['name','pos','neg','cl'],allTexlas.get_as_tags())

    self.texlaview.set_select_tag_cb(self.on_select)
    
    vbox.add(wl)

    self.add(vbox)

    hbox=gtk.HBox()
    self.new_but=make_button("new",self.on_new)
    hbox.add(self.new_but)
    self.del_but=make_button("del",self.on_del)
    hbox.add(self.del_but)
    self.rename_but=make_button("rename",self.on_rename)
    hbox.add(self.rename_but)
    #self.dup_but=make_button("dup",self.on_rename)
    #hbox.add(self.dup_but)
    vbox.pack_start(hbox,False)
    self.learn_but=make_button("learn",self.on_learn)
    vbox.pack_start(self.learn_but,False)

    hbox=gtk.HBox()
    self.organize_but=make_button('organize',lambda x:self.organize())
    hbox.add(self.organize_but)
    self.handle_all_cb=gtk.CheckButton('all',False)
    hbox.pack_start(self.handle_all_cb,False)
    vbox.pack_start(hbox,False)
    
    self.sel=[]

    self.in_on_select=False
    
    self.last_tags=None
    self.sync_view()

    database.change_mask.register(self.on_mask_change)
    allTexlas.change.register(self.sync_view)
    allTexlas.classifier_change.register(self.sync_view)


  def sync_view(self):    
    tags=allTexlas.get_as_tags()
    self.learn_but.set_sensitive(allTexlas.can_learn())
    self.organize_but.set_sensitive(allTexlas.get_classifier()!=None)
    self.del_but.set_sensitive(not allTexlas.is_read_only() and
                               allTexlas.n_texla()>1)
    self.rename_but.set_sensitive(not allTexlas.is_read_only())
    # avoid event deadlock (yes it's crappy)
    if (tags==self.last_tags
        and self.texlaview.get_selected_no()==allTexlas.selected):
      return
    self.last_tags=tags
    self.texlaview.set_tags_and_sel(tags,allTexlas.selected)
    
  def on_mask_change(self,im):
    if self.in_on_select: return
    # pdb.set_trace()
    imno=im.get_tex_imno()
    allTexlas.set_mask(imno,im.mask,im.size,im.get_tex_scale())

  def on_rename(self,arg):
    prompt='name'
    while True:
      name=ask_string(prompt)
      if name==None: return    # cancel
      if name not in allTexlas.all_names():
        break
      prompt='name exists'
    allTexlas.set_name(name)
    
  def get_focus(self):
    pass

  def on_new(self,arg):
    #name=ask_string('name')
    #if name==None: return    
    allTexlas.add_new_texla()

  def sync_masks_with_sel(self):
    """remove old masks & set new ones"""
    if allTexlas.is_read_only():
      for im in database.images:
        im.mask=ReadOnlyVectorMask
    else:
      d=allTexlas.mask_map(allTexlas.selected)    
      for im in database.images:
        if (im.tex_imno==None or not d.has_key(im.tex_imno)):
          im.mask=None
        else:
          im.mask=d[im.tex_imno]
    

  def on_select(self,no):
    self.in_on_select=True
    allTexlas.set_selected(no)
    self.sync_masks_with_sel()
    database.change_mask.broadcast(None,None)
    self.in_on_select=False

  def on_del(self,arg):
    if allTexlas.n_texla()==1: return
    self.in_on_select=True
    allTexlas.del_texla()
    self.sync_masks_with_sel()
    database.change_mask.broadcast(None,None)
    self.in_on_select=False
    
  def on_learn(self,arg):
    print "learn:"
    allTexlas.do_learn()
    
  def organize_images(self,images,sel):

    cl=allTexlas.get_classifier()

    if not cl: return self.unorganized(images)

    if len(images)<30 or self.handle_all_cb.get_active():
      to_class=images
      the_rest=[]      
    else:
      to_class=[im for im in images if im.mask and not im.mask.is_clear()]
      the_rest=[im for im in images if not im.mask or im.mask.is_clear()]
      (to_class_add,the_rest)=random_split(the_rest,30)
      to_class=to_class_add+to_class

    ct=[]
    pb=ProgressBarWindow("classifying textures...",len(to_class))
    skip=False
    for im in to_class:
      if skip:
        the_rest.append(im)
      else: 
        ct.append((textureManager.classify(cl,im.get_tex_imno()),im))
      if not pb.one_step(): skip=True
    pb.destroy()
    pb.print_time()
      
    unclassified=[im for (sc,im) in ct if sc==None]
    ct=[(sc,im) for (sc,im) in ct if sc!=None]
    ct.sort()
    ct.reverse()
    print ct
    ct=[(im,"%.3f"%sc) for (sc,im) in ct]
    
    ret=[('pos to neg',ct)]
    if unclassified:
      ret+=[('not classifiable',unclassified)]
    
    if the_rest:
      ret+=[('not classified',the_rest)]
    
    return ret
  

