


import gtk,pango

from extra_functions import *
from extra_widgets import *

import geometry

from image import *

from image_db import *

from texture import *

def dummy(x):
  print "dummy"
  


class SelectedWindow(gtk.Window):

  def __init__(self):
    gtk.Window.__init__(self)
    self.set_geometry_hints(min_width=400,min_height=300)
    self.connect("destroy", self.on_destroy)
    c=gtk.VBox()
    self.add(c)
    self.selection=[]

    # filename
    self.fname_entry=gtk.Label()
    # self.fname_entry.set_property("selectable",False)
    # self.fname_entry.set_ellipsize(pango.ELLIPSIZE_START)
    self.fname_entry.set_justify(gtk.JUSTIFY_RIGHT) # doesn't work ?
    c.pack_start(self.fname_entry,False)

    nb=gtk.Notebook()
    c.add(nb)
    
    l=gtk.HBox()
    nb.append_page(l,gtk.Label("general"))
    # c.add(l)
    
    # left
    # sw=gtk.ScrolledWindow()
    sw=gtk.Frame()
    l.add(sw)
    #sw.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
    tv=gtk.TextView()
    tv.set_property("editable",False)
    sw.add(tv)
    tb=gtk.TextBuffer()
    self.text_info=gtk.TextBuffer()
    tv.set_buffer(self.text_info)
    
    # right
    c2=gtk.VBox()
    c2.set_size_request(120,120)
    l.pack_start(c2,False)
    c2.pack_start(make_button("in GIMP",self.on_in_gimp),False)
    self.tag_list=TagList(['tags','which'])
    c2.add(self.tag_list)
    bl=gtk.HBox()
    c2.pack_start(bl,False)
    bl.add(make_button("+",self.on_assign_tags))
    bl.add(make_button("-",self.on_remove_tags))    
    bl.add(make_button("new...",self.on_new_tag))    

    # histograms page    
    l=self.make_histogram_widget()
    nb.append_page(l,gtk.Label("colors"))

    # keypoints page
    self.kp_widget=KPTab(self)
    nb.append_page(self.kp_widget,gtk.Label("keypoints"))
    
    # person page
    self.person_widget=PersonTab(self)
    nb.append_page(self.person_widget,gtk.Label("persons"))

    # texture page
    self.tex_widget=TextureTab(self)
    nb.append_page(self.tex_widget,gtk.Label("texture"))   
    
    #init
    self.on_set_selection(imageSelection.get())
    self.show_all()
    tagManager.register(self.on_tags_changed)
    imageSelection.add_callback(self.on_set_selection)
    self.tag_list.double_click_callback(self.on_tag_double_click)
    database.change_mask.register(self.on_mask_change)
    

  def update_text_info(self):
    im=self.selection[0]
    self.text_info.set_text(im.long_str())

  def set_callbacks(self,addthem):
    if len(self.selection)==1:
      im=self.selection[0]
      if addthem:
        im.change.register(self.update_text_info)
      else:
        im.change.unregister(self.update_text_info)
  
  def on_set_selection(self,sel):
    self.set_callbacks(False)
    self.selection=sel
    if len(sel)==0:
      self.fname_entry.set_text("no selection")
      self.text_info.set_text("")
    elif len(sel)==1:
      im=sel[0]
      self.fname_entry.set_text(im.fname)
      self.update_text_info()
    else:
      pref=common_prefix([im.fname for im in sel])
      self.fname_entry.set_text(pref+'...')
      self.text_info.set_text("%d images\n"%len(sel))
    self.set_callbacks(True)
    self.update_tags()
    self.hist_widget.queue_draw()
    self.kp_widget.redraw()
    self.person_widget.redraw()
    self.tex_widget.new_selection()

  def on_mask_change(self,im):
    # update text part
    pass 
        
  def on_in_gimp(self,org):
    if self.selection==[]: return
    esc=[escape_sh(im.fname) for im in self.selection]    
    os.system("gimp-remote -n "+reduce(lambda accu,x:accu+' '+x,esc,'')+' &')
 
  def on_destroy(self,arg):
    print "selection win on_destroy"
    tagManager.unregister(self.on_tags_changed)
    imageSelection.remove_callback(self.on_set_selection)
    self.set_callbacks(False)
    self.tex_widget.on_destroy()
    
  ##################################################################
  # Tags

  def update_tags(self):
    used_tags={}
    for tag in tagManager.all_tags:
      used_tags[tag]=0
    for im in self.selection:
      for tag in im.tags:
        used_tags[tag]+=1
    tagl=[]
    ntot=len(self.selection)
    for tag in tagManager.all_tags:
      if ntot==0: who=''
      elif ntot==1:
        who=used_tags[tag]==0 and ' ' or 'yes'
      else:
        if used_tags[tag]==0: who=' '
        elif used_tags[tag]<ntot: who='%d'%used_tags[tag]
        else: who='all'
      tagl.append((tag,who))
    self.tag_list.set_tags(tagl)

  def on_tags_changed(self):
    self.update_tags()

  def on_assign_tags(self,arg):
    for tag in self.tag_list.get_selected_tags():
      tagManager.add_tag_to(self.selection,tag)

  def on_remove_tags(self,arg):
    print "tags=",self.tag_list.get_selected_tags()
    for tag in self.tag_list.get_selected_tags():
      tagManager.remove_tag_from(self.selection,tag)
      self.selection[0].tags
  
  def on_new_tag(self,arg):
    tagname=ask_string("new tag name ?")
    if tagname:
      tagManager.new_tag(tagname)

  def on_tag_double_click(self,tag):
    tagManager.add_tag_to(self.selection,tag)

  ##################################################################
  # Histograms
    
  def make_histogram_widget(self):
    w=self.hist_widget=gtk.DrawingArea()
    w.connect('expose-event',self.hist_expose_event)
    w.connect('realize',self.hist_realize)
    return w
    
  def hist_realize(self,widget):
    self.hist_gc = widget.window.new_gc()
    
  def hist_expose_event(self, widget, event):      
    n=len(self.selection)
    if n==0: return
    (xd,yd)=widget.window.get_size()
    h=yd/n-2
    if h<1: h=1
    for i,im in enumerate(self.selection):    
      ch=im.get_hist()

      if ch.__class__.__name__=='ClusteredColorHistogram':
        f=xd/float(ch.totw)
        s=0
        for colbin in ch.clusters:
          (r,g,b)=(colbin.r,colbin.g,colbin.b)
          col=gtk.gdk.Color(int(256*r),int(256*g),int(256*b))
          self.hist_gc.set_rgb_fg_color(col)
          widget.window.draw_rectangle(
            self.hist_gc,True,
            int(s*f),yd*i/n,int(colbin.w*f)+1,h)
          s+=colbin.w
      elif ch.__class__.__name__=='FixedColorHistogram':
        curves=ch.getCurvesPy()
        vmax=max([max(x) for x in curves])
        for j,colname in enumerate(['Red','Green','Blue']):
          l=[]
          for b in range(ch.nbin):
            assert curves[j][b]>=0
            l.append((int(xd*(b+0.5)/ch.nbin),
                     (i+1)*yd/n-curves[j][b]*yd/vmax/n))
          self.hist_gc.set_rgb_fg_color(gtk.gdk.color_parse(colname))
          widget.window.draw_lines(self.hist_gc,l)          
      

    

class KPTab(gtk.VBox):

  ##################################################################
  # Keypoints

  def __init__(self,selwin):
    gtk.VBox.__init__(self)
    self.selwin=selwin
    self.w=w=gtk.DrawingArea()
    self.add(w)
    w.connect('expose-event',self.kp_expose_event)
    w.connect('realize',self.kp_realize)
    hb=gtk.HBox()
    self.pack_start(hb,False)    
    self.bs_raw=bs=gtk.CheckButton("raw")
    hb.pack_start(bs,False)
    bs.connect("clicked",self.redraw)
    sb=self.kp_scale=gtk.HScale()
    # self.pack_start(sb,False)
    hb.add(sb)
    sb.set_range(0,0.2)
    self.kp_thresh=1
    sb.set_value(0)
    sb.set_digits(3)
    sb.set_sensitive(False)
    sb.connect("value-changed",self.kp_thesh_changed)

    
  def kp_realize(self,widget):
    self.kp_gc = widget.window.new_gc()
    self.kp_gc.set_rgb_fg_color(gtk.gdk.color_parse('Red'))    
    self.kp_gc2 = widget.window.new_gc()
    self.kp_gc2.set_rgb_fg_color(gtk.gdk.color_parse('Blue'))
    
  def kp_expose_event(self, widget, event):
    sel=self.selwin.selection
    self.kp_scale.set_sensitive(len(sel)==2)
    if len(sel)==1:
      self.kp_expose_event_1(widget,sel[0])
    elif len(sel)==2:
      self.kp_expose_event_2(widget,sel[0],sel[1])
      self.kp_scale.set_sensitive(True)
    # else nothing

  def redraw(self,*ignored):
    self.w.queue_draw()

  def arrange2im(self,(xd,yd),(xd0,yd0),(xd1,yd1)):
    s0=sqrt(xd0*yd0)
    (xs0,ys0)=(xd0/s0,yd0/s0)
    s1=sqrt(xd1*yd1)
    (xs1,ys1)=(xd1/s1,yd1/s1)
    scale_ud=min(xd/xs0,xd/xs1,yd/(ys0+ys1))
    scale_rl=min(yd/ys0,yd/ys1,xd/(xs0+xs1))
    if scale_ud>scale_rl:
      updown=True
      sc=scale_ud
    else:
      updown=False
      sc=scale_rl
    (xd0,yd0)=(int(sc/s0*xd0),int(sc/s0*yd0))
    (xd1,yd1)=(int(sc/s1*xd1),int(sc/s1*yd1))
    return (updown,(xd0,yd0),(xd1,yd1))
    
    
  def draw_im_in(self,widget,im,(x0,y0),(xd,yd),with_pplg):
    (scale,drawim)=im.get_thumb(xd,yd,widget.window,self.kp_gc,True)
    if not drawim: return
    imdesc=im.get_imdesc()
    scale=scale/im.geom_scale
    widget.window.draw_drawable(self.kp_gc,drawim,0,0,x0,y0,-1,-1)

    for i in range(imdesc.getNDescs()):
      desc=imdesc.getDesc(i)
      (x,y)=(int(desc.getx()*scale),int(desc.gety()*scale))
      widget.window.draw_point(self.kp_gc,x+x0,y+y0)
      if not with_pplg: continue
      pts=geometry.desc_to_parallelogram(desc,1.0)
      pts=[(int(x*scale+x0),int(y*scale+y0)) for (x,y) in pts]
      widget.window.draw_polygon(self.kp_gc2,False,pts)
    return scale

  def kp_expose_event_1(self, widget,im):
    (xd,yd)=widget.window.get_size()
    self.draw_im_in(widget,im,(0,0),(xd,yd),True)

      
  def kp_expose_event_2(self,widget,im1,im2):
    (xd,yd)=widget.window.get_size()
    (updown,sz0,sz1)=self.arrange2im((xd,yd),im1.size,im2.size)
    scale1=self.draw_im_in(widget,im1,(0,0),sz0,False)
    (x0,y0)=updown and (0,sz0[1]) or (sz0[0],0)
    scale2=self.draw_im_in(widget,im2,(x0,y0),sz1,False)
    use_geom=True
    matches=crossRelations.get_2image_matches(
      im1,im2,not self.bs_raw.get_active())
    imdesc1=im1.get_imdesc()
    imdesc2=im2.get_imdesc()

    # print "thresh=",self.kp_thresh
    totmat=0
    
    for (p1,p2,dist) in matches:      
      if dist>self.kp_thresh: continue
      desc1=imdesc1.getDesc(p1)
      (x1,y1)=(int(scale1*desc1.getx()),int(scale1*desc1.gety()))
      desc2=imdesc2.getDesc(p2)
      (x2,y2)=(int(x0+scale2*desc2.getx()),int(y0+scale2*desc2.gety()))
      widget.window.draw_line(self.kp_gc2,x1,y1,x2,y2)
      totmat+=1
    # print "totmat=",totmat
    
  def kp_thesh_changed(self,sender):
    self.kp_thresh=0.2-sender.get_value()
    self.queue_draw()    



class PersonTab(gtk.DrawingArea):

  def __init__(self,selwin):
    gtk.DrawingArea.__init__(self)
    self.selwin=selwin
    self.connect('expose-event',self.person_expose_event)
    self.connect('realize',self.person_realize)

  def redraw(self):
    self.queue_draw()

  def person_realize(self,widget):
    self.person_gc = widget.window.new_gc()
    self.person_gc.set_rgb_fg_color(gtk.gdk.color_parse('Red'))
    self.person_gc.set_line_attributes(2,gtk.gdk.LINE_SOLID,
                                       gtk.gdk.CAP_ROUND,gtk.gdk.JOIN_ROUND)
    pcont=gtk.Widget.create_pango_context(widget)
    
    self.text_gc = widget.window.new_gc()
    self.pl=pango.Layout(pcont)
    
  def person_expose_event(self, widget, event):
    sel=self.selwin.selection
    if len(sel)!=1: return

    im=sel[0]
    
    (xd,yd)=widget.window.get_size()
    
    (scale,drawim)=im.get_thumb(xd,yd,widget.window,self.person_gc,True)

    widget.window.draw_drawable(self.person_gc,drawim,0,0,0,0,-1,-1)
    pers=im.get_persons()
    scale=scale/im.persons_scale
    
    for (score,sc2,xmin,ymin,xdim,ydim) in pers:
      widget.window.draw_rectangle(
        self.person_gc,False,
        int(xmin*scale),int(ymin*scale),
        int(xdim*scale),int(ydim*scale))
    
      self.pl.set_text("%.2f"%score)
      widget.window.draw_layout(
        self.text_gc,max(int(xmin*scale),0),int(ymin*scale),
        self.pl,gtk_color_black,gtk_color_white)










class TextureTab(gtk.HBox):
  """ shows texture classifier result & allows to draw positive/negative
  regions
  """
  
  
  def __init__(self,selwin):
    gtk.HBox.__init__(self)
    self.selwin=selwin

    self.w=w=gtk.DrawingArea()
    # self.w=w=ColoredWidget()
    self.add(w)
    w.connect('expose-event',self.expose_event)
    w.connect('realize',self.on_realize)
    self.connect('motion-notify-event', self.on_motion_notify)
    w.connect('button-press-event', self.on_button_press)
    w.connect('button-release-event', self.on_button_release)
    w.set_property("events",
                   gtk.gdk.BUTTON_PRESS_MASK|
                   gtk.gdk.BUTTON_RELEASE_MASK |
                   gtk.gdk.BUTTON_MOTION_MASK)
    vb=gtk.VBox()
    self.pack_start(vb,False)
    vb.pack_start(RadioButtons([
      'show','edit pos','edit neg'
      ],self.ty_change),False)

    self.bs_clear=bs=gtk.Button("del curve")
    vb.pack_start(bs,False)
    bs.connect("clicked",self.clear_path)       
    self.cb_get_opp=SilentCheckButton("full")
    vb.pack_start(self.cb_get_opp,False)
    self.cb_get_opp.connect("clicked",self.get_from_opp)
    self.im=None
    self.multiple=None # image list 
    # im XOR multiple

    self.mode=0 # drawing mode (0, VectorMask.POS, VectorMask.NEG)
    self.l=None
    vb.add(gtk.Label("")) # spacer
    self.cb_get_opp.cancel_toggle=False
    self.set_states()
    
    allTexlas.classifier_change.register(self.on_classifier_change)
    self.shades=None

    database.change_mask.register(self.on_mask_change)
    
  def on_destroy(self):
    database.change_mask.unregister(self.on_mask_change)
    allTexlas.classifier_change.unregister(self.on_classifier_change)

  def multiple_state(self):
    accu=[0]*4
    for im in self.multiple:
      accu[im.mask.copied+1]+=1
    return (accu[0],accu[1:])
      
  def set_states(self):
    if self.im:
      mask=self.im.mask
      self.bs_clear.set_sensitive(
        mask.can_edit(self.mode) and
        len(mask.cur_curve(self.mode))>0)
      self.cb_get_opp.set_sensitive(self.mode!=0 and not mask.is_read_only())
      self.cb_get_opp.set_inconsistent(False)
      self.cb_get_opp.set_active_silent(mask.is_copied(self.mode))
    elif self.multiple:
      self.bs_clear.set_sensitive(False)
      if self.mode==0:
        self.cb_get_opp.set_sensitive(False)
      else:
        (nro,mstate)=self.multiple_state()
        self.cb_get_opp.set_sensitive(nro==0)
        if nro>0: return          
        n=len(self.multiple)
        othermode=self.mode==VectorMask.POS and VectorMask.NEG or VectorMask.POS
        if mstate[othermode]==n:
          self.cb_get_opp.set_active_silent(True)
          self.cb_get_opp.set_inconsistent(False)
        elif mstate[0]==n:          
          self.cb_get_opp.set_active_silent(False)
          self.cb_get_opp.set_inconsistent(False)
        else:
          self.cb_get_opp.set_inconsistent(True)        
    else: # no selection
      self.bs_clear.set_sensitive(False)
      self.cb_get_opp.set_sensitive(False)

  def new_selection(self):

    self.im=None
    self.multiple=None
    
    sel=self.selwin.selection

    if len(sel)==1:
      im=sel[0]
      if im.mask==None: im.mask=VectorMask()
      self.im=im
      
    if len(sel)>=2:
      for im in sel: 
        if im.mask==None: im.mask=VectorMask()
      self.multiple=sel[:]
      
    self.for_size=None
    self.shades=None
    self.set_states()
    self.queue_draw()
    
  def ty_change(self,no):
    self.mode=no
    self.set_states()    
    self.queue_draw()

  def get_drawim(self,widget,gc):
    size=widget.window.get_size()
    if size!=self.for_size:
      (xd,yd)=size
      (scale,self.drawim)=self.im.get_thumb(xd,yd,widget.window,gc,True)
      self.for_size=size
      self.scale=scale
    return self.drawim

  def init_shades(self,size):
    (xd,yd)=size
    (scale,xdo,ydo)=self.im.get_thumb_size(xd,yd)
    if (self.shades==None or
        (xdo,ydo)!=(self.shades.get_width(),self.shades.get_height())):
      self.shades=gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB,False,8,xdo,ydo)
      textureManager.show_patch_classes(
        allTexlas.get_classifier(),
        self.im.get_tex_imno(),
        scale/self.im.get_tex_scale(),self.shades)
    
  def draw_shades(self,widget,gc):
    self.init_shades(widget.window.get_size())
    widget.window.draw_pixbuf(gc,self.shades,0,0,0,0)
   
  def expose_event(self,widget,event):
    if self.im:

      if self.mode==0:
        if allTexlas.get_classifier()==None:
          drawim=self.get_drawim(widget,self.gc)
          (xd,yd)=drawim.get_size()
          self.drawim_size=(xd,yd)
          widget.window.draw_drawable(self.gc,drawim,0,0,0,0,xd,yd)
        else:
          self.draw_shades(widget,self.gc)
      else:      
        mask=self.im.mask
        drawim=self.get_drawim(widget,self.gc)
        (xd,yd)=drawim.get_size()
        self.drawim_size=(xd,yd)
        widget.window.draw_drawable(self.gc,drawim,0,0,0,0,xd,yd)
        mask.draw_on_window(widget,self.gcg,self.scale,(0,0,xd,yd))
        if self.l:
          self.gc.set_rgb_fg_color(gtk.gdk.Color(0,0,0))      
          widget.window.draw_polygon(self.gc,False,self.l)

    elif self.multiple:
      self.pl.set_text("%d images"%len(self.multiple))
      widget.window.draw_layout(self.gc,0,20,self.pl)
      
  def on_realize(self,widget):
    self.gc=widget.window.new_gc()
    self.gcg=make_mask_gc(widget)
    pcont=gtk.Widget.create_pango_context(widget)
    self.pl=pango.Layout(pcont)
    
  def on_button_press(self,widget,ev):
    if self.im and self.im.mask.can_edit(self.mode):
      self.l=[]
      self.queue_draw()
    else:
      self.l=None
    
  def on_motion_notify(self,widget,ev):
    if self.l==None: return
    # print ev.get_coords()
    co=ev.get_coords()
    (xmax,ymax)=self.drawim_size
    
    self.l.append((clamp(0,int(co[0]),xmax),
                   clamp(0,int(co[1]),ymax)))
    self.queue_draw()
   
  def on_button_release(self,widget,ev):
    if self.l==None: return    
    if self.mode in (VectorMask.POS,VectorMask.NEG):
      scale=1.0/self.scale
      simplify_curve(self.l)
      curve=self.im.mask.cur_curve(self.mode)      
      curve.append([(x*scale,y*scale) for (x,y) in self.l])
    self.l=None
    database.change_mask.broadcast(None,self.im)

  def clear_path(self,arg):    
    if self.mode in (VectorMask.POS,VectorMask.NEG):
      curve=self.im.mask.cur_curve(self.mode)
      try:
        del curve[-1]
      except IndexError:
        return
    database.change_mask.broadcast(None,self.im)


  def toggle_from_opp(self,on,im,mask):
    if on:
      if self.mode==VectorMask.POS:
        mask.copied=VectorMask.NEG
      if self.mode==VectorMask.NEG:
        mask.copied=VectorMask.POS
      curve=mask.cur_curve(self.mode)
      del curve[:]
    else:
      mask.copied=0
    database.change_mask.broadcast(None,im)
    
  def get_from_opp(self,arg):
    if self.mode==0:
      return
    on=arg.get_active()
    # print "get_from_opp ",arg.get_active()

    if self.im:
      self.toggle_from_opp(on,self.im,self.im.mask)

    if self.multiple:
      for im in self.multiple:
        self.toggle_from_opp(on,im,im.mask)
       

  def on_mask_change(self,arg):
    if self.im:
      if self.im.mask==None: self.im.mask=VectorMask()
    elif self.multiple:
      for im in self.multiple:
        if im.mask==None: im.mask=VectorMask()        
    self.set_states()
    self.queue_draw()

  def on_classifier_change(self,*arg):
    if self.mode==0:
      self.shades=None
      self.queue_draw()
