


from TextureClassifier import *

from extra_functions import *

import cPickle,pdb,errno



# replace FeaturesContainer's constructor

FeaturesContainer.ref_init=FeaturesContainer.__init__
def fcinit(self,*args):  
  if len(args)==1 and type(args[0])==type([]):
    self.ref_init()
    self.addFromList(args[0])
  else:
    self.ref_init(*args)

FeaturesContainer.__init__=fcinit

# add a few methods

FeatureVector.gt_tex=lambda self:self.gtlabel>0.5

FeatureVector.f_tex=lambda self:self.flabel>0.5



if not FeaturesComputer.hasLava():
  print "Lava not available"

def make_fc():
  # mountain vs. beach
  
  pp=FeaturesComputer()
  pp.randomPatches=True
  
  pp.npatch=1500
  pp.addhue=True
  pp.scone=6
  pp.winsize=16
  pp.doHaar=False
  # pp.scaleBase=(1.2)**(-2)
  pp.scaleBase=0.7
  
  pp.doAngNorm=True
  pp.doAffNorm=True
  pp.maxnpix=1000000 # image no bigger than ~ 1024*768
  pp.dets=FeaturesComputer.HL
  # pp.frimclaDesc=False
   
  return pp




class TextureManager(DataDir):
  """ manages the texture descriptions """
  
  def __init__(self):
    DataDir.__init__(self,"texdescs")
    self.np=[]
    # self.fc=InterestFeaturesComputer(self.pp)
    self.fc=make_fc()
    self.ep=ExtraTreeParams()
    self.ep.noOfTrees=15
    self.classifier_cache={}

  def fname(self,n):
    return self.dirprefix+"%08d.featurelist"%n

  def compute_tex(self,im):
    pi=PreprocessedImage(im,None)
    pl=self.fc.computeFeatures(pi)
    pl.thisown=1
    n=len(self.np)
    print "%d patches"%len(pl)
    self.np.append(len(pl))
    pl.writeTo(open(self.fname(n),"w"))
    pl.clear()
    return n

  def get_npatch(self,i):
    return self.np[i]

  def get_tex(self,i):
    """ ownership of patches transferred!!! (free with pl.clear()) """
    pl=FeaturesContainer()
    pl.readFrom(open(self.fname(i),"r"))
    return pl
  
  def learn(self,ims): 
    allpl=FeaturesContainer()

    # build list of features to handle
    pll=FeaturesContainer()   
    
    for im in ims:
      imno=im[0]
      plc=self.get_tex(imno)
      pl=plc.toList()
      allpl.addFromList(pl) # remember those to be deleted
      if im[1] in ('allpos','allneg'):
        label=im[1]=='allpos' and LABEL_FG or LABEL_BG        
        for patch in pl:
          patch.classLabel=label        
        pll.addFromList(pl)
      elif im[1] in ('pos+inv','neg+inv'):
        label=im[1]=='pos+inv' and LABEL_FG or LABEL_BG
        mask=im[2]
        otherlabel=label==LABEL_FG and LABEL_BG or LABEL_FG
        plc.setGroundTruth(mask)
        for patch in pl:
          if patch.gtlabel==1.0:
            patch.classLabel=label
          else:
            patch.classLabel=otherlabel
        pll.addFromList(pl)        
      elif im[1]=='pos+neg':
        label=LABEL_FG
        plo=[]
        for mask in (im[2],im[3]): 
          if mask:
            plc.setGroundTruth(mask)
            for patch in pl:
              if patch.gtlabel==1.0:
                patch.classLabel=label
                plo.append(patch)
          label=LABEL_BG
        pll.addFromList(plo)
    # pdb.set_trace()

    npos=nneg=0
    for patch in pll.toList():
      if patch.classLabel==LABEL_FG: npos+=1
      elif patch.classLabel==LABEL_BG: nneg+=1
      else: assert False
    print "learn from %dp+%dn=%d patches"%(
      npos,nneg,len(pll))

    # build the classifier
    cl=TextureClassifier(pll,self.ep)
    print
    allpl.clear()

    return cl

  def strong_lims(self,npos,nneg):
    fpos=npos/float(nneg+npos)
    return (0.4*fpos,1-0.4*(1-fpos))

  def strong_lims_x(self,npos,nneg):
    return (0.2,1-0.2)

  def classify_patches(self,cl,pl):
    # conversion to C++ of next function
    (strong_lim_inf,strong_lim_sup)=self.strong_lims(cl.npos,cl.nneg)
    (n_classified,n_strong_pos,n_strong_neg,f_prop)=cl.classifyAndStats(
      pl,strong_lim_inf,strong_lim_sup)
    # print (n_classified,n_strong_pos,n_strong_neg,f_prop)
    if n_strong_neg+n_strong_pos<5:
      print "too few strong patches"
      f_prop=None
    else:
      #print "nb strong patches: %dp+%dn=%d/%d"%(
      #  n_strong_pos,n_strong_neg,n_strong_neg+n_strong_pos,n_classified)
      pass
    return f_prop

  def classify_patches_slow(self,cl,pl):
    """ """
    (strong_lim_inf,strong_lim_sup)=self.strong_lims(cl.npos,cl.nneg)
    # print "strong lim: %g-%g"%(strong_lim_inf,strong_lim_sup)
    cl.classify(pl)
    pll=pl.toList()
    pll_classified=[patch for patch in pll if patch.flabel!=-1]
    pll_strong=[patch for patch in pll_classified if
                patch.flabel<strong_lim_inf or patch.flabel>strong_lim_sup]
    if len(pll_strong)<5:
      print "too few strong patches"
      f_prop=None
    else:
      f_prop=reduce(lambda accu,patch:accu+patch.flabel,pll_classified,0.0)/len(pll_classified)
      print "nb strong patches: %dp+%dn=%d/%d"%(
        len([patch for patch in pll_classified if patch.f_tex()]),
        len([patch for patch in pll_classified if not patch.f_tex()]),    
        len(pll_strong),len(pll))
    return f_prop

  def classify(self,cl,i):
    f_prop=self.get_cache_entry(cl,i)
    if f_prop!='not in cache':
      return f_prop
    pl=self.get_tex(i)
    f_prop=self.classify_patches(cl,pl)
    self.add_cache_entry(cl,i,f_prop)
    pl.clear()
    return f_prop

  def show_patch_classes(self,cl,i,scale,ci):
    pl=self.get_tex(i)  
    cl.classify(pl)
    pl.showVotes(scale,ci)
    pl.clear()


  ##############################################"
  # I/O

  
  def save_to(self,fname):
    self.prepare_store(fname)
    f=open(fname,"w")
    # dirprefix not stored as it depends on yorg name
    cPickle.dump(obj_to_dict(self,['np']),f)

  def load_from(self,fname):
    f=open(fname,"r")
    dict_to_obj(cPickle.load(f),self)
    self.istmpdir=False
    self.dirprefix=self.make_dirprefix(fname)
    # cache not stored
    self.classifier_cache={}

  ############################################
  ##  cache (classification results for each image) management
  ##  line = results for a given image
    
  def get_cache_line(self,cl):
    try:
      line=self.classifier_cache[cl]
    except KeyError:
      line={}
      self.classifier_cache[cl]=line
    return line
    
  def add_cache_entry(self,cl,i,f_prop):
    line=self.get_cache_line(cl)
    if line.has_key(i):
      print "warning, cl %s, imno %d result %g replaced by %g"%(
        cl,i,line[i],f_prop)
    line[i]=f_prop
    
  def get_cache_entry(self,cl,i):
    line=self.get_cache_line(cl)
    try:
      return line[i]
    except KeyError:
      return 'not in cache' # None is a valid cache entry

  def invalidate_cache_line(self,cl):
    """when classifier is removed or updated"""
    try:
      del self.classifier_cache[cl]
    except KeyError:
      pass
  


textureManager=TextureManager()



class VectorMask:
  """
  copied=-1: empty and read-only
  copied=0: pos and neg set independently
  copied=POS: neg is not(pos) (neg_curves unused)
  copied=NEG: pos is not(neg) (pos_curves unused)
  pos_curves: array of curves
     curve: array of (x,y) tuples in image coordinates
  neg_curves:
  """
  POS=1
  NEG=2
  pos_col=gtk.gdk.Color(0,65535,0)
  neg_col=gtk.gdk.Color(65535,0,0)


  def __init__(self):
    self.copied=0
    self.pos_curves=[]
    self.neg_curves=[]

  def can_edit(self,editmode):
    return (self.copied>=0 and
            (editmode==VectorMask.POS and self.copied!=VectorMask.NEG or
             editmode==VectorMask.NEG and self.copied!=VectorMask.POS))

  def is_read_only(self):
    return self.copied<0;

  def cur_curve(self,editmode):
    if editmode==VectorMask.POS: return self.pos_curves
    if editmode==VectorMask.NEG: return self.neg_curves

    
  def is_copied(self,editmode):
    if (editmode not in (VectorMask.POS,VectorMask.NEG) or
        self.copied<1): return False
    return self.copied!=editmode

  def is_clear(self):
    return (self.copied<0 or
            (self.copied==0 and self.pos_curves==[] and self.neg_curves==[]))

  def has_some(self,editmode):
    if self.copied<0:
      return False
    elif self.copied==0:
      return self.cur_curve(editmode)!=[]
    else:
      return self.copied!=editmode

  def clamp_curves(self,xd,yd):
    for curve in self.pos_curves:
      clamp_curve(curve,xd,yd)
    for curve in self.neg_curves:
      clamp_curve(curve,xd,yd)

  def get_as_image(self,xd,yd,scale,editmode):
    gi=MaskImage(xd,yd)
    gi.clear()
    for curve in self.cur_curve(editmode):
      fillPolygon(gi,255,[(x*scale,y*scale) for (x,y) in curve])
    return gi

  def draw_filled_poly(self,widget,gc,curve,scale,(x0,y0)):
    ci=[(x0+int(x*scale),y0+int(y*scale)) for (x,y) in curve]
    gc.set_fill(gtk.gdk.STIPPLED)
    widget.window.draw_polygon(gc,True,ci)
    gc.set_fill(gtk.gdk.SOLID)
    widget.window.draw_polygon(gc,False,ci)

  def draw_polys(self,widget,gc,curves,scale,color,p0):
    gc.set_rgb_fg_color(color)     
    for curve in curves:
      self.draw_filled_poly(widget,gc,curve,scale,p0)
        
  def draw_rect(self,widget,gc,(x0,y0,xd,yd),color):
    gc.set_rgb_fg_color(color)    
    gc.set_fill(gtk.gdk.STIPPLED)
    widget.window.draw_rectangle(gc,True,x0,y0,xd,yd)
    
  def draw_on_window(self,widget,gc,scale,bbox):
    # print self
    p0=(bbox[0],bbox[1])
    if self.copied==0:
      self.draw_polys(widget,gc,self.pos_curves,scale,VectorMask.pos_col,p0)
      self.draw_polys(widget,gc,self.neg_curves,scale,VectorMask.neg_col,p0)
    elif self.copied==VectorMask.NEG:
      self.draw_rect(widget,gc,bbox,VectorMask.pos_col)
      self.draw_polys(widget,gc,self.neg_curves,scale,VectorMask.neg_col,p0)
    elif self.copied==VectorMask.POS:
      self.draw_rect(widget,gc,bbox,VectorMask.neg_col)
      self.draw_polys(widget,gc,self.pos_curves,scale,VectorMask.pos_col,p0)

  def __str__(self):
    return "VectorMask: copied=%d, %dp+%dn curves"%(
      self.copied,len(self.pos_curves),len(self.neg_curves))

ReadOnlyVectorMask=VectorMask()
ReadOnlyVectorMask.copied=-1

# use for display

def make_stipple_gc(widget,bm,xd,yd,colorname):
  gc=widget.window.new_gc()
  xbm=gtk.gdk.bitmap_create_from_data(
    widget.window,
    bm,xd,yd)
  gc.set_stipple(xbm)
  gc.set_rgb_fg_color(gtk.gdk.color_parse(colorname))
  return gc

def make_mask_gc(widget):
  return make_stipple_gc(widget,'\x02\x08',2,2,'Green')

# self.gcgi=self.make_stipple_gc(widget,'\x04\x01',2,2,'Red')


class ImWithMask:
  """
  size = (xd,yd) of the image (useful to render the masks)
  scale = scale at which the texture detection was done
  mask = VectorMask (!=None)
  """
  def __init__(self,mask,size,scale):
    self.size=size
    self.scale=scale
    self.mask=mask

class Texla:
  """ set of images with masks + maybe an associated classifier
  Texla = TEXture cLAssifier any suggestion for a better name welcome

  name = name that appears in the array (should be distinct)
  d = "ro" : read-only classifier
  d = texno -> ImWithMask map : mask built from these image segments

  cl = None : no classifier computed (yet)
  cl = ExtraTreeClassifier : corresponding classifier
  
  """
  def __init__(self,name='default',d=None):
    self.name=name
    if d==None:
      self.d={}
    else:
      self.d=d
    self.cl=None

  def store(self,f,basefname):
    cPickle.dump(obj_to_dict(self,['d','name']),f)
    f.write(self.cl and 'c' or 'n')
    if self.cl:
      self.cl.writeToFile(basefname+'_%s.frimcla'%self.name)


  def load(self,f,basefname):
    " call just after constructor "
    dict_to_obj(cPickle.load(f),self)
    has_cl=f.read(1)
    if has_cl=='c':
      self.cl=TextureClassifier(basefname+'_%s.frimcla'%self.name)
    elif has_cl=='n':
      pass
    else:
      raise 'bad has_cl'
  
  

def get_index_dat_line(f):
  l=f.readline()
  # print l
  if not l: return None
  i=l.find('#')
  if i>=0: l=l[:i]
  l=l.strip()
  if l=='': return get_index_dat_line(f)
  return l


def load_index_dat(fname):
  basename=fname[:fname.rfind("/")+1]
  f=open(fname,"r")
  if get_index_dat_line(f)!="texture classifier 102":
    raise "wrong header"
  tcsets=[]
  while True:
    l=get_index_dat_line(f)
    if not l: break
    if l!="BeginSet":
      raise "parse error 5"
    fc=FeaturesComputer()
    fc.readFrom(f)
    tcs=[]
    while True:
      l=get_index_dat_line(f)
      if not l: break
      if l=="EndSet":
        break
      try:
        [name,fname]=l.split()
      except ValueError:
        raise "parse error 2"
      fname=basename+fname
      print "loading class %s from %s"%(name,fname)
      if os.access(fname,os.R_OK):
        tc=TextureClassifier(fname)
      else:
        print "  could not load"
        tc=None
      tcs.append((name,tc))
    tcsets.append((fc,tcs))
  return tcsets


class AllTexlas:

  def __init__(self):
    self.texlas=[Texla()]
    self.change=ChangeBroadcaster() # state change

    self.classifier_change=ChangeBroadcaster()
    
    self.selected=0
    self.ro_texlas=[] # read-only texlas 

  def n_pos_neg(self,d):    
    n_pos=n_neg=0
    n_pos_mask=n_neg_mask=0
    for i in d.itervalues():
      mask=i.mask
      if mask and not mask.is_clear():
        if mask.copied==0:
          if mask.pos_curves: n_pos_mask+=1
          if mask.neg_curves: n_neg_mask+=1
        elif mask.copied==VectorMask.POS:
          if mask.pos_curves:
            n_pos_mask+=1
            n_neg_mask+=1
          else:
            n_neg+=1
        elif mask.copied==VectorMask.NEG:
          if mask.neg_curves:
            n_pos_mask+=1
            n_neg_mask+=1
          else:
            n_pos+=1
    return (n_pos,n_neg,n_pos_mask,n_neg_mask)

  def shortnum(self,n,nm):
    if n>0:
      if nm>0: return "%d+%dm"%(n,nm)
      else: return "%d"%n
    else:
      if nm>0: return "%dm"%nm
      else: return ""
   

  def get_as_tags(self):
    tags=[]
    for i in self.texlas:
      if i.d=='ro':
        s_pos='-'
        s_neg='-'
      else:
        (n_pos,n_neg,n_pos_mask,n_neg_mask)=self.n_pos_neg(i.d)
        s_pos=self.shortnum(n_pos,n_pos_mask)
        s_neg=self.shortnum(n_neg,n_neg_mask)
      tags.append((i.name,s_pos,s_neg,i.cl!=None and '*' or ' '))
    return tags

  def set_mask(self,key,mask,size,scale):
    d=self.texlas[self.selected].d
    assert d!='ro'
    if not mask or mask.is_clear():
      try:
        del d[key]
      except KeyError:
        pass
    else:
      d[key]=ImWithMask(mask,size,scale)
    self.change.broadcast()

  def add_new_texla(self):
    # pdb.set_trace()
    j=1
    while "new_%d"%j in [i.name for i in self.texlas]:
      j+=1
    name="new_%d"%j
    # print "adding",name
    self.texlas.append(Texla(name))
    self.change.broadcast()

  def del_texla(self):
    if self.texlas[self.selected].cl:
      textureManager.invalidate_cache_line(self.texlas[self.selected].cl)
    del self.texlas[self.selected]
    if self.selected==len(self.texlas): self.selected-=1    
    self.classifier_change.broadcast('del texla')
    self.change.broadcast()

  def n_texla(self):
    return len(self.texlas)

  def mask_map(self,no):
    assert self.texlas[no].d!='ro'
    m={}
    for imno,i in self.texlas[no].d.iteritems():
      m[imno]=i.mask
    return m

  def is_read_only(self):
    return self.texlas[self.selected].d=='ro'
  
  def set_selected(self,no):
    self.selected=no
    self.classifier_change.broadcast('change selection')
    # self.change.broadcast()
    
  def set_name(self,new_name):
    self.texlas[self.selected].name=new_name
    self.change.broadcast()

  def to_bm_masks(self,d):
    """ get a list of images + associated masks """
    ims=[]
    for  (imno,i) in d.iteritems():
      mask=i.mask
      (xd,yd)=i.size
      scale=i.scale
      (xd,yd)=(int(xd*scale),int(yd*scale))
      #scale=im.get_tex_scale()
      #(xd,yd)=im.size
      #(xd,yd)=(int(xd*scale),int(yd*scale))
      if mask.copied==0:          
        if mask.pos_curves!=[]:            
          pos_mask=mask.get_as_image(xd,yd,scale,VectorMask.POS)
        else:
          pos_mask=None
        if mask.neg_curves!=[]:
          neg_mask=mask.get_as_image(xd,yd,scale,VectorMask.NEG)            
        else:
          neg_mask=None
        ims.append((imno,'pos+neg',pos_mask,neg_mask))          
      elif mask.copied==VectorMask.POS:
        if mask.pos_curves==[]:
          ims.append((imno,'allneg'))
        else:          
          pos_mask=mask.get_as_image(xd,yd,scale,VectorMask.POS)
          ims.append((imno,'pos+inv',pos_mask))
      elif mask.copied==VectorMask.NEG:
        if mask.neg_curves==[]:
          ims.append((imno,'allpos'))          
        else:
          neg_mask=mask.get_as_image(xd,yd,scale,VectorMask.NEG)
          ims.append((imno,'neg+inv',neg_mask))                  
    return ims

  def can_learn(self):
    if self.texlas[self.selected].d=='ro': return False
    (n_pos,n_neg,n_pos_mask,n_neg_mask)=self.n_pos_neg(
      self.texlas[self.selected].d)
    return n_pos+n_pos_mask>0 and n_neg+n_neg_mask>0
  
  def do_learn(self):
    # pdb.set_trace()
    i=self.texlas[self.selected]
    assert i.d!='ro'
    ims=self.to_bm_masks(i.d)
    if ims==[]: return
    cl=textureManager.learn(ims)
    if self.texlas[self.selected].cl:
      textureManager.invalidate_cache_line(self.texlas[self.selected].cl)
    self.texlas[self.selected].cl=cl
    
    self.classifier_change.broadcast('learning done')
    
  def all_names(self):
    return [w.name for w in self.texlas]

  def get_classifier(self,no=-1):
    if no==-1: no=self.selected
    return self.texlas[no].cl

  def get_classifier_dict(self):
    cd={}
    for i in self.texlas:
      if i.cl!=None:
        cd[i.name]=i.cl
    return cd


  #######################################################################
  # I/O
  # the texla array is stored by serializing it, but of course the
  # classifier C pointers (if not None) will be invalid when loading.
  # The classifiers are in separate files, in the correct order
  # the npos and nneg fields are not stored by the TextureClassifier object
  # so they are set in a separate dictionary
  
  def get_classifier_cache_dict(self):
    """ allTexlas is serializing the classifier caches on behalf of the
    textureManager (because it knows about the defined classifiers) """
    d={}
    for texla in self.texlas:
      try:
        d[texla.name]=textureManager.classifier_cache[texla.cl]
        # print "saving cache for",texla.name
      except KeyError:
        pass
    return d

  def set_classifier_cache_dict(self,d):    
    for texla in self.texlas:
      try:
        textureManager.classifier_cache[texla.cl]=d[texla.name]
        # print "restoring cache for",texla.name
      except KeyError:
        pass       
    

  def save_to(self,fname):
    f=open(fname+'_index.raw',"w")
    classifier_cache=self.get_classifier_cache_dict()
    pos_neg=[]
    for i in self.texlas:
      if i.d!='ro' and i.cl!=None:
        pos_neg.append((i.cl.npos,i.cl.nneg))
    cPickle.dump({'pos_neg':pos_neg,'classifier_cache':classifier_cache,
                  'n_texlas':len(self.texlas)},f)
    for texla in self.texlas:
      texla.store(f,fname)
      
    n=0
    for i in self.texlas:
      if i.d!='ro' and i.cl!=None:
        i.cl.writeToFile(fname+'_%d.frimcla'%n)
        n+=1
    
  def load_from(self,fname):    
    f=open(fname+'_index.raw',"r")
    dic=cPickle.load(f)
    
    self.texlas=[]
    for i in range(dic['n_texlas']):
      texla=Texla()
      texla.load(f,fname)
      self.texlas.append(texla)
        
    self.selected=0 
    # remove stored ro texlas
    n=0
    while n<len(self.texlas):
      texla=self.texlas[n]
      if texla.d=='ro':
        del self.texlas[n]
      else:
        n+=1
    # add our own
    self.texlas=self.ro_texlas+self.texlas

    n=0
    pos_neg=dic['pos_neg']
    for i in self.texlas:
      if i.d!='ro' and i.cl!=None:
        (i.cl.npos,i.cl.nneg)=pos_neg[n]
        n+=1

    self.set_classifier_cache_dict(dic['classifier_cache'])
    
    self.change.broadcast()
    self.classifier_change.broadcast('loaded')

  def load_ro_texlas(self,index_name):
    try:
      tcsets=load_index_dat(index_name)      
    except IOError,e:
      if e.errno==errno.ENOENT:
        print "could not open read-only texlas index ",index_name        
        return 
      raise e
    # use only first tcsets
    tcset=tcsets[0]
    # drop featurescomputer
    tcs=tcset[1]
    for name,tc in tcs:
      texla=Texla(name,'ro')
      texla.cl=tc
      self.ro_texlas.append(texla)
    self.texlas+=self.ro_texlas
    
    
allTexlas=AllTexlas()
allTexlas.load_ro_texlas("default_classifiers/index.dat")
